package uk.ac.leeds.ccg.grids;
/**
 * -------------------------------------------------------------------------
 * This file is distributed as part of the uk.ac.leeds.ccg.grids library.
 * This library is a resource for manipulating spatial data stored as grids.
 * Copyright (C) 2002  Andy Turner, CCG, a.turner@geog.leeds.ac.uk
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 */
import uk.ac.leeds.ccg.grids.Utilities;
import uk.ac.leeds.ccg.raster.Raster;
import uk.ac.leeds.ccg.geotools.GeoRectangle;
import java.io.File;
import java.io.RandomAccessFile;
import java.io.PrintWriter;
import java.io.BufferedWriter;
import java.io.BufferedReader;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.io.InputStreamReader;
import java.io.StreamTokenizer;
import java.awt.Graphics2D;
import java.awt.Toolkit;
import java.awt.Image;
import java.awt.image.BufferedImage;
import java.awt.image.MemoryImageSource;
import java.util.Vector;
import java.util.HashSet;
import java.util.TreeMap;
import java.util.Iterator;

/**
 * ---------------------------
 * Last updated on 25 Nov 2002
 * @author  andyt
 * @version 1.0
 * ---------------------------
 * Description:
 * This class contains methods for manipulating 2D square cell grids containing double values.
 * The values are referenced to square shaped partitions of the spatial extent in a Euclidean
 * two-dimensional geometry defined by orthoganol X and Y axes. The spatial extent of the data
 * is given by a rectangle with its sides aligned with the axes. The values are also known as
 * cells and are aligned in rows and columns.
 *
 * Known extentions:
 * 1. Grid2DSquareCellDouble
 * 2. Grid2DSquareCellDoubleFile
 */

public abstract class AbstractGrid2DSquareCellDouble {
    
    /**
     * ------
     * Fields
     * ------
     */
    
    /**
     * For storing the number of columns in the grid.
     */
    protected int ncols;
    
    /**
     * For storing the number of rows in the grid.
     */
    protected int nrows;
    
    /**
     * If the Y axis is verticle and the X axis is horizontal then xllcorner is
     * for storing the left x-axis cordinate of the grid.
     */
    protected double xllcorner = 0.0d;
    
    /**
     * If the Y axis is verticle and the X axis is horizontal then yllcorner is
     * for storing the bottom y-axis cordinate of the grid.
     */
    protected double yllcorner = 0.0d;
    
    /**
     * For storing the cellsize - the width (or height) of each cell.
     */
    protected double cellsize = 1.0d;
    
    /**
     * For storing the noDataValue of the grid
     */
    protected double noDataValue = Double.NaN;
    
    /**
     * Reference to the instance of the GridStatistics object
     */
    protected AbstractGridStatistics gridStatistics = null;
    
    
    /**
     * =============
     * Basic methods
     * =============
     */
    
    /**
     * Returns a string description of the grid.
     */
    public abstract String toString();
    
    /**
     * Returns the dataType of the grid.
     * dataType = 0 means the data is held in memory
     * dataType = 1 means the data is held in a file
     * dataType = 2 means the data is held in a database
     */
    public abstract int getDataType();
    
    /**
     * Clears collection or deletes file that stores the grid values
     */
    public abstract void clear();
    
    /**
     * Returns the reference to the instance of the GridStatistics object
     */
    public final AbstractGridStatistics getGridStatistics() {
        return this.gridStatistics;
    }
    
    /**
     *  Initialises the GridStatistics which accompanies the grid
     */
    protected final AbstractGridStatistics initGridStatistics( int gridStatisticsType ) {
        AbstractGridStatistics stats;
        switch ( gridStatisticsType ) {
            case 0 :
                stats = new GridStatistics0( this );
                break;
            default :
                stats = new GridStatistics1( this );
                break;
        }
        setGridStatistics( stats );
        return stats;
    }
    
    /**
     * Replaces gridStatistics with newGridStatistics and returns the former gridStatistics
     */
    protected final AbstractGridStatistics setGridStatistics( AbstractGridStatistics newGridStatistics ) {
        AbstractGridStatistics oldGridStatistics = getGridStatistics();
        this.gridStatistics = newGridStatistics;
        return oldGridStatistics;
    }
    
    /**
     * Returns the number of columns in the grid.
     */
    public final int getNcols() {
        return this.ncols;
    }
    
    /**
     * Initialises ncols.
     */
    protected final void initNcols( int ncols ) {
        this.ncols = ncols;
    }
    
    /**
     * Returns the number of rows in the grid.
     */
    public final int getNrows() {
        return this.nrows;
    }
    
    /**
     * Initialises nrows.
     */
    protected final void initNrows( int nrows ) {
        this.nrows = nrows;
    }
    
    /**
     * Returns the x-coordinate of the lower left corner of the grid.
     */
    public final double getXllcorner() {
        return this.xllcorner;
    }
    
    /**
     * This effectively shifts the grid to a new origin
     * Replaces xllcorner with newXllcorner and returns its former value
     */
    public final double setXllcorner( double newXllcorner ) {
        double oldXllcorner = getXllcorner();
        this.xllcorner = newXllcorner;
        return oldXllcorner;
    }
    
    /**
     * Returns the y-coordinate of the lower left corner of the grid
     */
    public final double getYllcorner() {
        return this.yllcorner;
    }
    
    /**
     * This effectively shifts the grid to a new origin
     * Replaces yllcorner with newYllcorner and returns its former value
     */
    public final double setYllcorner( double newYllcorner ) {
        double oldYllcorner = getYllcorner();
        this.yllcorner = newYllcorner;
        return oldYllcorner;
    }
    
    /**
     * Returns the cell width (cell height) of the grid cells
     */
    public final double getCellsize() {
        return this.cellsize;
    }
    
    /**
     * This effectively scales the grid keeping the origin fixed
     * Replaces cellsize with newCellsize and returns its former value
     */
    public double setCellsize( double newCellsize ) {
        double oldCellsize = getCellsize();
        this.cellsize = newCellsize;
        return oldCellsize;
    }
    
    /**
     * Returns the noDataValue for the grid cells.
     */
    public final double getNoDataValue() {
        return this.noDataValue;
    }
    
    /**
     * Initialises noDataValue
     */
    protected final void initNoDataValue( double noDataValue ) {
        this.noDataValue = noDataValue;
    }
    
    /**
     * Sets noDataValue to newNoDataValue replacing existing noDataValues with newNoDataValue
     */
    public void setNoDataValue( double newNoDataValue ) {
        double oldNoDataValue = getNoDataValue();
        if ( newNoDataValue != oldNoDataValue ) {
            int ncols = getNcols();
            int nrows = getNrows();
            double d1 = noDataValue;
            double cellasProportionOfGrid = 1 / ( double ) ( ncols * nrows );
            double sparseness = 1.0d;
            double sum = 0.0d;
            double min = Double.POSITIVE_INFINITY;
            int minCount = nrows * ncols;
            double max = Double.NEGATIVE_INFINITY;
            int maxCount = minCount;
            for ( int i = 0 ; i < nrows ; i ++ ) {
                for ( int j = 0 ; j < ncols ; j ++ ) {
                    d1 = getCell( i, j );
                    if ( d1 == newNoDataValue || d1 == oldNoDataValue ) {
                        sparseness -= cellasProportionOfGrid;
                        initCell( i, j, newNoDataValue );
                    } else {
                        sum += d1;
                        if ( d1 < min ) { min = d1; minCount = 1; }
                        if ( d1 == min ) { minCount += 1; }
                        if ( d1 > max ) { max = d1; maxCount = 1; }
                        if ( d1 == max ) { maxCount += 1; }
                    }
                }
            }
            initNoDataValue( newNoDataValue );
            AbstractGridStatistics stats = getGridStatistics();
            stats.setSparseness( sparseness );
            stats.setSum( sum );
            stats.setMinCount( minCount );
            stats.setMin( min );
            stats.setMaxCount( maxCount );
            stats.setMax( max );
            stats.setIsUpToDate( true );
        }
        if ( this instanceof Grid2DSquareCellDouble ) {
            Grid2DSquareCellDouble grid = ( Grid2DSquareCellDouble ) this;
            grid.optimiseCollection( "memory" );
        }
    }
    
    /**
     * Sets numbers in the range [d1,d2] to noDataValue. All existing noDataValues
     * remain as noDataValues.
     */
    public void setToNoDataValue( double d1, double d2 ) {
        double d3;
        int ncols = getNcols();
        int nrows = getNrows();
        double noDataValue = getNoDataValue();
        for ( int i = 0; i < nrows * ncols; i ++ ) {
            d3 = getCell( i );
            if ( d3 >= d1 && d3 <= d2 ) {
                setCell( i, noDataValue );
            }
        }
    }
    
    /**
     * Sets all cells with centroids between [ xmin, xmax ] and [ ymin, ymax ] to noDataValues
     */
    public void setToNoDataValue( double xmin, double ymin, double xmax, double ymax ) {
        int startRowIndex = getRowIndex( ymax );
        int endRowIndex = getRowIndex( ymin );
        int startColIndex = getColIndex( xmin );
        int endColIndex = getColIndex( xmax );
        double noDataValue = getNoDataValue();
        for ( int i = startRowIndex; i <= endRowIndex; i ++ ) {
            for ( int j = startColIndex; j <= endColIndex; j ++ ) {
                setCell( i, j, noDataValue );
            }
        }
    }
    
    /**
     * Returns the height of the grid
     */
    public final double getHeight() {
        return getNrows() * getCellsize();
    }
    
    /**
     * Returns the width of the grid
     */
    public final double getWidth() {
        return getNcols() * getCellsize();
    }
    
    /**
     * Returns true iff x-coordinate x, y-coordinate y is in the Grid.
     */
    public final boolean inGrid( double x, double y ) {
        double xllcorner = getXllcorner();
        double yllcorner = getYllcorner();
        double cellsize = getCellsize();
        return ( x >= xllcorner && x < ( xllcorner + ( getNcols() * cellsize ) ) && y >= yllcorner && y < ( yllcorner + ( getNrows() * cellsize ) ) );
    }
    
    /**
     * Returns true iff row i, column j is in the Grid.
     */
    protected final boolean inGrid( int i, int j ) {
        return ( i >= 0 && i < getNrows() && j >= 0 && j < getNcols() );
    }
    
    /**
     * Returns true iff cellID is a valid cellID.
     */
    protected final boolean inGrid( int cellID ) {
        return ( cellID >= 0 && cellID < getNrows() * getNcols() );
    }
    
    /**
     * Returns the cellID if row and col are respectively row and column indexes
     * that reference a cell in the grid, else returns Integer.MIN_VALUE.
     * Note:
     * Cells are numbered from 0 to getNrows()*getNcols()-1 in row major order and
     * only have an ID if they are in the grid.
     */
    public final int getCellID( int row, int col ) {
        if ( inGrid( row, col ) ) {
            return ( row * getNcols() ) + col;
        }
        return Integer.MIN_VALUE;
    }
    public final int getCellID( double x, double y ) {
        return getCellID( getRowIndex( y ), getColIndex( x ) );
    }
    
    /**
     * Returns the row index for the input y-coordinate. Rows in the grid are
     * indexed from 0 to nrows-1.
     */
    public final int getRowIndex( double y ) {
        double cellsize = getCellsize();
        return ( ( int ) Math.floor( ( ( ( double ) getNrows() * cellsize ) + getYllcorner() - y ) / cellsize ) );
    }
    /**
     * Returns the row index for the input cellID.
     */
    public final int getRowIndex( int cellID ) {
        if ( inGrid( cellID ) ) {
            return ( int ) cellID / getNcols();
        }
        return Integer.MIN_VALUE; // This should not happen given a valid cellID.
    }
    
    /**
     * Returns the column index for the input x-coordinate. Columns in the grid are
     * indexed from 0 to ncols-1.
     */
    public final int getColIndex( double x ) {
        return ( int ) Math.floor( ( x - getXllcorner() ) / getCellsize() );
    }
    /**
     * Returns the column index for the input cellID.
     */
    public final int getColIndex( int cellID ) {
        if ( inGrid( cellID ) ) {
            return cellID - ( getRowIndex( cellID ) * getNcols() );
        }
        return Integer.MIN_VALUE; // This should not happen given a valid cellID.
    }
    
    /**
     * Returns the x-coordinate of the cell centroid.
     */
    public final double getCellX( int cellID ) {
        double cellsize = getCellsize();
        return getXllcorner() + ( cellsize * getColIndex( cellID ) ) + ( cellsize / 2.0d );
    }
    public final double getCellX( int i, int j ) {
        //return getCellX( getCellID( i, j ) );}
        double cellsize = getCellsize();
        return getXllcorner() + ( cellsize * j ) + ( cellsize / 2.0d );
    }
    
    /**
     * Returns the y-coordinate of the cell centroid.
     */
    public final double getCellY( int cellID ) {
        double cellsize = getCellsize();
        return getYllcorner() + ( cellsize * ( getNrows() - getRowIndex( cellID ) ) ) - ( cellsize / 2.0d );
    }
    public final double getCellY( int i, int j ) {
        //return getCellY( getCellID( i, j ) );
        double cellsize = getCellsize();
        return getYllcorner() + ( cellsize * ( getNrows() - i ) ) - ( cellsize / 2.0d );
    }
    
    /**
     * Returns the value at row i, column j.
     */
    public abstract double getCell( int i, int j );
    
    /**
     * Returns the value of the cell containing x-coordinate x, y-coordinate y
     */
    public double getCell( double x, double y ) {
        return getCell( getRowIndex( y ), getColIndex( x ) );
    }
    /**
     * Returns the value of the cell with cellID
     */
    public double getCell( int cellID ) {
        return getCell( getRowIndex( cellID ), getColIndex( cellID ) );
    }
    
    /**
     * Initilises the value at row i column j to d1.
     */
    protected abstract void initCell( int i, int j, double d1 );
    
    /**
     * Initialises the value at x-coordinate x, y-coordinate y to d1
     */
    protected final void initCell( double x, double y, double d1 ) {
        initCell( getRowIndex( y ), getColIndex( x ), d1 );
    }
    
    /**
     *Initialises the value at cellID to d1
     */
    protected final void initCell( int cellID, double d1 ) {
        initCell( getRowIndex( cellID ), getColIndex( cellID ), d1 );
    }
    
    /**
     * For setting the value of the cell at row i column j to d1 and returns the original value of the cell
     */
    public abstract double setCell( int i, int j, double d1 );
    
    /**
     * Sets the value at x-coordinate x, y-coordinate y to d1
     */
    public final double setCell( double x, double y, double d1 ) {
        return setCell( getRowIndex( y ), getColIndex( x ), d1 );
    }
    
    /**
     * Sets the value of the cell cellID to d1
     */
    public final double setCell( int cellID, double d1 ) {
        return setCell( getRowIndex( cellID ), getColIndex( cellID ), d1 );
    }
    
    /**
     * Adds d1 to cell at row i column j and returns the cell value that was replaced if cell is in Grid otherwise returns noDataValue.
     */
    public final double addToCell( int i, int j, double d1 ) {
        double noDataValue = getNoDataValue();
        if ( inGrid( i, j ) ) {
            double d2 = getCell( i, j );
            if ( d2 != noDataValue ) {
                if ( d1 != noDataValue ) {
                    setCell( i, j, d1 + d2 );
                } else {
                    setCell( i, j, d2 );
                }
            } else {
                if ( d1 != noDataValue ) {
                    setCell( i, j, d1 );
                }
            }
            return d2;
        }
        return noDataValue;
    }
    
    /**
     * Adds d1 to cell containing the x-coordinate x, y-coordinate y.
     */
    public final double addToCell( double x, double y, double d1 ) {
        return addToCell( getRowIndex( y ), getColIndex( x ), d1 );
    }
    
    /**
     * Adds d1 to cell with ID cellID
     */
    public final double addToCell( int cellID, double d1 ) {
        return addToCell( getRowIndex( cellID ), getColIndex( cellID ), d1 );
    }
    
    /**
     * Adds value to all cells in grid with IDs as Integers in cellIDs
     **/
    public final void addToGrid( AbstractGrid2DSquareCellDouble grid, HashSet cellIDs, double value ) {
        int cellID;
        double noDataValue = grid.getNoDataValue();
        Iterator iterator1 = cellIDs.iterator();
        double thisValue;
        while ( iterator1.hasNext() ) {
            cellID = ( ( Integer ) iterator1.next() ).intValue();
            thisValue = grid.getCell( cellID );
            if ( thisValue != noDataValue ) {
                grid.addToCell( cellID, value );
            }
        }
    }
    
    /**
     * Adds values to the grid from xYValue ascii file formatted as:
     * x y value
     * where x and y are the coordinates of the value to be added to the cell with ID getCellID( x, y )
     */
    public final void addToGrid( File xYValueFile ) {
        try {
            StreamTokenizer st = new StreamTokenizer( new BufferedReader( new InputStreamReader( new FileInputStream( xYValueFile ) ) ) );
            st.eolIsSignificant( false );
            st.parseNumbers();
            st.whitespaceChars( ',', ',' );
            st.wordChars( '"', '"' );
            int tokenType = st.nextToken();
            String alternator = new String( "x" );
            double x = 0.0d;
            double y = 0.0d;
            double value = 0.0d;
            while ( tokenType != StreamTokenizer.TT_EOF ) {
                switch ( tokenType ) {
                    case StreamTokenizer.TT_NUMBER :
                        if ( alternator.equals( "x" ) ) {
                            x = st.nval;
                            alternator = "y";
                        }
                        else if ( alternator.equals( "y" ) ) {
                            y = st.nval;
                            alternator = "value";
                        }
                        else {
                            addToCell( x, y, st.nval );
                            alternator = "x";
                        }
                        break;
                    default :
                        break;
                }
                tokenType = st.nextToken();
            }
        }
        catch ( java.io.IOException e ) {
            System.out.println( e  + " in addToGrid( xYValueFile ( " + xYValueFile.toString() + " ) )");
            System.exit( 0 );
        }
    }
    
    /**
     * Adds values to the grid from the inputGrid
     */
    public void addToGrid( AbstractGrid2DSquareCellDouble inputGrid ) {
        
        int nrows = getNrows();
        int ncols = getNcols();
        double xllcorner = getXllcorner();
        double yllcorner = getYllcorner();
        double cellsize = getCellsize();
        double noDataValue = getNoDataValue();
        
        int inputGridNrows = inputGrid.getNrows();
        int inputGridNcols = inputGrid.getNcols();
        double inputGridXllcorner = inputGrid.getXllcorner();
        double inputGridYllcorner = inputGrid.getYllcorner();
        double inputGridCellsize = inputGrid.getCellsize();
        double inputGridNoDataValue = inputGrid.getNoDataValue();
        
        Grid2DSquareCellDoubleFactory f = new Grid2DSquareCellDoubleFactory();
        Grid2DSquareCellDoubleFileFactory ff = new Grid2DSquareCellDoubleFileFactory();
        
        // If no intersection then do nothing
        if ( ( inputGridXllcorner > xllcorner + ( ( double ) ncols * cellsize ) ) ||
        ( inputGridYllcorner > yllcorner + ( ( double ) nrows * cellsize ) ) ||
        ( inputGridXllcorner + ( ( double ) inputGridNcols * inputGridCellsize ) < xllcorner ) ||
        ( inputGridYllcorner + ( ( double ) inputGridNrows * inputGridCellsize ) < yllcorner ) ) {
            System.out.println( "No intersection! in addToGrid( AbstractGrid2DSquareCellDouble(" + inputGrid.toString() + ") )" );
            return;
        }
        if ( inputGridCellsize < cellsize ) {
            AbstractGrid2DSquareCellDouble aggregateInputGrid = null;
            try {
                aggregateInputGrid = inputGrid.aggregate( cellsize, "sum", f, xllcorner, yllcorner, noDataValue );
            } catch ( java.lang.OutOfMemoryError e ) {
                aggregateInputGrid = inputGrid.aggregate( cellsize, "sum", ff, xllcorner, yllcorner, noDataValue );
            }
            addToGrid( aggregateInputGrid );
        } else {
            // If inputGridCellsize is the same and the grids align
            if ( inputGridCellsize == cellsize && ( ( inputGridXllcorner % cellsize ) == ( xllcorner % cellsize ) ) && ( ( inputGridYllcorner % cellsize ) == ( yllcorner % cellsize ) ) ) {
                double x;
                double y;
                for ( int i = 0; i < nrows; i ++ ) {
                    for ( int j = 0; j < nrows; j ++ ) {
                        x = getCellX( i, j );
                        y = getCellY( i, j );
                        addToCell( i, j, inputGrid.getCell( x, y ) );
                    }
                }
            } else {
                // Need to intersect
                AbstractGrid2DSquareCellDouble tempGrid1 = null;
                AbstractGrid2DSquareCellDouble tempGrid2 = null;
                try {
                    tempGrid1 = f.createGrid2DSquareCellDouble( nrows, ncols, xllcorner, yllcorner, cellsize, noDataValue, 1 );
                    tempGrid2 = f.createGrid2DSquareCellDouble( nrows, ncols, xllcorner, yllcorner, cellsize, noDataValue, 1 );
                } catch ( java.lang.OutOfMemoryError e ) {
                    tempGrid1 = ff.createGrid2DSquareCellDouble( nrows, ncols, xllcorner, yllcorner, cellsize, noDataValue, 0 );
                    tempGrid2 = ff.createGrid2DSquareCellDouble( nrows, ncols, xllcorner, yllcorner, cellsize, noDataValue, 0 );
                }
                double[] bounds = new double[ 4 ];
                int cellID1;
                int cellID2;
                int cellID3;
                int cellID4;
                double d1;
                double d2;
                double d3;
                double d4;
                double areaProportion = 0.0d;
                for ( int i = 0; i < nrows * ncols; i ++ ) {
                    bounds = getBounds( i );
                    cellID1 = inputGrid.getCellID( bounds[0], bounds[3] );
                    cellID2 = inputGrid.getCellID( bounds[2], bounds[3] );
                    cellID3 = inputGrid.getCellID( bounds[0], bounds[1] );
                    cellID4 = inputGrid.getCellID( bounds[2], bounds[1] );
                    if ( cellID1 == cellID2 && cellID2 == cellID3 ) {
                        d1 = inputGrid.getCell( cellID1 );
                        if ( d1 != inputGridNoDataValue ) {
                            areaProportion = ( cellsize * cellsize ) / ( inputGridCellsize * inputGridCellsize );
                            tempGrid1.addToCell( i, d1 * areaProportion );
                            tempGrid2.addToCell( i, areaProportion );
                        }
                    } else {
                        d1 = inputGrid.getCell( cellID1 );
                        d2 = inputGrid.getCell( cellID2 );
                        d3 = inputGrid.getCell( cellID3 );
                        d4 = inputGrid.getCell( cellID4 );
                        if ( cellID1 != Integer.MIN_VALUE && d1 != inputGridNoDataValue ) {
                            if ( cellID1 == cellID2 || cellID1 == cellID3 ) {
                                if ( cellID1 == cellID2 ) {
                                    areaProportion = ( Math.abs( bounds[3] - ( inputGrid.getCellY( cellID1 ) - ( inputGridCellsize / 2.0d ) ) ) * cellsize ) / ( inputGridCellsize * inputGridCellsize );
                                } else {
                                    areaProportion = ( Math.abs( ( inputGrid.getCellX( cellID1 ) + ( inputGridCellsize / 2.0d ) ) - bounds[0] ) * cellsize ) / ( inputGridCellsize * inputGridCellsize );
                                }
                            } else {
                                areaProportion = ( ( Math.abs( bounds[3] - ( inputGrid.getCellY( cellID1 ) - ( inputGridCellsize / 2.0d ) ) ) * Math.abs( ( inputGrid.getCellX( cellID1 ) + ( inputGridCellsize / 2.0d ) ) - bounds[0] ) ) / ( inputGridCellsize * inputGridCellsize ) );
                            }
                            tempGrid1.addToCell( i, d1 * areaProportion );
                            tempGrid2.addToCell( i, areaProportion );
                        }
                        if ( cellID2 != Integer.MIN_VALUE && d2 != inputGridNoDataValue ) {
                            if ( cellID2 != cellID1 ) {
                                if ( cellID2 == cellID4 ) {
                                    areaProportion = ( Math.abs( bounds[2] - ( inputGrid.getCellX( cellID2 ) - ( inputGridCellsize / 2.0d ) ) ) * cellsize ) / ( inputGridCellsize * inputGridCellsize );
                                } else {
                                    areaProportion = ( ( Math.abs( bounds[3] - ( inputGrid.getCellY( cellID2 ) - ( inputGridCellsize / 2.0d ) ) ) * Math.abs( bounds[2] - ( inputGrid.getCellX( cellID2 ) - ( inputGridCellsize / 2.0d ) ) ) ) / ( inputGridCellsize * inputGridCellsize ) );
                                }
                                tempGrid1.addToCell( i, d2 * areaProportion );
                                tempGrid2.addToCell( i, areaProportion );
                            }
                        }
                        if ( cellID3 != Integer.MIN_VALUE && d3 != inputGridNoDataValue ) {
                            if ( cellID3 != cellID1 ) {
                                if ( cellID3 == cellID4 ) {
                                    areaProportion = ( Math.abs( ( inputGrid.getCellY( cellID3 ) + ( inputGridCellsize / 2.0d ) ) - bounds[1] ) * cellsize ) / ( inputGridCellsize * inputGridCellsize );
                                } else {
                                    areaProportion = ( ( Math.abs( ( inputGrid.getCellY( cellID3 ) + ( inputGridCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( ( inputGrid.getCellX( cellID3 ) + ( inputGridCellsize / 2.0d ) ) - bounds[0] ) ) / ( inputGridCellsize * inputGridCellsize ) );
                                }
                                tempGrid1.addToCell( i, d3 * areaProportion );
                                tempGrid2.addToCell( i, areaProportion );
                            }
                        }
                        if ( cellID4 != Integer.MIN_VALUE && d4 != inputGridNoDataValue ) {
                            if ( cellID4 != cellID2 && cellID4 != cellID3 ) {
                                areaProportion = ( ( Math.abs( ( inputGrid.getCellY( cellID4 ) + ( inputGridCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( bounds[2] - ( inputGrid.getCellX( cellID4 ) - ( inputGridCellsize / 2.0d ) ) ) ) / ( inputGridCellsize * inputGridCellsize ) );
                                tempGrid1.addToCell( i, d4 * areaProportion );
                                tempGrid2.addToCell( i, areaProportion );
                            }
                        }
                        // Check fails due to rounding errors!
                        //if ( cellID1 != Integer.MIN_VALUE && cellID2 != Integer.MIN_VALUE && cellID3 != Integer.MIN_VALUE && cellID4 != Integer.MIN_VALUE && totalArea != 1.0 ) { System.out.println( "id = " + i + " : totalArea = " + totalArea + " (cellID1,cellID2,cellID3,cellID4) = (" + cellID1 + "," + cellID2 + "," + cellID3 + "," + cellID4 + ")" ); System.exit(0);}
                    }
                }
                System.out.println( "tempGrid1" );
                System.out.println( tempGrid1.toString() );
                System.out.println( "tempGrid2" );
                System.out.println( tempGrid2.toString() );
                // The values are normalised by dividing the aggregate Grid sum by the proportion of cells with grid values.
                for ( int i = 0; i < nrows * ncols; i ++ ) {
                    d1 = tempGrid2.getCell( i );
                    if ( ! ( d1 == 0.0d || d1 == noDataValue ) ) {
                        //setCell( i, tempGrid2.getCell( i ) );
                        addToCell( i, tempGrid1.getCell( i ) / d1 );
                        //addToCell( i, tempGrid1.getCell( i ) / ( Math.pow( ( inputGridCellsize / cellsize ), 2.0d ) / d1 ) );
                    }
                }
                tempGrid1.clear();
                tempGrid2.clear();
            }
        }
    }
    
    /**
     * Returns the bounding GeoRectangle of the grid.
     */
    public GeoRectangle getBoundingGeoRectangle() {
        double cellsize = getCellsize();
        return new GeoRectangle( getXllcorner(), getYllcorner(), cellsize * getNcols(), cellsize * getNrows() );
    }
    
    /**
     * Returns the bounding GeoRectangle of a cell.
     */
    public GeoRectangle getBoundingGeoRectangle( double x, double y ) {
        double cellsize = getCellsize();
        return new GeoRectangle( x - ( cellsize / 2.0d ), y - ( cellsize / 2.0d ), cellsize, cellsize );
    }
    public GeoRectangle getBoundingGeoRectangle( int row, int column ) {
        return getBoundingGeoRectangle( getCellX( row, column ), getCellY( row, column ) );
    }
    public GeoRectangle getBoundingGeoRectangle( int cellID ) {
        return getBoundingGeoRectangle( getCellX( cellID ), getCellY( cellID ) );
    }
    
    /**
     * Returns cell bounds (xmin, ymin, xmax, ymax)
     */
    public double[] getBounds( int cellID ) {
        double x = getCellX( cellID );
        double y = getCellY( cellID );
        double halfCellsize = getCellsize() / 2.0d;
        double[] bounds = new double[4];
        bounds[0] = x - halfCellsize;
        bounds[1] = y - halfCellsize;
        bounds[2] = x + halfCellsize;
        bounds[3] = y + halfCellsize;
        return bounds;
    }
    
    /**
     * Returns the distance between two cell centroids in map units
     * @ToDo Add distance method in Utilities and call it 
     */
    private final double distance( int cellID1, int cellID2 ) {
        return Math.sqrt( Math.pow( getCellX( cellID1 ) - getCellX( cellID2 ), 2.0d ) + Math.pow( getCellY( cellID1 ) - getCellY( cellID2 ), 2.0d ) );
    }
    
    /**
     * Returns the distance between a pair of coordinates
     * @param x1 - the x coordinte of one point
     * @param y1 - the y coordinte of one point
     * @param x2 - the x coordinte of another point
     * @param y2 - the y coordinte of another point
     */
    public final double distance( double x1, double y1, double x2, double y2 ) {
        return Math.sqrt( Math.pow( x1 - x2, 2.0d ) + Math.pow( y1 - y2, 2.0d ) );
    }
    
    /**
     * Returns the distance between two cell centroids in cell units
     */
    private final double cellDistance( int cellID1, int cellID2 ) {
        return ( distance( cellID1, cellID2 ) / getCellsize() );
    }
    
    /**
     * Returns the angle in radians to the y axis between a pair of coordinates
     * @param x1 - the x coordinte of one point
     * @param y1 - the y coordinte of one point
     * @param x2 - the x coordinte of another point
     * @param y2 - the y coordinte of another point
     * @ToDo Add angle method to Utilities and call it 
     */
    protected final double angle( int cellID1, int cellID2 ) {
        return angle( getCellX( cellID1 ), getCellY( cellID1 ), getCellX( cellID2 ), getCellY( cellID2 ) );
    }
    
    /**
     * Returns the angle in radians to the y axis between a pair of coordinates
     * @param x1 - the x coordinte of one point
     * @param y1 - the y coordinte of one point
     * @param x2 - the x coordinte of another point
     * @param y2 - the y coordinte of another point
     */
    public final double angle( double x1, double y1, double x2, double y2 ) {
        double xdiff = x1 - x2;
        double ydiff = y1 - y2;
        double angle;
        if ( xdiff == 0.0d && ydiff == 0.0d ) {
            angle = -1.0d;
        } else {
            if ( xdiff <= 0.0d ) {
                if ( xdiff == 0.0d ) {
                    if ( ydiff <= 0.0d ) {
                        angle = 0.0d;
                    } else {
                        angle = Math.PI;
                    }
                } else {
                    if ( ydiff <= 0.0d ) {
                        if ( ydiff == 0.0d ) {
                            angle = Math.PI / 2.0d;
                        } else {
                            angle = Math.atan( Math.abs( xdiff / ydiff ) );
                        }
                    } else {
                        angle = Math.PI - Math.atan( Math.abs( xdiff / ydiff ) );
                    }
                }
            } else {
                if ( ydiff <= 0.0d ) {
                    if ( ydiff == 0.0d ) {
                        angle = 3.0d * Math.PI / 2.0d;
                    } else {
                        angle = ( 2.0d * Math.PI ) - Math.atan( Math.abs( xdiff / ydiff ) );
                    }
                } else {
                    angle = Math.PI + Math.atan( Math.abs( xdiff / ydiff ) );
                }
            }
        }
        return angle;
    }
    
    
    /**
     * ==========================================================
     * Operations that modify this AbstractGrid2DSquareCellDouble
     * ==========================================================
     */
    
    /**
     * mask sets values in the grid to its noDataValue if values in the mask
     * have the mask noDataValue.
     * The mask has to have the same spatial frame as this.
     */
    public final void mask( AbstractGrid2DSquareCellDouble mask ) {
        double d1;
        double maskNoDataValue = mask.getNoDataValue();
        GeoRectangle maskBounds = mask.getBoundingGeoRectangle();
        double maskCellsize = mask.getCellsize();
        double noDataValue = getNoDataValue();
        GeoRectangle bounds = getBoundingGeoRectangle();
        double cellsize = getCellsize();
        int nrows = getNrows();
        int ncols = getNcols();
        if ( maskBounds.equals( bounds ) && maskCellsize == cellsize ) {
            for ( int i = 0; i < nrows * ncols; i ++ ) {
                d1 = mask.getCell( i );
                if ( d1 == maskNoDataValue ) {
                    setCell( i, noDataValue );
                }
            }
        } else {
            System.out.println( "!!!Warning: Spatial frame of mask is different in mask ( " + mask.toString() + " )" );
        }
    }
    
    
    /**
     * ====================================================================================================
     * Operations on this AbstractGrid2DSquareCellDouble that return another AbstractGrid2DSquareCellDouble
     * ====================================================================================================
     */
    
    /**
     * Normalise method
     * Returns a normalised grid
     */
    public final AbstractGrid2DSquareCellDouble normalise( double min, double max ) {
        try {
            return normalise( new Grid2DSquareCellDoubleFactory(), min, max );
        } catch ( OutOfMemoryError e ) {
            return normalise( new Grid2DSquareCellDoubleFileFactory(), min, max );
        }
    }
    public final AbstractGrid2DSquareCellDouble normalise( AbstractGrid2DSquareCellDoubleFactory gridFactory, double min, double max ) {
        int nrows = getNrows();
        int ncols = getNcols();
        double xllcorner = getXllcorner();
        double yllcorner = getYllcorner();
        double cellsize = getCellsize();
        double noDataValue = getNoDataValue();
        AbstractGrid2DSquareCellDouble grid = gridFactory.createGrid2DSquareCellDouble( nrows, ncols, xllcorner, yllcorner, cellsize, noDataValue );
        AbstractGridStatistics stats = getGridStatistics();
        double thisMin = stats.getMin();
        double thisMax = stats.getMax();
        double thisValue;
        // if range of either input or output is zero
        if ( ( thisMax - thisMin ) == 0.0d || ( max - min ) == 0.0d ) {
            for ( int i = 0; i < nrows * ncols; i ++ ) {
                thisValue = getCell( i );
                if ( thisValue != noDataValue ) {
                    grid.setCell( i, max );
                }
            }
        } else {
            for ( int i = 0; i < nrows * ncols; i ++ ) {
                thisValue = getCell( i );
                if ( thisValue != noDataValue ) {
                    grid.setCell( i, ( ( ( ( thisValue - thisMin ) / ( thisMax - thisMin ) ) ) * ( max - min ) ) + min );
                }
            }
        }
        return grid;
    }
    
    
    /*
     * Aggregation methods
     * ===================
     */
    
    /**
     * aggregate - returns a grid at a lower level of resolution where the values in the cells are either
     * the sum, mean, max or min
     * @ToDos
     * 1. Add normalisationLevel parameter
     * 2. Add visualisationLevel parameter
     * @param cellfactor - the number of times wider/higher the aggregated grid cells are
     * @param statistic - "sum", "mean", "max", or "min" depending on what aggregate of values are wanted
     */
    public AbstractGrid2DSquareCellDouble aggregate( int cellFactor, String statistic ) {
        try {
            return aggregate( cellFactor, statistic, new Grid2DSquareCellDoubleFactory() );
        } catch ( java.lang.OutOfMemoryError e ) {
            return aggregate( cellFactor, statistic, new Grid2DSquareCellDoubleFileFactory() );
        }
    }
    /**
     * Use this aggregate method rather aggregate( int cellFactor, String statistic ) if it is known what
     * type of Abstract2DSquareCellDoubleFactory is wanted.
     * @param gridFactory - the Abstract2DSquareCellDoubleFactory used to create grids
     * @see aggregate( int cellFactor, String statistic )
     */
    public AbstractGrid2DSquareCellDouble aggregate( int cellFactor, String statistic, AbstractGrid2DSquareCellDoubleFactory gridFactory ) {
        return aggregate( cellFactor, statistic, gridFactory, getXllcorner(), getYllcorner(), getNoDataValue() );
    }
    /**
     * Use this aggregate method to force origin of the result to be ( resultXllcorner, resultYllcorner )
     * @param resultXllcorner - the x-coordinate of the aggregate grid lower left corner
     * @param resultYllcorner - the y-coordinate of the aggregate grid lower left corner
     * @see aggregate( int cellFactor, String statistic )
     */
    public AbstractGrid2DSquareCellDouble aggregate( int cellFactor, String statistic, double resultXllcorner, double resultYllcorner ) {
        return aggregate( cellFactor, statistic, resultXllcorner, resultYllcorner, getNoDataValue() );
    }
    public AbstractGrid2DSquareCellDouble aggregate( int cellFactor, String statistic, double resultXllcorner, double resultYllcorner, double noDataValue ) {
        try {
            return aggregate( cellFactor, statistic, new Grid2DSquareCellDoubleFactory(), resultXllcorner, resultYllcorner, noDataValue );
        } catch ( java.lang.OutOfMemoryError e ) {
            return aggregate( cellFactor, statistic, new Grid2DSquareCellDoubleFileFactory(), resultXllcorner, resultYllcorner, noDataValue );
        }
    }
    public AbstractGrid2DSquareCellDouble aggregate( int cellFactor, String statistic, AbstractGrid2DSquareCellDoubleFactory gridFactory, double resultXllcorner, double resultYllcorner, double noDataValue ) {
        //System.out.println( "aggregate( cellFactor(" + cellFactor + "), statistic(" + statistic + "), gridFactory(" + gridFactory + "), resultXllcorner(" + resultXllcorner + "), resultYllcorner(" + resultYllcorner + "), noDataValue(" + noDataValue + ") )" );
        // Initial tests
        if ( cellFactor <= 0 ) {
            System.out.println( "Warning!!! cellFactor <= 0 : Returning null." );
            return null;
        }
        
        // Initialisation
        int nrows = getNrows();
        int ncols = getNcols();
        double xllcorner = getXllcorner();
        double yllcorner = getYllcorner();
        double cellsize = getCellsize();
        double resultCellsize = cellsize * ( double ) cellFactor;
        double width = cellsize * ncols;
        double height = cellsize * nrows;
        double cells = Math.pow( cellFactor, 2.0d );
        // Calculate resultNrows and resultHeight
        int resultNrows = 1;
        double resultHeight = resultCellsize;
        while ( ( yllcorner + resultHeight ) < ( yllcorner + height ) ) {
            resultNrows ++;
            resultHeight += resultCellsize;
        }
        // Calculate resultNcols and resultWidth
        int resultNcols = 1;
        double resultWidth = resultCellsize;
        while ( ( xllcorner + resultWidth ) < ( xllcorner + width ) ) {
            resultNcols ++;
            resultWidth += resultCellsize;
        }
        // Initialise result
        AbstractGrid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( resultNrows, resultNcols, resultXllcorner, resultYllcorner, resultCellsize, noDataValue );
        
        // sum
        // If there is a noDataValue the average of other values is substituted in the sum.
        if ( statistic.equalsIgnoreCase( "sum" ) ) {
            double d1 = 0.0d;
            double d2 = 0.0d;
            double d3 = 0.0d;
            for ( int i = 0; i < resultNrows; i ++ ) {
                for ( int j = 0; j < resultNcols; j ++ ) {
                    d2 = 0.0d;
                    d3 = 0.0d;
                    for ( int p = ( 0 - ( int ) ( ( yllcorner - yllcorner ) / cellsize ) ); p < ( cellFactor - ( int ) ( ( yllcorner - yllcorner ) / cellsize ) ); p ++ ) {
                        for ( int q = ( 0 - ( int ) ( ( xllcorner - xllcorner ) / cellsize ) ); q < ( cellFactor - ( int ) ( ( xllcorner - xllcorner ) / cellsize ) ); q ++ ) {
                            d1 = getCell( ( i * cellFactor ) + p, ( j * cellFactor ) + q );
                            if ( d1 != noDataValue ) {
                                d2 += d1;
                                d3 ++;
                            }
                        }
                    }
                    if ( d3 != 0.0d ) {
                        result.setCell( i, j, d2 + ( ( cells - d3 ) * d2 / cells ) );
                    }
                }
            }
        }

        // mean
        if ( statistic.equalsIgnoreCase( "mean" ) ) {
            double d1 = 0.0d;
            double d2 = 0.0d;
            double d3 = 0.0d;
            for ( int i = 0; i < resultNrows; i ++ ) {
                for ( int j = 0; j < resultNcols; j ++ ) {
                    d2 = 0.0d;
                    d3 = 0.0d;
                    for ( int p = ( 0 - ( int ) ( ( yllcorner - yllcorner ) / cellsize ) ); p < ( cellFactor - ( int ) ( ( yllcorner - yllcorner ) / cellsize ) ); p ++ ) {
                        for ( int q = ( 0 - ( int ) ( ( xllcorner - xllcorner ) / cellsize ) ); q < ( cellFactor - ( int ) ( ( xllcorner - xllcorner ) / cellsize ) ); q ++ ) {
                            d1 = getCell( ( i * cellFactor ) + p, ( j * cellFactor ) + q );
                            if ( d1 != noDataValue ) {
                                d2 += d1;
                                d3 ++;
                            }
                        }
                    }
                    if (d3 != 0.0d) {
                        result.setCell( i, j, ( d2 / d3 ) );
                    }
                }
            }
        }
        
        // min
        if ( statistic.equalsIgnoreCase( "min" ) ) {
            double min = Double.MAX_VALUE;
            double d1 = 0.0d;
            for ( int i = 0; i < resultNrows; i ++ ) {
                for ( int j = 0; j < resultNcols; j ++ ) {
                    min = Double.MAX_VALUE;
                    for ( int p = ( 0 - ( int ) ( ( yllcorner - yllcorner ) / cellsize ) ); p < ( cellFactor - ( int ) ( ( yllcorner - yllcorner ) / cellsize ) ); p ++ ) {
                        for ( int q = ( 0 - ( int ) ( ( xllcorner - xllcorner ) / cellsize ) ); q < ( cellFactor - ( int ) ( ( xllcorner - xllcorner ) / cellsize ) ); q ++ ) {
                            d1 = getCell( ( i * cellFactor ) + p, ( j * cellFactor ) + q );
                            if ( d1 != noDataValue ) {
                                min = Math.min( min, d1 );
                            }
                        }
                    }
                    if ( min != Double.MAX_VALUE ) {
                        result.setCell( i, j, min );
                    }
                }
            }
        }
        
        // max
        if ( statistic.equalsIgnoreCase( "max" ) ) {
            double max = Double.MIN_VALUE;
            double d1 = 0.0d;
            for ( int i = 0; i < resultNrows; i ++ ) {
                for ( int j = 0; j < resultNcols; j ++ ) {
                    max = Double.MIN_VALUE;
                    for ( int p = ( 0 - ( int ) ( ( yllcorner - yllcorner ) / cellsize ) ); p < ( cellFactor - ( int ) ( ( yllcorner - yllcorner ) / cellsize ) ); p ++ ) {
                        for ( int q = ( 0 - ( int ) ( ( xllcorner - xllcorner ) / cellsize ) ); q < ( cellFactor - ( int ) ( ( xllcorner - xllcorner ) / cellsize ) ); q ++ ) {
                            d1 = getCell( ( i * cellFactor ) + p, ( j * cellFactor ) + q );
                            if ( d1 != noDataValue ) {
                                max = Math.max( max, d1 );
                            }
                        }
                    }
                    if ( max != Double.MAX_VALUE ) {
                        result.setCell( i, j, max );
                    }
                }
            }
        }
        //System.out.println( result.toString() );
        return result;
    }
    /**
     * Aggregate grid
     * @param statistic - the type of aggregation
     * @param resultCellsize - output grid cellsize
     * @param resultXllcorner - output grid xllcorner
     * @param resultYllcorner - output grid yllcorner
     * @param noDataValue - output noDataValue
     */
    public AbstractGrid2DSquareCellDouble aggregate( double resultCellsize, String statistic, double resultXllcorner, double resultYllcorner, double noDataValue ) {
        try {
            return aggregate( resultCellsize, statistic, new Grid2DSquareCellDoubleFactory(), resultXllcorner, resultYllcorner, noDataValue );
        } catch ( java.lang.OutOfMemoryError e ) {
            return aggregate( resultCellsize, statistic, new Grid2DSquareCellDoubleFileFactory(), resultXllcorner, resultYllcorner, noDataValue );
        }
    }
    /**
     * @param gridFactory - the AbstracGrid2DSquareCellDoubleFactory for generating AbstractGrid2DSquareCellDouble grids
     * @ToDo Add a method changeSpatialFrame which does either aggregation or areal interpolation
     */
    public AbstractGrid2DSquareCellDouble aggregate( double resultCellsize, String statistic, AbstractGrid2DSquareCellDoubleFactory gridFactory, double resultXllcorner, double resultYllcorner, double noDataValue ) {
        //System.out.println( "aggregate( cellsize( " + resultCellsize + " ), statistic( " + statistic + " ), gridFactory( " + gridFactory.toString() + " ), resultXllcorner( " + resultXllcorner + " ), resultYllcorner( " + resultYllcorner + " ), noDataValue( " + noDataValue + " ) )" );
        // Initialistaion
        int nrows = getNrows();
        int ncols = getNcols();
        double xllcorner = getXllcorner();
        double yllcorner = getYllcorner();
        double cellsize = getCellsize();
        double width = cellsize * ncols;
        double height = cellsize * nrows;
        // Test this is an aggregation
        if ( resultCellsize <= cellsize ) {
            System.out.println( "!!!Warning: resultCellsize <= cellsize in aggregate( cellsize( " + resultCellsize + " ), statistic( " + statistic + " ), gridFactory( " + gridFactory.toString() + " ), resultXllcorner( " + resultXllcorner + " ), resultYllcorner( " + resultYllcorner + " ), noDataValue( " + noDataValue + " ) ). Returning null!" );
            return null;
        }
        // Test for intersection
        if ( ( resultXllcorner > xllcorner + ( ( double ) ncols * cellsize ) ) || ( resultYllcorner > yllcorner + ( ( double ) nrows * cellsize ) ) ) {
            System.out.println( "!!!Warning: No intersection in aggregate( cellsize( " + resultCellsize + " ), statistic( " + statistic + " ), gridFactory( " + gridFactory.toString() + " ), resultXllcorner( " + resultXllcorner + " ), resultYllcorner( " + resultYllcorner + " ), noDataValue( " + noDataValue + " ) ). Returning null!" );
            return null;
        }
        // If resultCellsize is an integer multiple of cellsize and this grid aligns with the result then use
        // a cellFactor aggregation as it should be faster.
        //System.out.println("resultCellsize % cellsize == " + ( resultCellsize % cellsize ) );
        //System.out.println("resultXllcorner % cellsize = " + ( resultXllcorner % cellsize ) + ", xllcorner % cellsize = " + ( xllcorner % cellsize ) );
        //System.out.println("resultYllcorner % cellsize = " + ( resultYllcorner % cellsize ) + ", yllcorner % cellsize = " + ( yllcorner % cellsize ) );
        if ( ( resultCellsize % cellsize == 0.0d ) && ( ( resultXllcorner % cellsize ) == ( xllcorner % cellsize ) ) && ( ( resultYllcorner % cellsize ) == ( yllcorner % cellsize ) ) ) {
            int cellFactor = ( int ) ( resultCellsize / cellsize );
            return aggregate( cellFactor, statistic, resultXllcorner, resultYllcorner, noDataValue );
        }
        // Calculate resultNrows and resultHeight
        int resultNrows = 1;
        double resultHeight = resultCellsize;
        while ( ( resultYllcorner + resultHeight ) < ( yllcorner + height ) ) {
            resultNrows ++;
            resultHeight += resultCellsize;
        }
        // Calculate resultNcols and resultWidth
        int resultNcols = 1;
        double resultWidth = resultCellsize;
        while ( ( resultXllcorner + resultWidth ) < ( xllcorner + width ) ) {
            resultNcols ++;
            resultWidth += resultCellsize;
        }
        //System.out.println( "resultNcols " + resultNcols + ", resultNrows " + resultNrows );
        //System.out.println( "inputGridNcols " + ncols + ", inputGridNrows " + nrows );
        // Initialise result
        AbstractGrid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( resultNrows, resultNcols, resultXllcorner, resultYllcorner, resultCellsize, noDataValue, 0 );
        AbstractGrid2DSquareCellDouble tempGrid1 = gridFactory.createGrid2DSquareCellDouble( resultNrows, resultNcols, resultXllcorner, resultYllcorner, resultCellsize, noDataValue, 0 );
        AbstractGrid2DSquareCellDouble tempGrid2 = gridFactory.createGrid2DSquareCellDouble( resultNrows, resultNcols, resultXllcorner, resultYllcorner, resultCellsize, noDataValue, 0 );

        // sum
        if ( statistic.equalsIgnoreCase( "sum" ) ) {
            double x = 0.0d;
            double y = 0.0d;
            double d1 = noDataValue;
            double areaProportion = 0.0d;
            double[] bounds = new double[4];
            int cellID1 = -1;
            int cellID2 = -1;
            int cellID3 = -1;
            int cellID4 = -1;
            //double totalArea = 0;
            for ( int i = 0; i < nrows * ncols; i ++ ) {
                d1 = getCell( i );
                if ( d1 != noDataValue ) {
                    x = getCellX( i );
                    y = getCellY( i );
                    bounds = getBounds( i );
                    cellID1 = result.getCellID( bounds[0], bounds[3] );
                    cellID2 = result.getCellID( bounds[2], bounds[3] );
                    cellID3 = result.getCellID( bounds[0], bounds[1] );
                    cellID4 = result.getCellID( bounds[2], bounds[1] );
                    if ( cellID1 == cellID2 && cellID2 == cellID3 ) {
                        if ( cellID1 != Integer.MIN_VALUE ) {
                            areaProportion = 1.0d;
                            tempGrid1.addToCell( x, y, d1 * areaProportion );
                            tempGrid2.addToCell( x, y, areaProportion );
                        }
                    } else {
                        if ( cellID1 != Integer.MIN_VALUE ) {
                            if ( cellID1 == cellID2 || cellID1 == cellID3 ) {
                                if ( cellID1 == cellID2 ) {
                                    areaProportion = ( Math.abs( bounds[3] - ( result.getCellY( cellID1 ) - ( resultCellsize / 2.0d ) ) ) * cellsize ) / ( cellsize * cellsize );
                                } else {
                                    areaProportion = ( Math.abs( ( result.getCellX( cellID1 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) * cellsize ) / ( cellsize * cellsize );
                                }
                            } else {
                                areaProportion = ( ( Math.abs( bounds[3] - ( result.getCellY( cellID1 ) - ( resultCellsize / 2.0d ) ) ) * Math.abs( ( result.getCellX( cellID1 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) ) / ( cellsize * cellsize ) );
                            }
                            tempGrid1.addToCell( cellID1, d1 * areaProportion );
                            tempGrid2.addToCell( cellID1, areaProportion );
                        }
                        if ( cellID2 != Integer.MIN_VALUE ) {
                            if ( cellID2 != cellID1 ) {
                                if ( cellID2 == cellID4 ) {
                                    areaProportion = ( Math.abs( bounds[2] - ( result.getCellX( cellID2 ) - ( resultCellsize / 2.0d ) ) ) * cellsize ) / ( cellsize * cellsize );
                                } else {
                                    areaProportion = ( ( Math.abs( bounds[3] - ( result.getCellY( cellID2 ) - ( resultCellsize / 2.0d ) ) ) * Math.abs( bounds[2] - ( result.getCellX( cellID2 ) - ( resultCellsize / 2.0d ) ) ) ) / ( cellsize * cellsize ) );
                                }
                                tempGrid1.addToCell( cellID2, d1 * areaProportion );
                                tempGrid2.addToCell( cellID2, areaProportion );
                            }
                        }
                        if ( cellID3 != Integer.MIN_VALUE ) {
                            if ( cellID3 != cellID1 ) {
                                if ( cellID3 == cellID4 ) {
                                    areaProportion = ( Math.abs( ( result.getCellY( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * cellsize ) / ( cellsize * cellsize );
                                } else {
                                    areaProportion = ( ( Math.abs( ( result.getCellY( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( ( result.getCellX( cellID3 ) + ( resultCellsize / 2.0d) ) - bounds[0] ) ) / ( cellsize * cellsize ) );
                                }
                                tempGrid1.addToCell( cellID3, d1 * areaProportion );
                                tempGrid2.addToCell( cellID3, areaProportion );
                            }
                        }
                        if ( cellID4 != Integer.MIN_VALUE ) {
                            if ( cellID4 != cellID2 && cellID4 != cellID3 ) {
                                areaProportion = ( ( Math.abs( ( result.getCellY( cellID4 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( bounds[2] - ( result.getCellX( cellID4 ) - ( resultCellsize / 2.0d ) ) ) ) / ( cellsize * cellsize ) );
                                tempGrid1.addToCell( cellID4, d1 * areaProportion );
                                tempGrid2.addToCell( cellID4, areaProportion );
                            }
                        }
                        // Check fails due to rounding errors!
                        //if ( cellID1 != Integer.MIN_VALUE && cellID2 != Integer.MIN_VALUE && cellID3 != Integer.MIN_VALUE && cellID4 != Integer.MIN_VALUE && totalArea != 1.0 ) { System.out.println( "id = " + i + " : totalArea = " + totalArea + " (cellID1,cellID2,cellID3,cellID4) = (" + cellID1 + "," + cellID2 + "," + cellID3 + "," + cellID4 + ")" ); System.exit(0);}
                    }
                }
            }
            // The values are normalised by dividing the aggregate Grid sum by the proportion of cells with grid values.
            for ( int i = 0; i < resultNrows * resultNcols; i ++ ) {
                d1 = tempGrid2.getCell( i );
                if ( d1 != 0.0d && d1 != noDataValue ) {
                    result.setCell( i, tempGrid1.getCell( i ) / ( Math.pow( ( resultCellsize / cellsize ), 2.0d ) / d1 ) );
                }
            }
            tempGrid1.clear();
            tempGrid2.clear();
        }
        
        // mean
        if ( statistic.equalsIgnoreCase( "mean" ) ) {
            // To calculate the mean and cope with NODATA it is necessary to pass
            // through the data twice or for each aggregated cell get all
            // intersecting cells. This is because each cells area as a proportion
            // of the non noDataValue area of an aggregated cell is needed. This
            // cannot be simply done as it all depends on NODATA. In the
            // implementation below the data is read through twice. First read
            // involves calculating NODATA in each aggregated cell.
            double x;
            double y;
            double d1;
            double area;
            double[] bounds = new double[4];
            int cellID1;
            int cellID2;
            int cellID3;
            int cellID4;
            //double totalArea = 0;
            for ( int i = 0; i < nrows * ncols; i ++ ) {
                d1 = getCell( i );
                if ( d1 != noDataValue ) {
                    x = getCellX( i );
                    y = getCellY( i );
                    bounds = getBounds( i );
                    cellID1 = result.getCellID( bounds[0], bounds[3] );
                    cellID2 = result.getCellID( bounds[2], bounds[3] );
                    cellID3 = result.getCellID( bounds[0], bounds[1] );
                    cellID4 = result.getCellID( bounds[2], bounds[1] );
                    if ( cellID1 == cellID2 && cellID2 == cellID3 ) {
                        if ( cellID1 != Integer.MIN_VALUE ) {
                            area = cellsize * cellsize;
                            tempGrid1.addToCell( x, y, area );
                        }
                    } else {
                        if ( cellID1 != Integer.MIN_VALUE ) {
                            if ( cellID1 == cellID2 || cellID1 == cellID3 ) {
                                if ( cellID1 == cellID2 ) {
                                    area = ( Math.abs( bounds[3] - ( result.getCellY( cellID1 ) - ( resultCellsize / 2.0d ) ) ) * cellsize );
                                } else {
                                    area = ( Math.abs( ( result.getCellX( cellID1 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) * cellsize );
                                }
                            } else {
                                area = ( Math.abs( bounds[3] - ( result.getCellY( cellID1 ) - ( resultCellsize / 2.0d ) ) ) * Math.abs( ( result.getCellX( cellID1 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) );
                            }
                            tempGrid1.addToCell( cellID1, area );
                        }
                        if ( cellID2 != Integer.MIN_VALUE ) {
                            if ( cellID2 != cellID1 ) {
                                if ( cellID2 == cellID4 ) {
                                    area = ( Math.abs( bounds[2] - ( result.getCellX( cellID2 ) - ( resultCellsize / 2.0d ) ) ) * cellsize );
                                } else {
                                    area = ( Math.abs( bounds[3] - ( result.getCellY( cellID2 ) - ( resultCellsize / 2.0d ) ) ) * Math.abs( bounds[2] - ( result.getCellX( cellID2 ) - ( resultCellsize / 2.0d ) ) ) );
                                }
                                tempGrid1.addToCell( cellID2, area );
                            }
                        }
                        if ( cellID3 != Integer.MIN_VALUE ) {
                            if ( cellID3 != cellID1 ) {
                                if ( cellID3 == cellID4 ) {
                                    area = ( Math.abs( ( result.getCellY( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * cellsize );
                                } else {
                                    area = ( Math.abs( ( result.getCellY( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( ( result.getCellX( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) );
                                }
                                tempGrid1.addToCell( cellID3, area );
                            }
                        }
                        if ( cellID4 != Integer.MIN_VALUE ) {
                            if ( cellID4 != cellID2 && cellID4 != cellID3 ) {
                                area = ( Math.abs( ( result.getCellY( cellID4 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( bounds[2] - ( result.getCellX( cellID4 ) - ( resultCellsize / 2.0d ) ) ) );
                                tempGrid1.addToCell( cellID4, area );
                            }
                        }
                    }
                }
            }
            for ( int i = 0; i < nrows * ncols; i ++ ) {
                double areaIntersect;
                d1 = getCell( i );
                if ( d1 != noDataValue ) {
                    x = getCellX( i );
                    y = getCellY( i );
                    bounds = getBounds( i );
                    cellID1 = result.getCellID( bounds[0], bounds[3] );
                    cellID2 = result.getCellID( bounds[2], bounds[3] );
                    cellID3 = result.getCellID( bounds[0], bounds[1] );
                    cellID4 = result.getCellID( bounds[2], bounds[1] );
                    if ( cellID1 == cellID2 && cellID2 == cellID3 ) {
                        if ( cellID1 != Integer.MIN_VALUE ) {
                            area = tempGrid1.getCell( x, y );
                            result.addToCell( x, y, d1 * ( cellsize * cellsize ) / area );
                        }
                    } else {
                        if ( cellID1 != Integer.MIN_VALUE ) {
                            if ( cellID1 == cellID2 || cellID1 == cellID3 ) {
                                if ( cellID1 == cellID2 ) {
                                    areaIntersect = ( Math.abs( bounds[3] - ( result.getCellY( cellID1 ) - ( resultCellsize / 2.0d ) ) ) * cellsize );
                                } else {
                                    areaIntersect = ( Math.abs( ( result.getCellX( cellID1 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) * cellsize );
                                }
                            } else {
                                areaIntersect = ( Math.abs( bounds[3] - ( result.getCellY( cellID1 ) - ( resultCellsize / 2.0d ) ) ) * Math.abs( ( result.getCellX( cellID1 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) );
                            }
                            area = tempGrid1.getCell( cellID1 );
                            result.addToCell( cellID1, d1 * areaIntersect / area );
                        }
                        if ( cellID2 != Integer.MIN_VALUE ) {
                            if ( cellID2 != cellID1 ) {
                                if ( cellID2 == cellID4 ) {
                                    areaIntersect = ( Math.abs( bounds[2] - ( result.getCellX( cellID2 ) - ( resultCellsize / 2.0d ) ) ) * cellsize );
                                } else {
                                    areaIntersect = ( Math.abs( bounds[3] - ( result.getCellY( cellID2 ) - ( resultCellsize / 2.0d ) ) ) * Math.abs( bounds[2] - ( result.getCellX( cellID2 ) - ( resultCellsize / 2.0d ) ) ) );
                                }
                                area = tempGrid1.getCell( cellID2 );
                                result.addToCell( cellID2, d1 * areaIntersect / area );
                            }
                        }
                        if ( cellID3 != Integer.MIN_VALUE ) {
                            if ( cellID3 != cellID1 ) {
                                if ( cellID3 == cellID4 ) {
                                    areaIntersect = ( Math.abs( ( result.getCellY( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * cellsize );
                                } else {
                                    areaIntersect = ( Math.abs( ( result.getCellY( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( ( result.getCellX( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) );
                                }
                                area = tempGrid1.getCell( cellID3 );
                                result.addToCell( cellID3, d1 * areaIntersect / area );
                            }
                        }
                        if ( cellID4 != Integer.MIN_VALUE ) {
                            if ( cellID4 != cellID2 && cellID4 != cellID3 ) {
                                areaIntersect = ( Math.abs( ( result.getCellY( cellID4 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( bounds[2] - ( result.getCellX( cellID4 ) - ( resultCellsize / 2.0d ) ) ) );
                                area = tempGrid1.getCell( cellID4 );
                                result.addToCell( cellID4, d1 * areaIntersect / area );
                            }
                        }
                    }
                }
            }
            tempGrid1.clear();
        }
        
        // max
        if ( statistic.equalsIgnoreCase( "max" ) ) {
            double x;
            double y;
            double d1;
            double d2;
            double[] bounds = new double[4];
            for ( int i = 0; i < nrows * ncols; i ++ ) {
                d1 = getCell( i );
                if ( d1 != noDataValue ) {
                    x = getCellX( i );
                    y = getCellY( i );
                    bounds = getBounds( i );
                    d2 = result.getCell( bounds[0], bounds[3] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[0], bounds[3], Math.max( d2, d1 ) );
                    } else {
                        result.setCell( bounds[0], bounds[3], d1 );
                    }
                    d2 = result.getCell( bounds[2], bounds[3] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[2], bounds[3], Math.max( d2, d1 ) );
                    } else {
                        result.setCell( bounds[2], bounds[3], d1 );
                    }
                    d2 = result.getCell( bounds[0], bounds[1] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[0], bounds[1], Math.max( d2, d1 ) );
                    } else {
                        result.setCell( bounds[0], bounds[1], d1 );
                    }
                    d2 = result.getCell( bounds[2], bounds[1] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[2], bounds[1], Math.max( d2, d1 ) );
                    } else {
                        result.setCell( bounds[2], bounds[1], d1 );
                    }
                }
            }
        }
        
        // min
        if ( statistic.equalsIgnoreCase( "min" ) ) {
            double x;
            double y;
            double d1;
            double d2;
            double[] bounds = new double[4];
            for ( int i = 0; i < nrows * ncols; i ++ ) {
                d1 = getCell( i );
                if ( d1 != noDataValue ) {
                    x = getCellX( i );
                    y = getCellY( i );
                    bounds = getBounds( i );
                    d2 = result.getCell( bounds[0], bounds[3] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[0], bounds[3], Math.min( d2, d1 ) );
                    } else {
                        result.setCell( bounds[0], bounds[3], d1 );
                    }
                    d2 = result.getCell( bounds[2], bounds[3] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[2], bounds[3], Math.min( d2, d1 ) );
                    } else {
                        result.setCell( bounds[2], bounds[3], d1 );
                    }
                    d2 = result.getCell( bounds[0], bounds[1] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[0], bounds[1], Math.min( d2, d1 ) );
                    } else {
                        result.setCell( bounds[0], bounds[1], d1 );
                    }
                    d2 = result.getCell( bounds[2], bounds[1] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[2], bounds[1], Math.min( d2, d1 ) );
                    } else {
                        result.setCell( bounds[2], bounds[1], d1 );
                    }
                }
            }
        }
        //System.out.println( result.toString() );
        return result;
    }
    
    
    /*
     * Non-aggregation methods
     * =======================
     */
    
    /**
     * regionStats methods for calculating statistics of a grid in an approximately circular region 
     * defined either by an integer cell distance or double distance.
     * The following stats are supported; sum, mean, min, max, stdev.
     * noDataValues are treated as follows:
     * sum - 0.0d
     * mean - mean
     * max - Double.MIN_VALUE
     * min - Double.MAX_VALUE
     * stdev - In mean as above for n they are ignored
     */
    public final AbstractGrid2DSquareCellDouble regionStats( String statistic, int cellDistance ) {
        return regionStats( statistic, cellDistance * getCellsize() );
    }
    public final AbstractGrid2DSquareCellDouble regionStats( String statistic, double distance ) {
        try {
            return regionStats( new Grid2DSquareCellDoubleFactory(), statistic, distance );
        } catch ( java.lang.OutOfMemoryError e ) {
            return regionStats( new Grid2DSquareCellDoubleFileFactory(), statistic, distance );
        }
    }
    public final AbstractGrid2DSquareCellDouble regionStats( AbstractGrid2DSquareCellDoubleFactory gridFactory, String statistic, double distance ) {
        System.out.println( "regionStats( AbstractGrid2DSquareCellDoubleFactory(" + gridFactory.toString() + "), statistic(" + statistic + "), distance(" + distance + ") )" );
        int ncols = getNcols();
        int nrows = getNrows();
        double xllcorner = getXllcorner();
        double yllcorner = getYllcorner();
        double cellsize = getCellsize();
        double noDataValue = getNoDataValue();
        // Initialise result
        AbstractGrid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( nrows, ncols, xllcorner, yllcorner, cellsize, noDataValue );
        int cellDistance = ( int ) Math.ceil( distance / cellsize );
        int cellID1 = -1;
        int cellID2 = -1;
        double thisCellDistance = 0.0d;
        double d1 = 0.0d;

        // sum
        // If all values are noDataValues then noDatavalue is returned.
        if ( statistic.equalsIgnoreCase( "sum" ) ) {
            double sum = 0.0d;
            int count1 = 0;
            int count2 = 0;
            for ( int i = 0; i < nrows; i ++ ) {
                for ( int j = 0; j < ncols; j ++ ) {
                    sum = 0.0d;
                    count1 = 0;
                    count2 = 0;
                    cellID1 = getCellID( i, j );
                    for ( int p = -cellDistance; p <= cellDistance; p ++ ) {
                        for ( int q = -cellDistance; q <= cellDistance; q ++ ) {
                            count1 ++;
                            d1 = getCell( ( i + p ), ( j + q ) );
                            if ( d1 != noDataValue ) {
                                cellID2 = getCellID( ( i + p ), ( j + q ) );
                                thisCellDistance = cellDistance( cellID1, cellID2 );
                                if ( thisCellDistance <= cellDistance ) {
                                    sum += d1;
                                }
                            } else {
                                count2 ++;
                            }
                        }
                    }
                    if ( count1 != count2 ) {
                        result.setCell( i, j, sum );
                    }
                }
            }
        }
        
        // max
        if ( statistic.equalsIgnoreCase( "max" ) ) {
            double max = Double.MIN_VALUE;
            for ( int i = 0; i < nrows; i ++ ) {
                for ( int j = 0; j < ncols; j ++ ) {
                    max = Double.MIN_VALUE;
                    cellID1 = getCellID( i, j );
                    for ( int p = -cellDistance; p <= cellDistance; p ++ ) {
                        for ( int q = -cellDistance; q <= cellDistance; q ++ ) {
                            d1 = getCell( ( i + p ), ( j + q ) );
                            if ( d1 != noDataValue ) {
                                cellID2 = getCellID( ( i + p ), ( j + q ) );
                                thisCellDistance = cellDistance( cellID1, cellID2 );
                                if ( thisCellDistance <= cellDistance ) {
                                    max = Math.max( max, d1 );
                                }
                            }
                        }
                    }
                    if ( max != Double.MIN_VALUE ) {
                        result.setCell( i, j, max );
                    }
                }
            }
        }
        
        // min
        if ( statistic.equalsIgnoreCase( "min" ) ) {
            double min = Double.MAX_VALUE;
            for ( int i = 0; i < nrows; i ++ ) {
                for ( int j = 0; j < ncols; j ++ ) {
                    min = Double.MAX_VALUE;
                    cellID1 = getCellID( i, j );
                    for ( int p = -cellDistance; p <= cellDistance; p ++ ) {
                        for ( int q = -cellDistance; q <= cellDistance; q ++ ) {
                            d1 = getCell( ( i + p ), ( j + q ) );
                            if ( d1 != noDataValue ) {
                                cellID2 = getCellID( ( i + p ), ( j + q ) );
                                thisCellDistance = cellDistance( cellID1, cellID2 );
                                if ( thisCellDistance <= cellDistance ) {
                                    min = Math.min( min, d1 );
                                }
                            }
                        }
                    }
                    if ( min != Double.MAX_VALUE ) {
                        result.setCell( i, j, min );
                    }
                }
            }
        }
        
        // mean
        if ( statistic.equalsIgnoreCase( "mean" ) ) {
            double sum = 0.0d;
            int count1 = 0;
            int count2 = 0;
            int count3 = 0;
            for ( int i = 0; i < nrows; i ++ ) {
                for ( int j = 0; j < ncols; j ++ ) {
                    sum = 0.0d;
                    count1 = 0;
                    count2 = 0;
                    count3 = 0;
                    cellID1 = getCellID( i, j );
                    for ( int p = -cellDistance; p <= cellDistance; p ++ ) {
                        for ( int q = -cellDistance; q <= cellDistance; q ++ ) {
                            count1 ++;
                            d1 = getCell( ( i + p ), ( j + q ) );
                            if ( d1 != noDataValue ) {
                                cellID2 = getCellID( ( i + p ), ( j + q ) );
                                thisCellDistance = cellDistance( cellID1, cellID2 );
                                if ( thisCellDistance <= cellDistance ) {
                                    sum += d1;
                                    count3 ++;
                                }
                            } else {
                                count2 ++;
                            }
                        }
                    }
                    if ( count1 != count2 ) {
                        result.setCell( i, j, sum / count3 );
                    }
                }
            }
        }
        
        // stddev
        if ( statistic.equalsIgnoreCase( "stddev" ) ) {
            AbstractGrid2DSquareCellDouble meanGrid = regionStats( gridFactory, "mean", distance );
            double sum = 0.0d;
            double mean = noDataValue;
            int count1 = 0;
            for ( int i = 0; i < nrows; i ++ ) {
                for ( int j = 0; j < ncols; j ++ ) {
                    mean = meanGrid.getCell( i, j );
                    if ( mean != noDataValue ) {
                        sum = 0.0d;
                        count1 = 0;
                        cellID1 = getCellID( i, j );
                        for ( int p = -cellDistance; p <= cellDistance; p ++ ) {
                            for ( int q = -cellDistance; q <= cellDistance; q ++ ) {
                                count1 ++;
                                d1 = getCell( ( i + p ), ( j + q ) );
                                if ( d1 != noDataValue ) {
                                    cellID2 = getCellID( ( i + p ), ( j + q ) );
                                    thisCellDistance = cellDistance( cellID1, cellID2 );
                                    if ( thisCellDistance <= cellDistance ) {
                                        sum += ( d1 - mean ) * ( d1 - mean );
                                        count1 ++;
                                    }
                                }
                            }
                        }
                        if ( count1 == 1 ) {
                            result.setCell( i, j, 0.0d );
                        } else {
                            result.setCell( i, j, Math.sqrt( sum / ( count1 - 1 ) ) );
                        }
                    }
                }
            }
        }
        // Warning!!! If statistics not a recognised word the result full of noDataValues is returned.
        //System.out.println( result.toString() );
        return result;
    }
    
    /**
     * Returns a smoothed AbstractGrid2DSquareCellDouble
     * ToDos:
     * 1. Add normalisationLevel parameter
     * 2. Add visualisationLevel parameter
     * 3. Enhance distance weightingFactor functionality so that other types of kernel
     *    can be applied.
     * @param cellDistance - the cell distance within which values will be used
     */
    public AbstractGrid2DSquareCellDouble smooth( int cellDistance ) {
        try {
            return smooth( cellDistance, 1.0d, new Grid2DSquareCellDoubleFactory() );
        } catch ( java.lang.OutOfMemoryError e ) {
            return smooth( cellDistance, 1.0d, new Grid2DSquareCellDoubleFileFactory() );
        }
    }
    /**
     * @param distanceWeightFactor - indicates the distance weightingFactor factor applied during smoothing
     */
    public AbstractGrid2DSquareCellDouble smooth( int cellDistance, double distanceWeightFactor ) {
        try {
            return smooth( cellDistance, distanceWeightFactor, new Grid2DSquareCellDoubleFactory() );
        } catch ( java.lang.OutOfMemoryError e ) {
            return smooth( cellDistance, distanceWeightFactor, new Grid2DSquareCellDoubleFileFactory() );
        }
    }
    /**
     * @param gridFactory - the AbstractFactory2DSquareCellDouble used to create results
     */
    public AbstractGrid2DSquareCellDouble smooth( int cellDistance, double distanceWeightFactor, AbstractGrid2DSquareCellDoubleFactory gridFactory ) {
        return smooth( ( double ) cellDistance * getCellsize(), distanceWeightFactor, gridFactory );
    }
    /**
     * @param distance - the map distance within which values will be used
     */
    public AbstractGrid2DSquareCellDouble smooth( double distance, double distanceWeightFactor, AbstractGrid2DSquareCellDoubleFactory gridFactory ) {
        //System.out.println( "smooth( distance(" + distance + "), distanceWeightFactor(" + distanceWeightFactor + "), gridFactory(" + gridFactory + ") )" );
        // Initialisation
        int ncols = getNcols();
        int nrows = getNrows();
        double xllcorner = getXllcorner();
        double yllcorner = getYllcorner();
        double cellsize = getCellsize();
        double noDataValue = getNoDataValue();
        int cellDistance = ( int ) Math.ceil( distance / cellsize );
        int cellID1 = 0;
        int cellID2 = 0;
        double thisCellDistance = 0.0d;
        double d1 = 0.0d;
        double d2 = 0.0d;
        double d3 = 0.0d;
        AbstractGrid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( nrows, ncols, xllcorner, yllcorner, cellsize, noDataValue );
        // Calculation
        for ( int i = 0; i < nrows; i ++ ) {
            for ( int j = 0; j < ncols; j ++ ) {
                d1 = 0.0d;
                d2 = 0.0d;
                cellID1 = getCellID( i, j );
                for ( int p = -cellDistance; p <= cellDistance; p ++ ) {
                    for ( int q = -cellDistance; q <= cellDistance; q ++ ) {
                        d3 = getCell( ( i + p ), ( j + q ) );
                        if (d3 != noDataValue) {
                            cellID2 = getCellID( ( i + p ), ( j + q ) );
                            thisCellDistance = cellDistance( cellID1, cellID2 );
                            if ( thisCellDistance <= cellDistance ) {
                                if ( thisCellDistance != 0.0d ) {
                                    d1 += ( d3 / Math.pow( thisCellDistance, distanceWeightFactor ) );
                                } else {
                                    d1 += d3;
                                }
                                d2 += 1.0d;
                            }
                        }
                    }
                }
                if ( d2 != 0.0d ) {
                    result.setCell( i, j, ( d1 / d2 ) );
                }
            }
        }
        //System.out.println( result.toString() );
        return result;
    }
    
    
    /*
     * DEM methods
     * ===========
     */
    
    /**
     * Slope
     * The AbstractGrid2DSquareCellDouble[0] returned is the average slope over the region
     * The AbstractGrid2DSquareCellDouble[1] returned is the aspect over the region
     * aspect is a function of the slope between each cell in the region and the cell being calculated.
     * @param grid - the AbstractGrid2DSquareCellDouble frame  within which values will be used
     */
    public AbstractGrid2DSquareCellDouble[] slope() {
        // Default distance to contain centroids of immediate neighbours
        double cellsize = getCellsize();
        double distance = cellsize + ( cellsize / 2.0d );
        double distanceWeightFactor = 1.0d;
        try {
            return slope( distance, distanceWeightFactor, new Grid2DSquareCellDoubleFactory() );
        } catch ( java.lang.OutOfMemoryError e ) {
            return slope( distance, distanceWeightFactor, new Grid2DSquareCellDoubleFileFactory() );
        }
    }
    public AbstractGrid2DSquareCellDouble[] slope( double distance, double distanceWeightFactor, AbstractGrid2DSquareCellDoubleFactory gridFactory ) {
        return slope( this, distance, distanceWeightFactor, gridFactory );
    }
    /**
     * @param outGrid - defines the returned AbstractGrid2DSquareCellDouble resolutions and dimensions
     */
    public AbstractGrid2DSquareCellDouble[] slope( AbstractGrid2DSquareCellDouble outGrid, double distance, double distanceWeightFactor, AbstractGrid2DSquareCellDoubleFactory gridFactory ) {
        // Initialisation
        int ncols = getNcols();
        int nrows = getNrows();
        double xllcorner = getXllcorner();
        double yllcorner = getYllcorner();
        double cellsize = getCellsize();
        double noDataValue = getNoDataValue();
        int outGridNcols = outGrid.getNcols();
        int outGridNrows = outGrid.getNrows();
        double outGridXllcorner = outGrid.getXllcorner();
        double outGridYllcorner = outGrid.getYllcorner();
        double outGridCellsize = outGrid.getCellsize();
        double outGridNoDataValue = outGrid.getNoDataValue();
        int cellDistance = ( int ) Math.ceil( distance / cellsize );
        double thisDistance;
        double x1;
        double y1;
        double z1;
        double x2;
        double y2;
        double z2;
        double x3;
        double y3;
        double z3;
        double xdiff;
        double ydiff;
        double zdiff;
        double slopeCounter = 0.0d;
        double aspect;
        boolean noDataTrigger;
        boolean aggregated = false;
        double averageZdiff;
        // Initialise result
        AbstractGrid2DSquareCellDouble slopeGrid = gridFactory.createGrid2DSquareCellDouble( outGridNrows, outGridNcols, outGridXllcorner, outGridYllcorner, outGridCellsize, outGridNoDataValue );
        AbstractGrid2DSquareCellDouble aspectGrid = gridFactory.createGrid2DSquareCellDouble( outGridNrows, outGridNcols, outGridXllcorner, outGridYllcorner, outGridCellsize, outGridNoDataValue );
        AbstractGrid2DSquareCellDouble[] result = new AbstractGrid2DSquareCellDouble[2];
        result[0] = slopeGrid;
        result[1] = aspectGrid;
        // Calculate slope and aspect
        // NB. There are various strategies to reduce bias caused by noDataValues
        // In this case:
        // 1. If the cell for which slope and aspect is being calculated is noDataValue then a noDataValue
        //    is returned.
        // 2. If one of the cells in the calculation of slope and aspect is a noDataValue then its 
        // difference in height is taken as the average difference in height for those cells with values.
        // Aggregation here copes with noData bias better
        if ( outGridCellsize > cellsize ) {
            outGrid = aggregate( outGridCellsize, "mean", outGridXllcorner, outGridYllcorner, noDataValue );
            aggregated = true;
        }
        int rowIndex = 0;
        int colIndex = 0;
        for ( int i = 0; i < outGridNrows; i ++ ) {
            for ( int j = 0; j < outGridNcols; j ++ ) {
                x1 = outGrid.getCellX( i, j );
                y1 = outGrid.getCellY( i, j );
                if ( aggregated ) {
                    z1 = outGrid.getCell( i, j );
                } else {
                    z1 = getCell( x1, y1 );
                }
                if ( z1 != noDataValue ) {
                    rowIndex = getRowIndex( y1 );
                    colIndex = getColIndex( x1 );
                    noDataTrigger = false;
                    x3 = 0.0d;
                    y3 = 0.0d;
                    z3 = 0.0d;
                    slopeCounter = 0.0d;
                    averageZdiff = 0.0d;
                    for ( int p = - ( cellDistance + 1 ); p <= ( cellDistance + 1 ); p ++ ) {
                        for ( int q = - ( cellDistance + 1 ); q <= ( cellDistance + 1 ); q ++ ) {
                            y2 = getCellY( ( rowIndex + p ), ( colIndex + q ) );
                            x2 = getCellX( ( rowIndex + p ), ( colIndex + q ) );
                            thisDistance = distance( x1, y1, x2, y2 );
                            if ( thisDistance <= distance && thisDistance != 0.0d ) {
                                z2 = getCell( ( rowIndex + p ), ( colIndex + q ) );
                                if ( z2 != noDataValue ) {
                                    xdiff = x1 - x2;
                                    ydiff = y1 - y2;
                                    zdiff = z1 - z2;
                                    averageZdiff += zdiff;
                                    z3 += zdiff / Math.pow( thisDistance, distanceWeightFactor );
                                    x3 += xdiff * zdiff / Math.pow( thisDistance, distanceWeightFactor );
                                    y3 += ydiff * zdiff / Math.pow( thisDistance, distanceWeightFactor );
                                    slopeCounter += 1.0d ;
                                    //slopeCounter += 1.0d / Math.pow( thisDistance, distanceWeightFactor );
                                } else {
                                    noDataTrigger = true;
                                }
                            }
                        }
                    }
                    if ( noDataTrigger && slopeCounter > 0.0d ) {
                        averageZdiff = averageZdiff / slopeCounter;
                        //averageZdiff = z3 / slopeCounter;
                        for ( int p = - ( cellDistance + 1 ); p <= ( cellDistance + 1 ); p ++ ) {
                            for ( int q = - ( cellDistance + 1 ); q <= ( cellDistance + 1 ); q ++ ) {
                                y2 = getCellY( ( rowIndex + p ), ( colIndex + q ) );
                                x2 = getCellX( ( rowIndex + p ), ( colIndex + q ) );
                                thisDistance = distance( x1, y1, x2, y2 );
                                if ( thisDistance <= distance && thisDistance != 0.0d ) {
                                    z2 = getCell( ( rowIndex + p ), ( colIndex + q ) );
                                    if ( z2 == noDataValue ) {
                                        xdiff = x1 - x2;
                                        ydiff = y1 - y2;
                                        z3 += averageZdiff / Math.pow( thisDistance, distanceWeightFactor );
                                        x3 += xdiff * averageZdiff / Math.pow( thisDistance, distanceWeightFactor );
                                        y3 += ydiff * averageZdiff / Math.pow( thisDistance, distanceWeightFactor );
                                        slopeCounter += 1.0d ;
                                        //slopeCounter += 1.0d / Math.pow( thisDistance, distanceWeightFactor );
                                    }
                                }
                            }
                        }
                    }
                    if ( slopeCounter > 0.0d ) {
                        slopeGrid.setCell( i, j, z3 / slopeCounter );
                        aspectGrid.setCell( i, j, angle( x1, y1, ( x3 + x1 ), ( y3 + y1 ) ) );
                    } else {
                        slopeGrid.setCell( i, j, 0.0d );
                        //aspectGrid.setCell( i, j, 0.0d );
                        aspectGrid.setCell( i, j, -1.0d );
                    }
                }
            }
        }
        return result;
    }
    
    /**
     * Returns an AbstractGrid2DSquareCellDouble whose values are the filled DEM
     * @param x the x-coordinate of the outflow cell
     * @param y the y-coordinate of the outflow cell
     */
    public AbstractGrid2DSquareCellDouble fillDEMHollows( double x, double y ) {
        try {
            return fillDEMHollows( new Grid2DSquareCellDoubleFactory(), x, y );
        } catch ( java.lang.OutOfMemoryError e ) {
            return fillDEMHollows( new Grid2DSquareCellDoubleFileFactory(), x, y );
        }
    }
    public AbstractGrid2DSquareCellDouble fillDEMHollows( AbstractGrid2DSquareCellDoubleFactory gridFactory, double x, double y ) {
        if ( ! inGrid( x, y ) ) {
            System.out.println( "Warning! ( x, y ) = ( " + x + ", " + y + " ) is not in the grid. Returning null." );
            return null;
        }
        // Initialisation
        AbstractGrid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( this );
        Integer outflowCellKey = new Integer( getCellID( x, y ) );
        int nrows = result.getNrows();
        int ncols = result.getNcols();
        double noDataValue = result.getNoDataValue();
        AbstractGridStatistics stats = result.getGridStatistics();
        double minHeight = stats.getMin();
        double heightTollerance = Utilities.minTollerance( minHeight, stats.getMax() ) * 10.0d;
        System.out.println( "heightTollerance " + heightTollerance );
        // Force the cellValue to which all cells will flow to be the lowest
        System.out.println( "The minimum height of the dem is " + minHeight );
        
        result.setCell( x, y, minHeight - heightTollerance );
        System.out.println( "Set ( " + x + ", " + y + " ) to " + ( minHeight - heightTollerance ) );
        System.out.println( "Set ( " + result.getRowIndex( y ) + ", " + result.getColIndex( x ) + " ) to " + ( minHeight - heightTollerance ) );
        // Initialise hollows
        HashSet hollows2 = initHollows( gridFactory );
        // Discount outflow as a hollow
        hollows2.remove( outflowCellKey );
        int numberOfHollows = hollows2.size();
        boolean calculated1 = false;
        boolean calculated2 = false;
        if ( numberOfHollows == 0 ) {
            calculated1 = true;
        }
        double height;
        int iteration1 = 0;
        int iteration2;
        int counter1 = 0;
        // maxIterations is hardcoded!
        int maxIterations = 2000;
        int row;
        int col;
        int cellID1;
        int cellID2;
        int cellID3;
        Iterator iterator1;
        Iterator iterator2;
        Integer cellKey1;
        Integer cellKey2;
        Integer cellKey3;
        HashSet toVisitSet1 = null;
        HashSet toVisitSet2 = null;
        HashSet toVisitSet3 = null;
        HashSet visitedSet1 = null;
        HashSet visitedSet2 = null;
        HashSet hollows1 = null;
        HashSet hollowsVisited = null;
        HashSet hollowSet = null;
        double height0;
        
        // Fill in hollows
        while ( ! calculated1 ) {
            if ( iteration1 < maxIterations ) {
                iteration1 ++;
                numberOfHollows = hollows2.size();
                System.out.println( "Iteration " + iteration1 + " out of a maximum " + maxIterations + ": Number of hollows " + numberOfHollows );
                if ( numberOfHollows > 0 ) {
                    visitedSet1 = new HashSet();
                    hollowsVisited = new HashSet();
                    hollowsVisited.add( outflowCellKey );
        
                    // Raise all hollows by heightTollerance
                    addToGrid( result, hollows2, heightTollerance );
                    
                    // Recalculate hollows in hollows2 neighbourhood
                    toVisitSet1 = new HashSet();
                    iterator1 = hollows2.iterator();
                    while ( iterator1.hasNext() ) {
                        cellKey1 = ( Integer ) iterator1.next();
                        cellID1 = cellKey1.intValue();
                        row = getRowIndex( cellID1 );
                        col = getColIndex( cellID1 );
                        for ( int p = -1; p < 2; p ++ ) {
                            for ( int q = -1; q < 2; q ++ ) {
                                if ( ! ( p == 0 && q == 0 ) ) {
                                    if ( inGrid( row + p, col + q ) ) {
                                        cellID2 = getCellID( row + p, col + q );
                                        cellKey2 = new Integer( cellID2 );
                                        toVisitSet1.add( cellKey2 );
                                    }
                                }
                            }
                        }
                    }
                    // If heightTollerance is really a minimimum this step is not necessary?
                    hollows1 = calculateHollows( result, toVisitSet1 );
                    hollows2.clear();
                    toVisitSet1.clear();
                    
                    // Trace bottom of each hollow and raise to the height of the lowest cell around it.
                    iterator1 = hollows1.iterator();
                    while ( iterator1.hasNext() ) {
                        cellKey1 = ( Integer ) iterator1.next();
                        if ( ! hollowsVisited.contains( cellKey1 ) ) {
                            hollowSet = new HashSet();
                            hollowSet.add( cellKey1 );
                            cellID1 = cellKey1.intValue();
                            row = getRowIndex( cellID1 );
                            col = getColIndex( cellID1 );
                            toVisitSet1 = new HashSet();
                            
                            // Step 1: Add all cells in adjoining hollows to hollowSet
                            for ( int p = -1; p < 2; p ++ ) {
                                for ( int q = -1; q < 2; q ++ ) {
                                    if ( ! ( p == 0 && q == 0 ) ) {
                                        if ( inGrid( row + p, col + q ) ) {
                                            cellID2 = getCellID( row + p, col + q );
                                            cellKey2 = new Integer( cellID2 );
                                            toVisitSet1.add( cellKey2 );
                                        }
                                    }
                                }
                            }
                            toVisitSet1.remove( outflowCellKey );
                            visitedSet2 = new HashSet();
                            visitedSet2.add( cellKey1 );
                            toVisitSet3 = new HashSet();
                            toVisitSet3.addAll( toVisitSet1 );
                            calculated2 = false;
                            while ( ! calculated2 ) {
                                toVisitSet2 = new HashSet();
                                iterator2 = toVisitSet1.iterator();
                                while ( iterator2.hasNext() ) {
                                    cellKey2 = ( Integer ) iterator2.next();
                                    visitedSet2.add( cellKey2 );
                                    cellID2 = cellKey2.intValue();
                                    row = getRowIndex( cellID2 );
                                    col = getColIndex( cellID2 );
                                    for ( int p = -1; p < 2; p ++ ) {
                                        for ( int q = -1; q < 2; q ++ ) {
                                            if ( ! ( p == 0 && q == 0 ) ) {
                                                if ( inGrid( row + p, col + q ) ) {
                                                    cellID3 = getCellID( row + p, col + q );
                                                    cellKey3 = new Integer( cellID3 );
                                                    visitedSet1.add( cellKey3 );
                                                    // If a hollow then add to hollow set and visit neighbours if not done already
                                                    if ( hollows1.contains( cellKey3 ) ) {
                                                        hollowSet.add( cellKey3 );
                                                        for ( int r = -1; r < 2; r ++ ) {
                                                            for ( int s = -1; s < 2; s ++ ) {
                                                                if ( ! ( r == 0 && s == 0 ) ) {
                                                                    if ( inGrid( row + p + r, col + q + s) ) {
                                                                        toVisitSet2.add( new Integer( getCellID( row + p + r, col + q + s ) ) );
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                toVisitSet2.remove( outflowCellKey );
                                toVisitSet3.addAll( toVisitSet2 );
                                toVisitSet1 = toVisitSet2;
                                toVisitSet1.removeAll( visitedSet2 );
                                if ( toVisitSet1.isEmpty() ) {
                                    calculated2 = true;
                                }
                            }
                            
                            // Step 2 Examine neighbours of each hollow
                            toVisitSet3.removeAll( hollowSet );
                            // NB. toVisitSet3 contains all cells which neighbour the traced hollow
                            calculated2 = false;
                            minHeight = Double.MAX_VALUE;
                            height0 = result.getCell( row, col );
                            while ( ! calculated2 ) {
                                toVisitSet2 = new HashSet();
                                toVisitSet2.addAll( toVisitSet3 );
                                iterator2 = toVisitSet3.iterator();
                                // Step 2.1 Calculate height of the lowest neighbour minHeight
                                while ( iterator2.hasNext() ) {
                                    cellKey2 = ( Integer ) iterator2.next();
                                    cellID2 = cellKey2.intValue();
                                    row = getRowIndex( cellID2 );
                                    col = getColIndex( cellID2 );
                                    height = result.getCell( row, col );
                                    if ( height != noDataValue ) {
                                        minHeight = Math.min( minHeight, height );
                                    }
                                }
                                // Step 2.2 Treat cells:
                                // If minHeight is higher then add cells with this height to the 
                                // hollow set and their neighbours to toVisitSet2
                                if ( minHeight > height0 ) {
                                    iterator2 = toVisitSet3.iterator();
                                    while ( iterator2.hasNext() ) {
                                        cellKey2 = ( Integer ) iterator2.next();
                                        cellID2 = cellKey2.intValue();
                                        row = getRowIndex( cellID2 );
                                        col = getColIndex( cellID2 );
                                        height = result.getCell( row, col );
                                        if ( height == minHeight ) {
                                            hollowSet.add( cellKey2 );
                                            toVisitSet2.remove( cellKey2 );
                                            for ( int r = -1; r < 2; r ++ ) {
                                                for ( int s = -1; s < 2; s ++ ) {
                                                    if ( ! ( r == 0 && s == 0 ) ) {
                                                        if ( inGrid( row + r, col + s ) ) {
                                                            cellID3 = getCellID( row + r, col + s );
                                                            cellKey3 = new Integer( cellID3 );
                                                            toVisitSet2.add( cellKey3 );
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    height0 = minHeight;
                                    toVisitSet2.removeAll( hollowSet );
                                    //toVisitSet2.remove( outflowCellKey );
                                    toVisitSet3 = toVisitSet2;
                                } else {
                                    calculated2 = true;
                                }
                            }
                            
                            // Step 3 Raise all cells in hollowSet
                            hollowSet.remove( outflowCellKey );
                            iterator2 = hollowSet.iterator();
                            while ( iterator2.hasNext() ) {
                                cellKey2 = ( Integer ) iterator2.next();
                                cellID2 = cellKey2.intValue();
                                row = getRowIndex( cellID2 );
                                col = getColIndex( cellID2 );
                                result.setCell( row, col, height0 );
                            }
                            hollowsVisited.addAll( hollowSet );
                            visitedSet1.addAll( hollowSet );
                       }
                    }
                    hollows2 = calculateHollows( result, visitedSet1 );
                } else {
                    calculated1 = true;
                }
            } else {
                calculated1 = true;
            }
        }
        return result;
    }
    
    /**
     * Returns a HashSet containing cellKeys which identifies cells in this which
     * neighbouring cells are either the same value, higher or noDataValues
     */
    protected HashSet initHollows( AbstractGrid2DSquareCellDoubleFactory gridFactory ) {
        HashSet result = new HashSet();
        String dataDirectory = gridFactory.getDataDirectory();
        int nrows = getNrows();
        int ncols = getNcols();
        double noDataValue = getNoDataValue();
        AbstractGrid2DSquareCellDouble initialHollowsGrid = null;
        double[ ] heights = new double[ 9 ];
        int k;
        // Initialise hollows
        for ( int i = 0; i < nrows; i ++ ) {
            for ( int j = 0; j < ncols; j ++ ) {
                heights[ 0 ] = getCell( i, j );
                if ( heights[ 0 ] != noDataValue ) {
                    k = 0;
                    for ( int p = -1; p < 2; p ++ ) {
                        for ( int q = -1; q < 2; q ++ ) {
                            if ( ! ( p == 0 && q == 0 ) ) {
                                k ++;
                                heights[ k ] = getCell( i + q, j + p );
                            }
                        }
                    }
                    if ( ( heights[ 1 ] >= heights[ 0 ] || heights[ 1 ] == noDataValue ) &&
                    ( heights[ 2 ] >= heights[ 0 ] || heights[ 2 ] == noDataValue ) &&
                    ( heights[ 3 ] >= heights[ 0 ] || heights[ 3 ] == noDataValue ) &&
                    ( heights[ 4 ] >= heights[ 0 ] || heights[ 4 ] == noDataValue ) &&
                    ( heights[ 5 ] >= heights[ 0 ] || heights[ 5 ] == noDataValue ) &&
                    ( heights[ 6 ] >= heights[ 0 ] || heights[ 6 ] == noDataValue ) &&
                    ( heights[ 7 ] >= heights[ 0 ] || heights[ 7 ] == noDataValue ) &&
                    ( heights[ 8 ] >= heights[ 0 ] || heights[ 8 ] == noDataValue ) ) {
                        result.add( new Integer( getCellID( i, j ) ) );
                    }
                }
            }
        }
        return result;
    }
    
    /**
     * Returns a HashSet containing ID Keys of any hollows found in the demGrid at the cell locations 
     * given by cellKeys
     */
    private HashSet calculateHollows( AbstractGrid2DSquareCellDouble demGrid, HashSet cellKeys ) {
        double noDataValue = demGrid.getNoDataValue();
        HashSet result = new HashSet();
        HashSet visited1 = new HashSet();
        Integer cellKey1;
        int cellID1;
        int row;
        int col;
        double[ ] heights = new double[ 9 ];
        int k;
        Iterator iterator1 = cellKeys.iterator();
        while ( iterator1.hasNext() ) {
            cellKey1 = ( Integer ) iterator1.next();
            if ( ! visited1.contains( cellKey1 ) ) {
                cellID1 = cellKey1.intValue();
                row = getRowIndex( cellID1 );
                col = getColIndex( cellID1 );
                // Examine neighbourhood
                for ( int a = -1; a < 2; a ++ ) {
                    for ( int b = -1; b < 2; b ++ ) {
                        visited1.add( new Integer( getCellID( row + a, col + b ) ) );
                        heights[ 0 ] = demGrid.getCell( row + a, col + b );
                        if ( heights[ 0 ] != noDataValue ) {
                            k = 0;
                            for ( int p = -1; p < 2; p ++ ) {
                                for ( int q = -1; q < 2; q ++ ) {
                                    if ( ! ( p == 0 && q == 0 ) ) {
                                        k ++;
                                        heights[ k ] = demGrid.getCell( row + a + q, col + b + p );
                                    }
                                }
                            }
                            if ( ( heights[ 1 ] >= heights[ 0 ] || heights[ 1 ] == noDataValue ) &&
                            ( heights[ 2 ] >= heights[ 0 ] || heights[ 2 ] == noDataValue ) &&
                            ( heights[ 3 ] >= heights[ 0 ] || heights[ 3 ] == noDataValue ) &&
                            ( heights[ 4 ] >= heights[ 0 ] || heights[ 4 ] == noDataValue ) &&
                            ( heights[ 5 ] >= heights[ 0 ] || heights[ 5 ] == noDataValue ) &&
                            ( heights[ 6 ] >= heights[ 0 ] || heights[ 6 ] == noDataValue ) &&
                            ( heights[ 7 ] >= heights[ 0 ] || heights[ 7 ] == noDataValue ) &&
                            ( heights[ 8 ] >= heights[ 0 ] || heights[ 8 ] == noDataValue ) ) {
                                result.add( new Integer( getCellID( row + a, col + b ) ) );
                            }
                        }
                    }
                }
            }
        }
        return result;
    }
    
    /**
     * Returns a grid containing values which indicate the flow direction in the immediate 8 cell neighbourhood.
     * The max flow direction is a pointer to the cell which is the most downhill.
     * 1 2 3
     * 4 0 5
     * 6 7 8
     * If there is no downhill slope then the flow direction is 0.
     */
    protected AbstractGrid2DSquareCellDouble initMaxFlowDirection( AbstractGrid2DSquareCellDoubleFactory gridFactory, AbstractGrid2DSquareCellDouble demGrid ) {
        String dataDirectory = gridFactory.getDataDirectory();
        int gridStatisticsType = 0;
        int nrows = demGrid.getNrows();
        int ncols = demGrid.getNcols();
        double xllcorner = demGrid.getXllcorner();
        double yllcorner = demGrid.getYllcorner();
        double cellsize = demGrid.getCellsize();
        double noDataValue = demGrid.getNoDataValue();
        AbstractGrid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( nrows, ncols, xllcorner, yllcorner, cellsize, noDataValue, gridStatisticsType );
        Integer cellKey1;
        int cellID1;
        int row;
        int col;
        int k;
        int[] flowDirections = new int[ 9 ];
        int flowDirection = 0;
        double[ ] z = new double[ 9 ];
        double minz;
        int minzCount;
        int minzCountNoDataValue;
        for ( int i = 0; i < nrows; i ++ ) {
            for ( int j = 0; j < ncols; j ++ ) {
                z[ 0 ] = demGrid.getCell( i, j );
                if ( z[ 0 ] != noDataValue ) {
                    minz = Double.MAX_VALUE;
                    minzCount = 0;
                    minzCountNoDataValue = 0;
                    flowDirection = 0;
                    k = 0;
                    for ( int p = -1; p < 2; p ++ ) {
                        for ( int q = -1; q < 2; q ++ ) {
                            if ( ! ( p == 0 && q == 0 ) ) {
                                k ++;
                                z[ k ] = demGrid.getCell( i + p, j + q );
                                if ( z[ k ] != noDataValue ) {
                                    if ( z[ k ] <= minz && z[ k ] < z[ 0 ] ) {
                                        if ( z[ k ] == minz ) {
                                            minzCount ++;
                                        } else {
                                            minz = z[ k ];
                                            minzCount = 1;
                                            flowDirection = k;
                                        }
                                    }
                                } else {
                                    minzCountNoDataValue ++;
                                }
                            }
                        }
                    }
                    // If more than one flowDirection randomly assign one
                    if ( minzCount + minzCountNoDataValue > 1 ) {
                        int[] min = new int[ minzCount + minzCountNoDataValue ];
                        int minID = 0;
                        double random = Math.random();
                        for ( int k2 = 1; k2 < z.length; k2 ++ ) {
                            if ( z[ k2 ] == minz || z[ k2 ] == noDataValue ) {
                                min[ minID ] = k2;
                                minID ++;
                            }
                        }
                        flowDirection = min[ ( int ) Math.floor( random * ( minzCount +  minzCountNoDataValue ) ) ];
                    }
                    result.setCell( i, j, ( double ) flowDirection );
                }
            }
        }
        return result;
    }
    
    /*
     * geometricDensity methods
     * ========================
     */
    
    /**
     * geometricDensity methods
     * The algorithm for generating a geometric density surface is described in:
     * Turner A (2000) Density Data Generation for Spatial Data Mining Applications.
     * http://www.geog.leeds.ac.uk/people/a.turner/papers/geocomp00/gc_017.htm
     *
     * Notes:
     * 1. If normalisationLevel = 0 then there is no normalisation.
     * 2. If normalisationLevel = 1 then the input and the output are forced into the
     *   range [0.0d,1.0d].
     * 3. If normalization level = 2 then the input the output and the density output
     *   at each geometric scale (iteration) are forced into the range [0.0d,1.0d].
     * 4. The cellDistance parameter allows control for the number of iterations.
     *   By default it is the maximum of the rows and columns. It is effectively
     *   a limit beyond which cells of the grid are unrelated.
     * 5. The chunkSize parameter controls the size of each chunk processed. By
     *   default it is the cellDistance. If it is not >= cellDistance then it is set
     *   to cellDistance.
     */
    
    public final void geometricDensity() {
        // Default cellDistance to be half the maximum of the number of rows and number of colums.
        geometricDensity( Math.max( getNrows(), getNcols() ) / 2 ) ;
    }
    
    public final void geometricDensity( int cellDistance ) {
        // Default weightingFactor to one.
        geometricDensity( Math.max( getNrows(), getNcols() ) / 2, 0 ) ;
    }
    
    /**
     * @param weightingFactor - used to combine density results from different scales
     *  weightingFactor > 1 weights higher scale density more
     *  weightingFactor = 1 weights density at all scales the same
     *  weightingFactor < 1 weights lower scale density more
     */
    public final void geometricDensity( int cellDistance, double weightingFactor ) {
        AbstractGrid2DSquareCellDouble density;
        AbstractGrid2DSquareCellDoubleFactory factory;
        AbstractGrid2DSquareCellDouble[] densityArray = null;
        int nrows = getNrows();
        int ncols = getNcols();
        double noDataValue = getNoDataValue();
        double weight;
        double d1;
        boolean chunkProcess = false;
        try {
            densityArray = geometricDensity( cellDistance, new Grid2DSquareCellDoubleFactory() ) ;
        } catch ( java.lang.OutOfMemoryError e ) {
            if ( cellDistance < ( Math.max( getNrows(), getNcols() ) / 2 ) ) {
                System.out.println( e.toString() + "... Attempting to process in chunks..." );
                int chunkSize = cellDistance;
                density = geometricDensity( cellDistance, weightingFactor, chunkSize );
                for ( int cellID = 0; cellID < nrows * ncols; cellID ++ ) {
                    this.setCell( cellID, density.getCell( cellID ) );
                }
                chunkProcess = true;
            } else {
                System.out.println( e.toString() + "... Processing using available filespace..." );
                densityArray = geometricDensity( cellDistance, new Grid2DSquareCellDoubleFileFactory() );
            }
        }
        if ( ! chunkProcess ) {
            for ( int cellID = 0; cellID < nrows * ncols; cellID ++ ) {
                d1 = getCell( cellID );
                // This should mask
                if ( d1 != noDataValue ) {
                    for ( int i = 0; i < densityArray.length; i ++ ) {
                        weight = Math.pow( weightingFactor, i );
                        density = densityArray[i];
                        this.addToCell( cellID, densityArray[i].getCell( cellID ) * weight );
                    }
                }
            }
        }
    }
    
    /**
     * geometricDensity
     * @param cellDistance - the distance in cell units within which values for an output cell at the
     *  largest geometry are factored
     * @param gridFactory - the factory from which AbstractGrid2DSquareCellDoubles are made
     */
    public final AbstractGrid2DSquareCellDouble[] geometricDensity( int cellDistance, AbstractGrid2DSquareCellDoubleFactory gridFactory ) {
        //System.out.println( "geometricDensity( cellDistance(" + cellDistance + "), gridFactory(" + gridFactory + ") )" );
        String dataDirectory = gridFactory.getDataDirectory();
        AbstractGridStatistics gridStatistics = getGridStatistics();
        double sparseness = gridStatistics.getSparseness();
        int nrows = getNrows();
        int ncols = getNcols();
        double xllcorner = getXllcorner();
        double yllcorner = getYllcorner();
        double cellsize = getCellsize();
        double noDataValue = getNoDataValue();
        double d1;
        double d2;
        double d3;
        int height = nrows;
        int width = ncols;
        int i1;
        int numberOfIterations;
        int doubler = 1;
        int growth = 1;
        // Calculate number of iterations and initialise result.
        numberOfIterations = 0;
        i1 = Math.min( cellDistance, ( Math.max( nrows, ncols ) / 2 ) );
        for ( int i = 0; i < cellDistance; i++ ) {
            if ( i1 > 1 ) {
                i1 = i1 / 2;
                numberOfIterations ++;
            } else {
                break;
            }
        }
        AbstractGrid2DSquareCellDouble[] result = new AbstractGrid2DSquareCellDouble[numberOfIterations];
        AbstractGrid2DSquareCellDouble density = gridFactory.createGrid2DSquareCellDouble( nrows, ncols, xllcorner, yllcorner, cellsize, noDataValue );
        // If all values are noDataValues no calculation neccessary.
        if ( sparseness == 1.0d ) {
            for ( int i = 0; i < numberOfIterations; i ++ ) {
                result[i] = density;
            }
            return result;
        }
        // Initialise temporary numerator and normaliser grids
        AbstractGrid2DSquareCellDouble g2 = gridFactory.createGrid2DSquareCellDouble( nrows, ncols, xllcorner, yllcorner, cellsize, 0.0d );
        AbstractGrid2DSquareCellDouble g3 = gridFactory.createGrid2DSquareCellDouble( nrows, ncols, xllcorner, yllcorner, cellsize, 0.0d );
        for ( int i = 0; i < nrows; i ++ ) {
            for ( int j = 0; j < ncols; j ++ ) {
                d1 = getCell( i, j );
                if ( d1 != noDataValue ) {
                    g2.initCell( i, j, d1 );
                    g3.initCell( i, j, 1.0d );
                }
            }
        }
        // Densification
        AbstractGrid2DSquareCellDouble g4;
        AbstractGrid2DSquareCellDouble g5;
        AbstractGrid2DSquareCellDouble g6;
        AbstractGrid2DSquareCellDouble g7;
        for ( int iteration = 0; iteration < numberOfIterations; iteration ++ ) {
            //System.out.println( "Iteration " + ( iteration + 1 ) + " out of " + numberOfIterations );
            height += doubler;
            width += doubler;
            growth *= 2;
            // Step 1: Aggregate
            g4 = gridFactory.createGrid2DSquareCellDouble( height, width, 0.0d, 0.0d, cellsize, 0.0d );
            g5 = gridFactory.createGrid2DSquareCellDouble( height, width, 0.0d, 0.0d, cellsize, 0.0d );
            g6 = gridFactory.createGrid2DSquareCellDouble( height, width, 0.0d, 0.0d, cellsize, 0.0d );
            for ( int p = 0; p < doubler; p ++ ) {
                for ( int q = 0; q < doubler; q ++ ) {
                    for ( int i = 0; i < height; i += doubler ) {
                        for ( int j = 0; j < width; j += doubler ) {
                            d1 = g2.getCell( ( i + p ), ( j + q ) ) + g2.getCell( ( i + p ), ( j + q - doubler ) ) + g2.getCell( ( i + p - doubler ), ( j + q ) ) + g2.getCell( ( i + p - doubler ), ( j + q - doubler ) );
                            g4.initCell( ( i + p ), ( j + q ), d1 );
                            d2 = g3.getCell( ( i + p ), ( j + q ) ) + g3.getCell( ( i + p ), ( j + q - doubler ) ) + g3.getCell( ( i + p - doubler ), ( j + q ) ) + g3.getCell( ( i + p - doubler ), ( j + q - doubler ) );
                            g5.initCell( ( i + p ), ( j + q ), d2 );
                            if (d2 != 0.0d) {
                                g6.initCell( i, j, ( d1 / d2 ) );
                            }
                        }
                    }
                }
            }
            g2.clear();
            g3.clear();
            // Step 2: Average over output region.
            // 1. This is probably the slowest part of the algorithm and gets slower
            //    with each iteration. Using alternative data structures and
            //    processing strategies this step can probably be speeded up a lot.
            density = gridFactory.createGrid2DSquareCellDouble( nrows, ncols, 0.0d, 0.0d, cellsize, 0.0d );
            for ( int i = 0; i < nrows; i += doubler ) {
                for ( int p = 0; p < doubler; p ++ ) {
                    for ( int j = 0; j < ncols; j += doubler ) {
                        for ( int q = 0; q < doubler; q ++ ) {
                            d1 = 0.0d;
                            d2 = 0.0d;
                            for ( int a = 0; a < growth; a ++ ) {
                                for ( int b = 0; b < growth; b ++ ) {
                                    if ( g6.inGrid( ( i + p + a ), ( j + q + b ) ) ) {
                                        d1 += g6.getCell( ( i + p + a ), ( j + q + b ) );
                                        d2 += 1.0d;
                                    }
                                }
                            }
                            if ( d2 != 0.0d ) {
                                density.initCell( ( i + p ), ( j + q ), ( d1 / d2 ) );
                            }
                        }
                    }
                }
            }
            g6.clear();
            result[iteration] = density;
            doubler *= 2;
            g2 = g4;
            g3 = g5;
        }
        return result;
    }
    
    public final AbstractGrid2DSquareCellDouble geometricDensity( int cellDistance, double weightingFactor, int chunkSize ) {
        try {
            return geometricDensity( cellDistance, weightingFactor, chunkSize, new Grid2DSquareCellDoubleFactory() );
        } catch ( java.lang.OutOfMemoryError e ) {
            return geometricDensity( cellDistance, weightingFactor, chunkSize, new Grid2DSquareCellDoubleFileFactory() );
        }
    }
    
    public final AbstractGrid2DSquareCellDouble geometricDensity( int cellDistance, double weightingFactor, int chunkSize, AbstractGrid2DSquareCellDoubleFactory gridFactory ) {
        //System.out.println( "geometricDensity( cellDistance(" + cellDistance + "), weightingFactor(" + weightingFactor + "), chunkSize(" + chunkSize + "), gridFactory (" + gridFactory + " ) )" );
        // Initialisation
        String dataDirectory = gridFactory.getDataDirectory();
        int nrows = getNrows();
        int ncols = getNcols();
        double xllcorner = getXllcorner();
        double yllcorner = getYllcorner();
        double cellsize = getCellsize();
        double noDataValue = getNoDataValue();
        AbstractGrid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( nrows, ncols, xllcorner, yllcorner, cellsize, noDataValue, 1 );
        int chunks = 0;
        int colChunks = 0;
        int rowChunks = 0;
        int startRow = 0;
        int endRow = 0;
        int startCol = 0;
        int endCol = 0;
        int i1;
        int numberOfIterations, doubler, growth, lastrows, lastcols, thisrows, thiscols;
        double n1,d1,d2;
        int startRowIndex, endRowIndex, startColIndex, endColIndex;
        boolean alternator;
        // Initialise temporary files f1
        File f1 = Utilities.makeTemporaryFile( dataDirectory, "f1" );
        try {
            RandomAccessFile rf1 = new RandomAccessFile( f1, new String( "rw" ) );
            // Split the problem into chunks
            if ( chunkSize < ( cellDistance * 3 ) ) {
                chunkSize = cellDistance * 3;
            }
            // Calculate the number of the chunks of this size
            if ( ( ncols % chunkSize ) > 0 ) {
                colChunks = ( ncols / chunkSize ) + 1;
            } else {
                colChunks = ncols / chunkSize;
            }
            if ( ( nrows % chunkSize ) > 0 ) {
                rowChunks = ( nrows / chunkSize ) + 1;
            } else {
                rowChunks = nrows / chunkSize;
            }
            chunks  = rowChunks * colChunks;
            //System.out.println( "chunks(" + chunks + "), rowChunks(" + rowChunks + "), colchunks(" + colChunks + ") chunkSize(" + chunkSize + ") cellDistance(" + cellDistance + ")" );
            // Set mallarchery loop variables
            startRow = ( 0 - cellDistance );
            endRow = ( chunkSize - 1 ) + cellDistance;
            startCol = ( 0 - cellDistance - chunkSize );
            endCol = ( cellDistance - 1 );
            int extraCols = 0;
            int extraRows = 0;
            int tendRow = endRow;
            int tncols = 0;
            int tnrows = 0;
            for ( int mallarchery = 0; mallarchery < chunks; mallarchery ++ ) {
                //System.out.println( "Chunk " + ( mallarchery + 1 ) + " out of " + chunks );
                // Set mallarchery loop variables
                startCol += chunkSize;
                endCol += chunkSize;
                // Correct for short Col chunk
                if ( ( endCol - cellDistance ) > ( ncols - 1 ) ) {
                    extraCols = endCol - ( ( ncols - 1 ) + cellDistance );
                    tncols = chunkSize - extraCols;
                    endCol = ( ncols - 1 ) + cellDistance;
                } else {
                    extraCols = 0;
                    tncols = chunkSize;
                }
                // Correct for short Row chunks
                if ( ( endRow - cellDistance ) > ( nrows - 1 ) ) {
                    tendRow = endRow;
                    extraRows = tendRow - ( ( nrows - 1 ) + cellDistance );
                    tnrows = chunkSize - extraRows;
                    endRow = ( nrows - 1 ) + cellDistance;
                } else {
                    tendRow = endRow;
                    extraRows = 0;
                    tnrows = chunkSize;
                }
                //System.out.println( "Values used:" );
                //System.out.println( "(nrows,ncols)=(" + nrows + "," + ncols + ")" );
                //System.out.println( "(cellDistance,chunkSize)=(" + cellDistance + "," + chunkSize + ")" );
                //System.out.println( "(startRow,startCol)=(" + startRow + "," + startCol + ")" );
                //System.out.println( "(endRow,endCol)=(" + endRow + "," + endCol + ")" );
                //System.out.println( "(extraRows,extraCols)=(" + extraRows + "," + extraCols + ")");
                //System.out.println( "(tnrows,tncols)=(" + tnrows + "," + tncols + ")");
                // Try to create and process chunk using fast access memory.
                //System.out.println("Attempting to process chunk in memory...");
                AbstractGrid2DSquareCellDouble chunk;
                int gridStatisticsType = 0;
                try {
                    chunk = new Grid2DSquareCellDouble( this, startRow, startCol, endRow, endCol, gridStatisticsType );
                } catch ( java.lang.OutOfMemoryError e ) {
                    chunk = new Grid2DSquareCellDoubleFile( Utilities.makeTemporaryFile( dataDirectory, "chunk" ), this, startRow, startCol, endRow, endCol, gridStatisticsType );
                }
                chunk.geometricDensity( cellDistance, weightingFactor );
                // Write out this chunks result to rf1.
                chunk.toBinaryFile( rf1, true, cellDistance, cellDistance, ( cellDistance + tnrows - 1 ), ( cellDistance + tncols - 1 ) );
                // Tidy up
                chunk.clear();
                // Set mallarchery loop variables
                // Set startRow and endRow for next iteration
                startCol += chunkSize;
                endCol += chunkSize;
                if ( ( startCol + cellDistance) > ( ncols - 1 ) ) {
                    startRow += chunkSize;
                    endRow += chunkSize;
                    startCol = ( 0 - cellDistance - chunkSize );
                    endCol = ( cellDistance - 1 );
                } else {
                    startCol -= chunkSize;
                    endCol -= chunkSize;
                    endRow = tendRow;
                }
            }
            // Read temporary file back in and write out result to g2.
            // This is effectively a row by row grid merge operation
            for ( int p = 0; p < rowChunks; p ++ ) {
                //System.out.println("rowChunk "+p);
                for ( int i = 0; i < chunkSize; i ++ ) {
                    for ( int q = 0; q < colChunks; q ++ ) {
                        //System.out.println( "colChunk " + q );
                        // if not beyond last row
                        if ( ( ( p * chunkSize ) + i ) < nrows ) {
                            // The last rowChunk and the last colChunk are special cases
                            if ( ( p + 1 ) == rowChunks ) {
                                if ( ( q + 1 ) == colChunks ) {
                                    rf1.seek( ( ( p * ( chunkSize * ( ( chunkSize * ( colChunks - 1 ) ) + tncols ) ) ) + ( q * chunkSize * tnrows ) + ( i * tncols ) ) * 8 );
                                } else {
                                    rf1.seek( ( ( p * ( chunkSize * ( ( chunkSize * ( colChunks - 1 ) ) + tncols ) ) ) + ( q * chunkSize * tnrows ) + ( i * chunkSize ) ) * 8 );
                                }
                            } else {
                                if ( ( q + 1 ) == colChunks ) {
                                    rf1.seek( ( ( p * ( chunkSize * ( ( chunkSize * ( colChunks - 1 ) ) + tncols ) ) ) + ( q * chunkSize * chunkSize ) + ( i * tncols ) ) * 8 );
                                } else {
                                    rf1.seek( ( ( p * ( chunkSize * ( ( chunkSize * ( colChunks - 1 ) ) + tncols ) ) ) + ( q * chunkSize * chunkSize ) + ( i * chunkSize ) ) * 8 );
                                }
                            }
                            for ( int j = 0; j < chunkSize; j ++ ) {
                                // if not beyond last column
                                if ( ( ( q * chunkSize ) + j ) < ncols ) {
                                    //g2.initCell(((int)p*chunkSize)+i,(int)(q*chunkSize)+j,rf1.readDouble());
                                    //result.setCell( ( ( p * chunkSize ) + i ), ( ( q * chunkSize ) + j ), rf1.readDouble() );
                                    result.setCell( ( ( p * chunkSize ) + i ), ( ( q * chunkSize ) + j ), rf1.readDouble() );
                                    //System.out.println( ( ( p * chunkSize ) + i ), ( ( q * chunkSize ) + j ), rf1.readDouble() );
                                }
                            }
                        }
                    }
                }
            }
            // Tidy up
            rf1.close();
            f1.delete();
        } catch ( java.io.IOException e ) {
            System.out.println( e.toString() + " in geometricDensity( cellDistance(" + cellDistance + "), weightingFactor(" + weightingFactor + "), chunkSize(" + chunkSize + "), gridFactory (" + gridFactory + ") )" );
            System.exit( 0 );
        }
        // Return result
        return result;
    }
    
    
    /**
     * ============================
     * Operations on multiple grids
     * ============================
     */
    
    /**
     * This compares this AbstractGrid2DSquareCellDouble values with inputGrid values and returns the answer
     * as a AbstractGrid2DSquareCellDouble
     * @param inputGrid - the Abstract2DSquareCellDouble from which the difference will be calculated
     * @param compare - the name of the function to be used to compare the grids options include:
     * diff - the distance weighted sum of this cells minus the distance weighted sum of the inputGrid cells
     * abs - the sum of all the absolute differences in distance weighted cell values
     * pearson - the pearsons rank correlation coefficient (Pearson product moment correlation coefficient)
     * correl -
     * rsq -
     * @TODO
     * 1. Check and ensure that reasonable answers are returned for grids with different spatial frames.
     * 2. In some cases the kernels distance weights can be calculated once then passed over the grid rather
     *    than being recalculated each time.
     * Notes
     * 1. There are many different kernels that can be used:
     *    Math.pow( distance, distanceWeightFactor ) / Math.pow( bandWidth or radius, distanceWeightFactor )
     */
    public final AbstractGrid2DSquareCellDouble compare( AbstractGrid2DSquareCellDouble inputGrid, String comparator ) {
        try {
            return compare( new Grid2DSquareCellDoubleFactory(), inputGrid, comparator );
        } catch ( OutOfMemoryError oome1 ) {
            String dataDirectory = System.getProperty( "java.io.tmpdir" );
            if ( inputGrid instanceof Grid2DSquareCellDoubleFile ) {
                dataDirectory = ( ( Grid2DSquareCellDoubleFile ) inputGrid ).getDataDirectory();
            }
            System.out.println( "Run out of memory! Attempting to reprocess using filespace in directory " + dataDirectory );
            return compare( new Grid2DSquareCellDoubleFileFactory( dataDirectory ), inputGrid, comparator );
        }
    }
    /**
     * @param gridFactory - the Abstract2DSquareCellDoubleFactory used to create the output grid
     */
    public final AbstractGrid2DSquareCellDouble compare( AbstractGrid2DSquareCellDoubleFactory gridFactory, AbstractGrid2DSquareCellDouble inputGrid, String comparator ) {
        return compare( gridFactory, inputGrid, comparator, 0 );
    }
    /**
     * @param cellDistance - the cell distance within which this AbstractGrid2DSquareCellDouble and inputGrid
     * are to be compared ( default = 0 )
     */
    public final AbstractGrid2DSquareCellDouble compare( AbstractGrid2DSquareCellDoubleFactory gridFactory, AbstractGrid2DSquareCellDouble inputGrid, String comparator, int cellDistance ) {
        return compare( gridFactory, inputGrid, comparator, cellDistance, 0 );
    }
    /**
     * @param cellDistance - the cell distance within which this AbstractGrid2DSquareCellDouble and inputGrid
     * are to be compared ( default = 0 )
     */
    public final AbstractGrid2DSquareCellDouble compare( AbstractGrid2DSquareCellDoubleFactory gridFactory, AbstractGrid2DSquareCellDouble inputGrid, String comparator, int cellDistance, double distanceWeightFactor ) {
        return compare( gridFactory, inputGrid, comparator, cellDistance * getCellsize(), distanceWeightFactor );
    }
    /**
     * @param distanceWeightFactor - indicates the distance weightingFactor factor applied while comparing
     * AbstractGrid2DSquareCellDouble and inputGrid values ( default = 0.0d )
     */
    public final AbstractGrid2DSquareCellDouble compare( AbstractGrid2DSquareCellDoubleFactory gridFactory, AbstractGrid2DSquareCellDouble inputGrid, String comparator, double distance, double distanceWeightFactor ) {
        // @ToDo Add in checks here that in general there will be enough cells for the comparison
        // Initialisation
        int inputGridNrows = inputGrid.getNrows();
        int inputGridNcols = inputGrid.getNcols();
        double inputGridXllcorner = inputGrid.getXllcorner();
        double inputGridYllcorner = inputGrid.getYllcorner();
        double inputGridCellsize = inputGrid.getCellsize();
        double inputGridNoDataValue = inputGrid.getNoDataValue();
        int nrows = getNrows();
        int ncols = getNcols();
        double xllcorner = getXllcorner();
        double yllcorner = getYllcorner();
        double cellsize = getCellsize();
        double noDataValue = getNoDataValue();
        int thisCellDistance = ( int ) Math.ceil( distance / cellsize );
        int inputGridCellDistance = ( int ) Math.ceil( distance / inputGridCellsize );
        AbstractGrid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( nrows, ncols, xllcorner, yllcorner, cellsize, noDataValue );
        // Intersection check
        if ( ( inputGridXllcorner > xllcorner + ( ( double ) ncols * cellsize ) ) ||
        ( inputGridYllcorner > yllcorner + ( ( double ) nrows * cellsize ) ) ||
        ( inputGridXllcorner + ( ( double ) inputGridNcols * inputGridCellsize ) < xllcorner ) ||
        ( inputGridYllcorner + ( ( double ) inputGridNrows * inputGridCellsize ) < yllcorner ) ) {
            System.out.println( "No intersection! in compare( gridFactory(" + gridFactory + "), inputGrid(" + inputGrid + "), comparator(" + comparator + "), distance(" + distance + "), distanceWeightFactor(" + distanceWeightFactor + ") )" );
            return null;
        }
        
        // Calculation
        
        int cellID1;
        double thisDistance;
        double x0;
        double y0;
        double x1;
        double y1;
        double d1;
        double d2;
        double d3;
        double d4;
        double d5;
        double d6;
        
        // Perform difference operation
        if ( comparator.equalsIgnoreCase( "diff" ) ) {
            for ( int i = 0; i < nrows; i ++ ) {
                for ( int j = 0; j < ncols; j ++ ) {
                    cellID1 = getCellID( i, j );
                    x0 = getCellX( cellID1 );
                    y0 = getCellY( cellID1 );
                    d1 = 0.0d;
                    d2 = 0.0d;
                    for ( int p = - thisCellDistance; p <= thisCellDistance; p ++ ) {
                        for ( int q = - thisCellDistance; q <= thisCellDistance; q ++ ) {
                            d3 = getCell( ( i + p ), ( j + q ) );
                            if ( d3 != noDataValue ) {
                                x1 = getCellX( ( i + p ), ( j + q ) );
                                y1 = getCellY( ( i + p ), ( j + q ) );
                                thisDistance = distance( x0, y0, x1, y1 );
                                if ( thisDistance <= distance ) {
                                    if ( thisDistance != 0.0d ) {
                                        d1 += d3 / ( Math.pow( distance, distanceWeightFactor ) / Math.pow( thisDistance, distanceWeightFactor ) );
                                    } else {
                                        d1 += d3;
                                    }
                                    d2 += 1.0d;
                                }
                            }
                        }
                    }
                    d4 = 0.0d;
                    d5 = 0.0d;
                    d6 = 0.0d;
                    for ( int p = - inputGridCellDistance; p <= inputGridCellDistance; p ++ ) {
                        for ( int q = - inputGridCellDistance; q <= inputGridCellDistance; q ++ ) {
                            x1 = x0 + ( p * inputGridCellsize );
                            y1 = y0 + ( q * inputGridCellsize );
                            d6 = inputGrid.getCell( x1, y1 );
                            if ( d6 != noDataValue ) {
                                thisDistance = distance( x0, y0, x1, y1 );
                                if ( thisDistance <= distance ) {
                                    if ( thisDistance != 0.0d ) {
                                        d4 += d6 / ( Math.pow( distance, distanceWeightFactor ) / Math.pow( thisDistance, distanceWeightFactor ) );
                                    } else {
                                        d4 += d6;
                                    }
                                    d5 += 1.0d;
                                }
                            }
                        }
                    }
                    if ( d2 != 0.0d && d5 != 0.0d ) {
                        //result.setCell( i, j, d1 - d4 );
                        result.setCell( i, j, ( d1 / d2 ) - ( d4 / d5 ) );
                    }
                }
            }
        }
        
        // abs requires both this and the result to have the same spatial frame
        if ( comparator.equalsIgnoreCase( "abs" ) ) {
            AbstractGrid2DSquareCellDouble dummy = null;
            if ( ! ( inputGridCellsize == cellsize && ( ( inputGridXllcorner % cellsize ) == ( xllcorner % cellsize ) ) && ( ( inputGridYllcorner % cellsize ) == ( yllcorner % cellsize ) ) ) ) {
                dummy = gridFactory.createGrid2DSquareCellDouble( nrows, ncols, xllcorner, yllcorner, cellsize, noDataValue );
                dummy.addToGrid( inputGrid );
            } else {
                dummy = inputGrid;
            }
            int cellID2;
            for ( int i = 0; i < nrows; i ++ ) {
                for ( int j = 0; j < ncols; j ++ ) {
                    cellID1 = getCellID( i, j );
                    d1 = 0.0d;
                    d2 = 0.0d;
                    for ( int p = - thisCellDistance; p <= thisCellDistance; p ++ ) {
                        for ( int q = - thisCellDistance; q <= thisCellDistance; q ++ ) {
                            cellID2 = getCellID( i + p, j + q );
                            d3 = getCell( i + p, j + q );
                            d4 = dummy.getCell( i + p, j + q );
                            if ( d3 != noDataValue && d4 != noDataValue ) {
                                thisDistance = distance( cellID1, cellID2 );
                                if ( thisDistance <= distance ) {
                                    if ( thisDistance != 0.0d ) {
                                        d1 += Math.abs( d3 - d4 ) / ( Math.pow( distance, distanceWeightFactor ) / Math.pow( thisDistance, distanceWeightFactor ) );
                                    } else {
                                        d1 += d3;
                                    }
                                    d2 += 1.0d;
                                }
                            }
                        }
                    }
                    if ( d2 != 0.0d ) {
                        result.setCell( i, j, d1 / d2 );
                    }
                }
            }
        }
        
        // Calculates the pearsons correlation coefficient
        // Answers are the same as for momentcorrelation comparator!
        if ( comparator.equalsIgnoreCase( "pearsons" ) ) {
            double sumx = 0.0d;
            double sumy = 0.0d;
            double sumxsquared = 0.0d;
            double sumysquared = 0.0d;
            double sumxy = 0.0d;
            double weight;
            double sumWeight;
            int n;
            double numerator;
            double denominator;
            AbstractGrid2DSquareCellDouble dummy = null;
            if ( ! ( inputGridCellsize == cellsize && ( ( inputGridXllcorner % cellsize ) == ( xllcorner % cellsize ) ) && ( ( inputGridYllcorner % cellsize ) == ( yllcorner % cellsize ) ) ) ) {
                dummy = gridFactory.createGrid2DSquareCellDouble( nrows, ncols, xllcorner, yllcorner, cellsize, noDataValue );
                dummy.addToGrid( inputGrid );
            } else {
                dummy = inputGrid;
            }
            int cellID2;
            for ( int i = 0; i < nrows; i ++ ) {
                for ( int j = 0; j < ncols; j ++ ) {
                    cellID1 = getCellID( i, j );
                    sumx = 0.0d;
                    sumy = 0.0d;
                    sumxsquared = 0.0d;
                    sumysquared = 0.0d;
                    sumxy = 0.0d;
                    sumWeight = 0.0d;
                    n = 0;
                    for ( int p = - thisCellDistance; p <= thisCellDistance; p ++ ) {
                        for ( int q = - thisCellDistance; q <= thisCellDistance; q ++ ) {
                            if ( inGrid( i + p, j + q ) ) {
                                cellID2 = getCellID( i + p, j + q );
                                d3 = getCell( i + p, j + q );
                                d4 = dummy.getCell( i + p, j + q );
                                if ( ! ( d3 == noDataValue || d4 == noDataValue ) ) {
                                    thisDistance = distance( cellID1, cellID2 );
                                    if ( thisDistance <= distance ) {
                                        //if ( thisDistance != 0.0d ) {
                                        weight = Math.pow( 1.0d - ( Math.pow( thisDistance, distanceWeightFactor ) / Math.pow( distance, distanceWeightFactor ) ), distanceWeightFactor );
                                        //    weight = Math.pow( thisDistance, distanceWeightFactor ) / Math.pow( distance, distanceWeightFactor );
                                        //} else {
                                        //    weight = 1.0d;
                                        //}
                                        sumx += d3 * weight;
                                        sumy += d4 * weight;
                                        sumxsquared += Math.pow( d3, 2.0d ) * weight;
                                        sumysquared += Math.pow( d4, 2.0d ) * weight;
                                        sumxy += d3 * d4 * weight;
                                        sumWeight += weight;
                                        n ++;
                                    }
                                }
                            }
                        }
                    }
                    numerator = ( sumWeight * sumxy ) - ( sumx * sumy );
                    denominator = Math.sqrt( Math.abs( (  ( sumWeight * sumxsquared ) - Math.pow( sumx, 2.0d )  ) * ( ( sumWeight * sumysquared ) - Math.pow( sumy, 2.0d ) ) ) );
                    if ( denominator != 0.0d ) {
                        result.setCell( i, j, numerator / denominator );
                    }
                }
            }
        }
        
        // Calculates the moment correlation
        // Answers are the same as for pearsons comparator!
        if ( comparator.equalsIgnoreCase( "momentcorrelation" ) ) {System.out.println("momentumcorrelation");
        double meanx = 0.0d;
        double meany = 0.0d;
        double sdx = 0.0d;
        double sdy = 0.0d;
        double mxy = 0.0d;
        double weight;
        double sumWeight;
        double denominator;
        AbstractGrid2DSquareCellDouble dummy = null;
        if ( ! ( inputGridCellsize == cellsize && ( ( inputGridXllcorner % cellsize ) == ( xllcorner % cellsize ) ) && ( ( inputGridYllcorner % cellsize ) == ( yllcorner % cellsize ) ) ) ) {
            dummy = gridFactory.createGrid2DSquareCellDouble( nrows, ncols, xllcorner, yllcorner, cellsize, noDataValue );
            dummy.addToGrid( inputGrid );
        } else {
            dummy = inputGrid;
        }
        int cellID2;
        // Calculate
        for ( int i = 0; i < nrows; i ++ ) {
            for ( int j = 0; j < ncols; j ++ ) {
                cellID1 = getCellID( i, j );
                meanx = 0.0d;
                meany = 0.0d;
                sumWeight = 0.0d;
                // Calculate meanx and meany
                // NB. It may save time storing the IDs and weights for calculating the moments
                for ( int p = - thisCellDistance; p <= thisCellDistance; p ++ ) {
                    for ( int q = - thisCellDistance; q <= thisCellDistance; q ++ ) {
                        if ( inGrid( i + p, j + q ) ) {
                            cellID2 = getCellID( i + p, j + q );
                            d3 = getCell( i + p, j + q );
                            d4 = dummy.getCell( i + p, j + q );
                            if ( d3 != noDataValue && d4 != noDataValue ) {
                                thisDistance = distance( cellID1, cellID2 );
                                if ( thisDistance <= distance ) {
                                    // The following weight is just one example of a kernel weight that can be used!
                                    //if ( thisDistance != 0.0d ) {
                                    weight = Math.pow( 1.0d - ( Math.pow( thisDistance, distanceWeightFactor ) / Math.pow( distance, distanceWeightFactor ) ), distanceWeightFactor );
                                    //weight = Math.pow( thisDistance, distanceWeightFactor ) / Math.pow( distance, distanceWeightFactor );
                                    //} else {
                                    //    weight = 1.0d;
                                    //}
                                    meanx += d3 * weight;
                                    meany += d4 * weight;
                                    sumWeight += weight;
                                }
                            }
                        }
                    }
                }
                if ( sumWeight != 0.0d ) {
                    meanx = meanx / sumWeight;
                    meany = meany / sumWeight;
                }
                
                sdx = 0.0d;
                sdy = 0.0d;
                mxy = 0.0d;
                weight = 0.0d;
                sumWeight = 0.0d;
                // Calculate moments sdx, sdy and mxy
                for ( int p = - thisCellDistance; p <= thisCellDistance; p ++ ) {
                    for ( int q = - thisCellDistance; q <= thisCellDistance; q ++ ) {
                        if ( inGrid( i + p, j + q ) ) {
                            cellID2 = getCellID( i + p, j + q );
                            d3 = getCell( i + p, j + q );
                            d4 = dummy.getCell( i + p, j + q );
                            if ( d3 != noDataValue && d4 != noDataValue ) {
                                thisDistance = distance( cellID1, cellID2 );
                                if ( thisDistance <= distance ) {
                                    // Again the following weight is just one example of a kernel weight that can be used!
                                    //if ( thisDistance != 0.0d ) {
                                    weight = Math.pow( 1.0d - ( Math.pow( thisDistance, distanceWeightFactor ) / Math.pow( distance, distanceWeightFactor ) ), distanceWeightFactor );
                                    //weight = Math.pow( thisDistance, distanceWeightFactor ) / Math.pow( distance, distanceWeightFactor );
                                    //} else {
                                    //    weight = 1.0d;
                                    //}
                                    sdx += Math.pow( d3 - meanx, 2.0d ) * weight;
                                    sdy += Math.pow( d4 - meany, 2.0d ) * weight;
                                    mxy += ( d3 - meanx ) * ( d4 - meany ) * weight;
                                    sumWeight += weight;
                                }
                            }
                        }
                    }
                }
                if ( sumWeight != 0.0d ) {
                    sdx = sdx / sumWeight;
                    sdy = sdy / sumWeight;
                    mxy = mxy / sumWeight;
                }
                denominator = Math.sqrt( sdx ) * Math.sqrt( sdy );
                if ( denominator != 0.0d ) {
                    result.setCell( i, j, mxy / denominator );
                }
            }
        }
        }
        result.mask( this );
        result.mask( inputGrid );
        return result;
    }
    
    /**
     * Conversion methods
     */
    
    /**
     * Writes grid out to file in ArcInfo gridascii format and returns the File
     * to which it was written. The asciiFile is written with a 6 line header
     * followed by getNrows() lines containing the values of the grid written
     * out in column order. If the grid is already stored as a file then the
     * asciiFile is written out in the same directory else to the temporary
     * directory by default.
     */
    public File toAsciiFile() {
        String dataDirectory = System.getProperty( "java.io.tmp.dir" );
        if ( this instanceof Grid2DSquareCellDoubleFile ) {
            dataDirectory = ( ( Grid2DSquareCellDoubleFile ) this ).getDataDirectory();
        }
        return toAsciiFile( Utilities.makeTemporaryFile( dataDirectory ) );
    }
    public File toAsciiFile( File file ) {
        //System.out.println( "toAsciiFile( File(" + file.toString() + ") )" );
        try {
            PrintWriter pw = new PrintWriter( new BufferedWriter( new OutputStreamWriter( new FileOutputStream( file ) ) ) );
            // Write header info
            int nrows = getNrows();
            int ncols = getNcols();
            pw.println( "ncols " + getNcols() );
            pw.println( "nrows " + getNrows() );
            pw.println( "xllcorner " + getXllcorner() );
            pw.println( "yllcorner " + getYllcorner() );
            pw.println( "cellsize " + getCellsize() );
            pw.println( "noDataValue "+ getNoDataValue() );
            // Write values
            double d1;
            for ( int i = 0; i < nrows; i ++ ) {
                for ( int j = 0 ; j < ncols; j ++ ) {
                    d1 = getCell( i, j );
                    //if ( d1 == Double.NaN ) {
                    if ( new String("" + d1 ).equalsIgnoreCase( "NaN" ) ) {
                        pw.print( noDataValue + " " );
                        System.out.println( "Warning!!! NaN encountered at row " + i + ", column " + j + " set to noDataValue " + noDataValue + "." );
                    } else {
                        pw.print( d1 + " " );
                    }
                    //pw.print( getCell( i, j ) + " " );
                }
                pw.println( "" );
            }
            // Close output
            pw.flush();
            pw.close();
        } catch ( java.io.FileNotFoundException e ) {
            System.out.println( e.toString() + " in toAsciiFile( File(" + file + ") )" );
            System.exit( 0 );
        }
        return file;
    }
    
    /**
     * Writes grid cell values in row major order as binary doubles to file using a
     * RandomAccessFile and returns the RandomAccessFile. If append is true then the
     * values are appended to the file. If append is false then the values are
     * written from the start of the file after which nrows, ncols, xllcorner,
     * yllcorner, cellsize, noDataValue, upToDate, sparseness, sum, min, minCount,
     * max, maxCount are written and the file length is set.
     */
    public RandomAccessFile toBinaryFile() {
        String dataDirectory = System.getProperty( "java.io.tmp.dir" );
        if ( this instanceof Grid2DSquareCellDoubleFile ) {
            dataDirectory = ( ( Grid2DSquareCellDoubleFile ) this ).getDataDirectory();
        }
        return toBinaryFile( Utilities.makeTemporaryFile( dataDirectory ) );
    }
    public RandomAccessFile toBinaryFile( File file ) {
        return toBinaryFile( file, false );
    }
    public RandomAccessFile toBinaryFile( File file, boolean append ) {
        return toBinaryFile( file, append, 0, 0, getNrows() - 1, getNcols() - 1 );
    }
    public RandomAccessFile toBinaryFile( File file, boolean append, int startRow, int startCol, int endRow, int endCol ) {
        RandomAccessFile raf = null;
        try {
            raf = new RandomAccessFile( file, "rw" );
            if ( this instanceof Grid2DSquareCellDoubleFile && startRow == 0 && startCol == 0 && endRow == ( getNrows() - 1 ) && endCol == ( getNcols() - 1 ) ) {
                File gridFile = ( ( Grid2DSquareCellDoubleFile ) this ).getFile();
                String os = new String( System.getProperty( "os.name" ) );
                if ( os.equalsIgnoreCase( "SunOS" ) ) {
                    Runtime runtime = java.lang.Runtime.getRuntime();
                    runtime.exec( "cp " + gridFile.toString() + " " + file.toString() );
                    // Failed optimsiation for Windows XP OS
                    //} else if ( os.equalsIgnoreCase( "Windows XP" ) ) {
                    //    Runtime runtime = java.lang.Runtime.getRuntime();
                    //    runtime.exec( "C:/WINDOWS/System32/cmd.exe copy d:/" + gridFile.toString() + " /B d:/" + file.toString() + " /B " );
                } else {
                    //System.out.println( os );
                    BufferedInputStream bis = new BufferedInputStream( new FileInputStream( gridFile ) );
                    BufferedOutputStream bos = new BufferedOutputStream( new FileOutputStream( file ) );
                    for ( int i = 0; i < gridFile.length(); i ++ ) {
                        bos.write( bis.read() );
                    }
                    bos.flush();
                    bos.close();
                    bis.close();
                }
            } else {
                return toBinaryFile( raf, append, startRow, startCol, endRow, endCol );
            }
        } catch ( java.io.IOException e ) {
            System.out.println( e.toString() + " in toBinaryFile( File(" + file + "), startRow(" + startRow + "), startCol(" + startCol + "), endRow(" + endRow + "), endCol(" + endCol + ") )" );
            System.exit( 0 );
        }
        return raf;
    }
    public RandomAccessFile toBinaryFile( RandomAccessFile raf ) {
        return toBinaryFile( raf, false );
    }
    public RandomAccessFile toBinaryFile( RandomAccessFile raf, boolean append ) {
        return toBinaryFile( raf, append, 0, 0, getNrows() - 1, getNcols() - 1 );
    }
    public RandomAccessFile toBinaryFile( RandomAccessFile raf, boolean append, int startRow, int startCol, int endRow, int endCol ) {
        //System.out.println( "toBinaryFile( RandomAccessFile(" + raf.toString() + "), append(" + append + "), startRow(" + startRow + "), startCol(" + startCol + "), endRow(" + endRow + "), endCol(" + endCol + ") )" );
        AbstractGridStatistics stats = this.getGridStatistics();
        try {
            if ( append ) {
                //raf.seek( raf.length() - 13 * 8 );
                raf.seek( raf.length() );
            } else {
                raf.seek( 0 );
            }
            for ( int i = startRow; i <= endRow; i ++ ) {
                for ( int j = startCol; j <= endCol; j ++ ) {
                    raf.writeDouble( getCell( i, j ) );
                }
            }
            if ( ! append ) {
                int nrows = endRow - startRow + 1;
                int ncols = endCol - startCol + 1;
                double cellsize = getCellsize();
                raf.writeDouble( ( double ) nrows );
                raf.writeDouble( ( double ) ncols );
                raf.writeDouble( getCellX( startCol ) - ( cellsize / 2 ) );
                raf.writeDouble( getCellY( ( endRow * ncols ) + 1 ) - ( cellsize / 2 ) );
                raf.writeDouble( cellsize );
                raf.writeDouble( getNoDataValue() );
                double[] gridStats = stats.toArray();
                for ( int i = 0; i < gridStats.length; i ++ ) {
                    raf.writeDouble( gridStats[ i ] );
                }
            }
        } catch ( java.io.IOException e ) {
            System.out.println( e.toString() + " in toBinaryFile( RandomAccessFile( " + raf.toString() + "), append(" + append + "), startRow(" + startRow + "), startCol(" + startCol + "), endRow(" + endRow + "), endCol(" + endCol + ") )" );
            System.exit( 0 );
        }
        return raf;
    }
    
    /**
     * Returns a uk.ac.leeds.ccg.raster.Raster constructed from an array of grid
     * cell values.
     */
    public Raster toRaster() {
        int nrows = getNrows();
        int ncols = getNcols();
        double cellsize = getCellsize();
        GeoRectangle gr1 = new GeoRectangle( getXllcorner(), getYllcorner(), ( double ) ( ncols * cellsize ), ( double ) ( nrows * cellsize ) );
        Raster r1 = new Raster( gr1, cellsize );
        try {
            double[] data = new double[ nrows * ncols ];
            for ( int i = 0; i < nrows * ncols; i ++ ) {
                data[ i ] = getCell( i );
            }
            //r1.setNoDataValue(getNoDataValue());
            r1.setMissingValueCode( getNoDataValue() );
            r1.setData( data );
        } catch ( OutOfMemoryError e ) {
            System.out.println( "" + e + " in toRaster()" );
            System.exit( 0 );
        }
        return r1;
    }
    
    /**
     * Writes this grid as a Grey scale image
     * @param file - the image file that will be written
     * @param type - the name of the type of image to be written e.g. "png", "jpeg"
     */
    public void toImage( File file, String type ) {
        // Test what writers are available as this may vary on different systems!
        boolean writerAvailable = Utilities.isImageWriterAvailable( type );
        if ( ! writerAvailable ) {
            System.out.println( "Warning!!! Failed to write grid as " + type + " to File(" + file.toString() + ") as writer is unavailable." );
        } else {
            double noDataValue = getNoDataValue();
            int ncols = getNcols();
            int nrows = getNrows();
            AbstractGridStatistics stats = getGridStatistics();
            double max = stats.getMax();
            double min = stats.getMin();
            // Read all data into a array and scale all value into range 0 - 255
            try {
                double d1;
                int i1 = 0;
                int size = ncols * nrows;
                //System.out.println( "Scale grid values and do byte operation into an RGB int array" );
                int[] gridImageArray = new int[ size ];
                for ( int i = 0; i < size; i ++ ) {
                    d1 = getCell( i );
                    // Construct an RGB integer by byte operation : 32 bytes, first 8 bytes is transparency value
                    // second, third, forth 8 byte is Red, green, blue value, which will be used by MemoryImageSource
                    // class.
                    i1 = 255;
                    // Set noDataValue as blue
                    if (  d1 == noDataValue ) {
                        gridImageArray[ i ] = ( 255 << 24 ) | ( 0 << 16 ) | ( 0 << 8 ) | i1;
                    } else {
                        if ( max != min ) {
                            i1 = new Double( 255.0d * ( d1 - min ) / ( max - min ) ).intValue();
                            gridImageArray[ i ] = ( 255 << 24 ) | ( i1 << 16 ) | ( i1 << 8 ) | i1;
                        } else {
                            // Set white
                            gridImageArray[ i ] = ( 255 << 24 ) | ( i1 << 16 ) | ( i1 << 8 ) | i1;
                        }
                    }
                }
                // Use gridImageArray to create a MemoryImageSource
                // Construct a BufferedImage using Toolkit
                // because Image class does not implement the RenderedImage
                // interface and imageio needs it.
                MemoryImageSource gridImageSource = new MemoryImageSource( ncols, nrows, gridImageArray, 0, ncols );
                Image tempImage = Toolkit.getDefaultToolkit().createImage( gridImageSource );
                BufferedImage gridImage = new BufferedImage( ncols, nrows, BufferedImage.TYPE_INT_RGB );
                Graphics2D g = ( Graphics2D ) gridImage.getGraphics();
                g.drawImage( tempImage, 0, 0, new java.awt.Panel() );
                try {
                    javax.imageio.ImageIO.write( gridImage, type, file );
                } catch( java.io.IOException e1 ) {
                    e1.printStackTrace();
                    System.out.println( "Warning!!! Failed to write grid as " + type + " to File(" + file.toString() + ")" );
                }
                g.dispose();
                gridImage.flush();
                tempImage.flush();
            } catch ( java.lang.OutOfMemoryError e2 ) { // Try doing in tiles!
                e2.printStackTrace();
                System.out.println( "Warning!!! OutOfMemory attempting to write grid as " + type + " to File(" + file.toString() + ")" );
            }
        }
    }
    
}