package uk.ac.leeds.ccg.grids;
/**
 * -------------------------------------------------------------------------
 * This file is distributed as part of the uk.ac.leeds.ccg.grids library.
 * This library is a resource for manipulating spatial data stored as grids.
 * Copyright (C) 2002  Andy Turner
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 */
import uk.ac.leeds.ccg.raster.Raster;
import uk.ac.leeds.ccg.geotools.GeoRectangle;
import java.io.File;
import java.io.RandomAccessFile;
import java.io.StreamTokenizer;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.FileInputStream;
/**
 * ---------------------------
 * Last updated on 25 Nov 2002
 * @author andyt
 * @version 1.0
 * ---------------------------
 * Description:
 * This class holds methods for generating 2D square cell grids containing double values.
 *
 * Known extentions:
 * 1. Grid2DSquareCellDoubleFactory
 * 2. Grid2DSquareCellDoubleFileFactory
 */
public abstract class AbstractGrid2DSquareCellDoubleFactory {
    
    /**
     * ------
     * Fields
     * ------
     */
    
    /**
     * For storing what the constructed grids will have as noDataValue
     * The default is Double.NaN
     */
    double noDataValue = Double.NaN;

    /**
     * ------------------------------
     * Get and set methods for fields
     * ------------------------------
     */
    
    /**
     * Returns noDataValue
     */
    public double getNoDataValue() {
        return noDataValue;
    }

    /**
     * Sets noDataValue
     */
    public void setNoDataValue( double value ) {
        noDataValue = value;
    }

    /**
     * -------------------------------------------------------------------
     * Constructer methods to construct from a specification of dimensions
     * -------------------------------------------------------------------
     */
    
    /**
     * For no particular reason default nrows to 64 and ncols to 64
     */
    public AbstractGrid2DSquareCellDouble createGrid2DSquareCellDouble() {
        return createGrid2DSquareCellDouble( 64, 64 );
    }
    /**
     * Default xllcorner and yllcorner to 0.0, cellsize to 1.0
     */
    public AbstractGrid2DSquareCellDouble createGrid2DSquareCellDouble( int nrows, int ncols ) {
        return createGrid2DSquareCellDouble( nrows, ncols, 0.0d, 0.0d, 1.0d, noDataValue );
    }
    /**
     * Default gridStatisticsType to 0
     */
    public AbstractGrid2DSquareCellDouble createGrid2DSquareCellDouble( int nrows, int ncols, double xllcorner, double yllcorner, double cellsize, double noDataValue ) {
        int gridStatisticsType = 0;
        return createGrid2DSquareCellDouble( nrows, ncols, xllcorner, yllcorner, cellsize, noDataValue, gridStatisticsType );
    }
    public abstract AbstractGrid2DSquareCellDouble createGrid2DSquareCellDouble( int nrows, int ncols, double xllcorner, double yllcorner, double cellsize, double noDataValue, int gridStatisticsType );
    
    /**
     * -----------------------------------------------------------------------
     * Constructer methods to construct from an AbstractGrid2DSquareCellDouble
     * -----------------------------------------------------------------------
     */
    
    /**
     * Default StartRowIndex and StartColIndex to 0, endRowIndex to nrows - 1, endColIndex to ncols - 1
     */
    public AbstractGrid2DSquareCellDouble createGrid2DSquareCellDouble( AbstractGrid2DSquareCellDouble grid ) {
        return createGrid2DSquareCellDouble( grid, 0, 0, grid.getNrows() - 1, grid.getNcols() - 1 );
    }
    /**
     * Default gridStatisticsType to 0
     */
    public AbstractGrid2DSquareCellDouble createGrid2DSquareCellDouble( AbstractGrid2DSquareCellDouble grid, int StartRowIndex, int StartColIndex, int EndRowIndex, int EndColIndex ) {
        return createGrid2DSquareCellDouble( grid, StartRowIndex, StartColIndex, EndRowIndex, EndColIndex, 0 );
    }
    /**
     * Default gridStatisticsType to 0
     * @param extent - defines the extent of the constructed grid
     */
    public AbstractGrid2DSquareCellDouble createGrid2DSquareCellDouble( AbstractGrid2DSquareCellDouble grid, GeoRectangle extent ) {
        return createGrid2DSquareCellDouble( grid, extent, 0 );
    }
    public AbstractGrid2DSquareCellDouble createGrid2DSquareCellDouble( AbstractGrid2DSquareCellDouble grid, GeoRectangle extent, int gridStatisticsType ) {
        double x = extent.getX();
        double y = extent.getY();
        return createGrid2DSquareCellDouble( grid, grid.getRowIndex( y ) , grid.getColIndex( x ), grid.getRowIndex( y + extent.getHeight() ), grid.getColIndex( x + extent.getWidth() ), gridStatisticsType );
    }
    public abstract AbstractGrid2DSquareCellDouble createGrid2DSquareCellDouble( AbstractGrid2DSquareCellDouble grid, int StartRowIndex, int StartColIndex, int EndRowIndex, int EndColIndex, int gridStatisticsType );
    
    /**
     * ------------------------------------------------------
     * Constructer methods to construct from a File
     * The type of file is currently recognised by its suffix
     * ------------------------------------------------------
     */
    
    /**
     * Default gridStatisticsType to 0
     */
    public AbstractGrid2DSquareCellDouble createGrid2DSquareCellDouble( File gridFile ) {
        return createGrid2DSquareCellDouble( gridFile, 0 );
    }
    /**
     * Default startRowIndex and startColIndex to 0, endRowIndex to nrows - 1, endColIndex to ncols - 1
     */
    public AbstractGrid2DSquareCellDouble createGrid2DSquareCellDouble( File gridFile, int gridStatisticsType ) {
        String filename = gridFile.getName();
        int nrows = 64;
        int ncols = 64;
        if ( filename.endsWith( "asc" ) ) {
            try {
                InputStreamReader isr = new InputStreamReader( new FileInputStream( gridFile ) ); 
                StreamTokenizer st = new StreamTokenizer( new BufferedReader( isr ) );
                st.parseNumbers();
                st.eolIsSignificant( false );
                int type = st.nextToken();
                st.nextToken();
                ncols = ( int ) st.nval;
                st.nextToken();
                st.nextToken();
                nrows = ( int ) st.nval;
                isr.close();
            } catch ( java.io.IOException e ) {
                System.out.println( e.toString() + "in createGrid2DSquareCellDouble( File(" + gridFile.toString() + ") )" );
                System.exit( 0 );
            }
            return createGrid2DSquareCellDouble( gridFile, 0, 0, nrows - 1, ncols - 1, gridStatisticsType );
        }
        if ( filename.endsWith( "grd" ) ) {
            try {
                RandomAccessFile raf = null;
                raf = new RandomAccessFile( gridFile, "r" );
                raf.seek( raf.length() - ( 13 * 8 ) );
                nrows = ( int ) raf.readDouble();
                ncols = ( int ) raf.readDouble();
                raf.close();
            } catch ( java.io.IOException e ) {
                System.out.println( e.toString() + "in createGrid2DSquareCellDouble( File(" + gridFile.toString() + ") )" );
                System.exit( 0 );
            }
            return createGrid2DSquareCellDouble( gridFile, 0, 0, nrows - 1, ncols - 1, gridStatisticsType );
        }
        return null;
    }
    public abstract AbstractGrid2DSquareCellDouble createGrid2DSquareCellDouble( File gridFile, int startRowIndex, int startColIndex, int endRowIndex, int endColIndex, int gridStatisticsType );
    
    /**
     * -------------------------------------------------------
     * Constructer methods to construct from a geotools Raster
     * -------------------------------------------------------
     */
    
    /**
     * Default gridStatisticsType to 0
     */
    public AbstractGrid2DSquareCellDouble createGrid2DSquareCellDouble( Raster raster ) {
        return createGrid2DSquareCellDouble( raster , 0 );
    }
    public abstract AbstractGrid2DSquareCellDouble createGrid2DSquareCellDouble( Raster raster, int gridStatisticsType );
    
    /**
     * ------------------------
     * Required utility methods
     * ------------------------
     */
    public abstract void setFile( File file );
    public abstract String getDataDirectory();
    
}
