package uk.ac.leeds.ccg.grids;
/**
 * -------------------------------------------------------------------------
 * This file is distributed as part of the uk.ac.leeds.ccg.grids library.
 * This library is a resource for manipulating spatial data stored as grids.
 * Copyright (C) 2002  Andy Turner
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 */
/**
 * ---------------------------
 * Last updated on 25 Nov 2002
 * @author andyt
 * @version 1.0
 * ---------------------------
 * Description:
 * Instatiated extentions of this are attached to an instance of an extended 
 * AbstractGrid2DSquareCellDouble instance
 *
 * Known extentions:
 * 1. GridStatistics0
 * 2. GridStatistics1
 */

public abstract class AbstractGridStatistics {
    
    /**
     * ------
     * Fields
     * ------
     */

    /**
     * Reference to the AbstractGrid2DSquareCellDouble that holds a reference to this
     */
    private AbstractGrid2DSquareCellDouble grid;
    
    /**
     * Stores the proportion of cells in grid with values equal to the grid noDataValue.
     * sparseness = 1.0d means all values are noDataValues
     * sparseness = 0.0d means all values are not noDataValues
     */
    private double sparseness;
    
    /**
     * For storing the sum of all values in the Grid
     */
    protected double sum;
    
    /**
     * For storing the minimum value in the Grid
     */
    protected double min;
    /**
     * Stores the number of minimum values in the Grid
     */
    protected int minCount;
    
    /**
     * For storing the maximum value in the Grid
     */
    protected double max;
    /**
     * Stores the number of maximum values in the Grid
     */
    protected int maxCount;
    
    /**
     * ---------------
     * General methods
     * ---------------
     */

    /**
     * Initialisation method
     */
    protected void init( AbstractGrid2DSquareCellDouble grid ) {
        setGrid( grid );
        setSparseness( 1.0d );
        this.max = Double.MIN_VALUE;
        this.maxCount = 0;
        this.min = Double.MAX_VALUE;
        this.minCount = 0;
        this.sum = 0.0d;
    }
    
    /**
     * Returns grid
     */
    public AbstractGrid2DSquareCellDouble getGrid() {
        return this.grid;
    }
    
    /**
     * Sets grid
     */
    protected void setGrid( AbstractGrid2DSquareCellDouble grid ) {
        this.grid = grid;
    }
    
    /**
     * Returns the type of gridStatistics:
     * 0 - GridStatistics0
     * 1 - GridStatistics1
     */
    public abstract int getGridStatisticsType();
    
    /**
     * Returns true iff all grid statistics are up-to-date.
     */
    public abstract boolean getIsUpToDate();
    
    /**
     * Sets upToDate flags for statistics
     */
    protected abstract void setIsUpToDate( boolean value );
    
    /**
     * Returns sparseness
     */
    public double getSparseness() {
        return this.sparseness;
    }
    
    /**
     * Sets sparseness
     */
    protected void setSparseness( double sparseness ) {
        this.sparseness = sparseness;
    }
    
    /**
     * Returns the sum of all values in the grid
     */
    public double getSum() {
        if ( ! getIsUpToDate() ) {
            doStats();
        }
        return this.sum;
    }
    
    /**
     * Sets sum to value or not
     */
    protected abstract void setSum( double value );
    
    /**
     * Returns the minimum value in the Grid
     */
    public abstract double getMin();
    
    /**
     * Sets min to value or not
     */
    protected abstract void setMin( double value );
    
    /**
     * Returns a count of the number of cells in the grid with the minimum value
     */
    public abstract int getMinCount();
    
    /**
     * Sets minCount to value or not
     */
    protected abstract void setMinCount( int value );
    
    /**
     * Sets min to the minimum of its current value and the value passed in and modifies minCount or not
     */
    protected abstract void doMin( double value );
    
    /**
     * Returns the maximum value in the grid
     */
    public abstract double getMax();
    
    /**
     * Sets max to value or not
     */
    protected abstract void setMax( double value );
    
    /**
     * Returns a count of the number of cells in the grid with the maximum value
     */
    public abstract int getMaxCount();
    
    /**
     * Sets maxCount to value or not
     */
    protected abstract void setMaxCount( int value );
    
    /**
     * Sets max to the maximum of its current value and the value passed in and modifies maxCount or not
     */
    protected abstract void doMax( double d1 );
    
    /**
     * Updates statistics from newStats they are up-to-date else
     * sets isUpToDate to false.
     */
    protected void updateGridStatisticsFrom( AbstractGridStatistics newStats ) {
        boolean isUpToDate = newStats.getIsUpToDate();
        setSparseness( newStats.getSparseness() );
        if ( isUpToDate ) {
            this.sum = newStats.getSum();
            this.min = newStats.getMin();
            this.minCount = newStats.getMinCount();
            this.max = newStats.getMax();
            this.maxCount = newStats.getMaxCount();
        }
        setIsUpToDate( isUpToDate );
    }
    
    /**
     * Recalculates sparseness, sum, min, minCount, recalculateMin, max, maxCount and
     * recalculateMax if this is not up-to-date
     */
    protected void doStats() {
        if ( ! getIsUpToDate() ) {
            AbstractGrid2DSquareCellDouble grid = getGrid();
            int nrows = grid.getNrows();
            int ncols = grid.getNcols();
            double sparseness = 1.0d;
            double cellAsProportionOfGrid = 1.0d / ( double ) ( nrows * ncols );
            double sum = 0.0d;
            double max = Double.NEGATIVE_INFINITY;
            int maxCount = nrows * ncols;
            double min = Double.POSITIVE_INFINITY;
            int minCount = maxCount;
            double noDataValue = grid.getNoDataValue();
            double d1;
            for ( int i = 0; i < nrows; i ++ ) {
                for ( int j = 0; j < ncols; j ++ ) {
                    d1 = grid.getCell( i, j );
                    if ( d1 != noDataValue ) {
                        sparseness -= cellAsProportionOfGrid;
                        sum += d1;
                        if ( d1 > max ) { maxCount = 1; max = d1; }
                        if ( d1 == max ) { maxCount ++; }
                        if ( d1 < min ) { minCount = 1;min = d1; }
                        if ( d1 == min ) { minCount ++; }
                    }
                }
            }
            this.sparseness = sparseness;
            this.sum = sum;
            this.max = max;
            this.maxCount = maxCount;
            this.min = min;
            this.minCount = minCount;
            setIsUpToDate( true );
        }
    }
        
    /**
     * -----------------
     * Converter methods
     * -----------------
     */
    
    /**
     * Returns a description of the instance
     */
    public String toString() {
        String result;
        if ( this instanceof GridStatistics0 ) {
            GridStatistics0 dummy = ( GridStatistics0 ) this;
            result = "GridStatistics0( isUpToDate( " + getIsUpToDate() + " ), recalculateMin( " + dummy.getRecalculateMin() + " ), recalculateMax( " + dummy.getRecalculateMax() + " )";
        } else {
            result = "GridStatistics1( isUpToDate( " + getIsUpToDate() + " )";
        }
        return result + ", sparseness( " + getSparseness() + " ), sum( " + this.sum + " ), min( " + this.min + " ), minCount( " + this.minCount + " ), max( " + this.max + " ), maxCount( " + this.maxCount + " ) )";
    }
    
    /**
     * Returns stats as a double[]
     */
    protected double[] toArray( ) {
        double [] result = new double[ 7 ];
        if ( getIsUpToDate() ) {
            result[ 0 ] = 1.0d;
        } else {
            result[ 0 ] = 0.0d;
        }
        result[ 1 ] = getSparseness();
        result[ 2 ] = this.sum;
        result[ 3 ] = this.min;
        result[ 4 ] = this.minCount;
        result[ 5 ] = this.max;
        result[ 6 ] = this.maxCount;
        return result;
    }
    
}
