package uk.ac.leeds.ccg.grids;
/**
 * -------------------------------------------------------------------------
 * This file is distributed as part of the uk.ac.leeds.ccg.grids library.
 * This library is a resource for manipulating spatial data stored as grids.
 * Copyright (C) 2002  Andy Turner
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 */
import uk.ac.leeds.ccg.geotools.GeoRectangle;
import uk.ac.leeds.ccg.raster.Raster;
import java.io.File;
import java.io.RandomAccessFile;
import java.io.StreamTokenizer;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.FileInputStream;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Arrays;
/**
 * ---------------------------
 * Last updated on 25 Nov 2002
 * @author andyt
 * @version 1.0
 * ---------------------------
 * Description:
 * Class for storing and manipulating an AbstractGrid2DSquareCellDouble grid in memory.
 */
public class Grid2DSquareCellDouble extends AbstractGrid2DSquareCellDouble {
    
    /**
     * A double array for storing all the grid values in row major order.
     */
    private double[] gridArray;
    
    /**
     * A hashtable for storing grid values with their ID as the key. The ID is the
     * row order position in the grid as an Integer.
     */
    private Hashtable gridHashtable;
    
    /**
     * For indicating if the grid is stored as a Hashtable or a double[].
     */
    private boolean isHashtable = true;
    
    /**
     * Creates new Grid2DSquareCellDouble grid
     * @param nrows - The number of rows of cells in the constructed grid
     * @param ncols - The number of columns of cells in the constructed grid
     * @param xllcorner - The smallest (left-most) x-coordinate of the grid
     * @param yllcorner - The smallest (bottom-most) x-coordinate of the grid
     * @param cellsize - The cellsize of the grid
     * @param noDataValue - The noDataValue of the grid
     * @param gridStatistics1_0 -
     */
    protected Grid2DSquareCellDouble( int nrows, int ncols, double xllcorner, double yllcorner, double cellsize, double noDataValue, int gridStatisticsType ) {
        //System.out.println( "Grid2DSquareCellDouble( nrows( " + nrows + " ), ncols( " + ncols + " ), xllcorner( " + xllcorner + " ), yllcorner( " + yllcorner + " ), cellsize( " + cellsize + " ), noDataValue( " + noDataValue + " ), gridStatistics1_0( " + gridStatistics1_0 + " ) ) {" );
        initNrows( nrows );
        initNcols( ncols );
        setXllcorner( xllcorner );
        setYllcorner( yllcorner );
        setCellsize( cellsize );
        initNoDataValue( noDataValue );
        setGridHashtable( new Hashtable() );
        initGridStatistics( gridStatisticsType );
        //System.out.println( toString() );
    }
    
    /**
     * Constructs from an AbstractGrid2DSquareCellDouble.
     * @param inputGrid - the inputGrid AbstractGrid2DSquareCellDouble
     * The constructed grid has endColIndex-startColIndex + 1 columns and
     * endRowIndex-startRowIndex + 1 rows.
     */
    protected Grid2DSquareCellDouble( AbstractGrid2DSquareCellDouble inputGrid, int startRowIndex, int startColIndex, int endRowIndex, int endColIndex, int gridStatisticsType ) {
        //System.out.println( "Grid2DSquareCellDouble( AbstractGrid2DSquareCellDouble( " + inputGrid.toString() + " ), startRowIndex( " + startRowIndex + " ), startColIndex( " + startColIndex + " ), endRowIndex( " + endRowIndex + " ), endColIndex( " + endColIndex + " ), gridStatistics1_0( " + gridStatistics1_0 + " ) ) {" );
        int inputGridNrows = inputGrid.getNrows();
        int inputGridNcols = inputGrid.getNcols();
        double cellsize = inputGrid.getCellsize();
        int nrows = ( endRowIndex - startRowIndex + 1 );
        int ncols = ( endColIndex - startColIndex + 1 );
        initNrows( nrows );
        initNcols( ncols );
        setXllcorner( inputGrid.getXllcorner() + ( double ) startColIndex * cellsize );           
        //setXllcorner( inputGrid.getCellX( startColIndex ) - ( cellsize / 2.0d ) );
        setYllcorner( inputGrid.getYllcorner() + ( double ) ( inputGridNrows - endRowIndex - 1 ) * cellsize );
        setCellsize( cellsize );
        double noDataValue = inputGrid.getNoDataValue();
        initNoDataValue( noDataValue );
        AbstractGridStatistics inputGridStats = inputGrid.getGridStatistics();
        // Initialise grid stats
        initGridStatistics( gridStatisticsType );
        double sparseness = 1.0d;
        double cellAsProportionOfGrid = 1.0d / ( double ) ( nrows * ncols );
        double inputGridSparseness = inputGridStats.getSparseness();
        if ( inputGrid instanceof Grid2DSquareCellDouble ) {
            Grid2DSquareCellDouble inGrid = ( Grid2DSquareCellDouble ) inputGrid;
            boolean inGridIsHashtable = inGrid.getIsHashtable();
            if ( startRowIndex <= 0 && startColIndex <= 0 && endRowIndex >= ( inputGridNrows - 1 ) && endColIndex >= ( inputGridNcols - 1 ) ) {
                sparseness = ( ( inputGridSparseness * inputGridNrows * inputGridNcols ) + ( ( nrows * ncols ) - ( inputGridNrows * inputGridNcols ) ) ) / ( nrows * ncols );
                AbstractGridStatistics stats = getGridStatistics();
                stats.updateGridStatisticsFrom( inputGridStats );
                stats.setSparseness( sparseness );
                if ( inGridIsHashtable ) {
                    if ( startRowIndex == 0 && startColIndex == 0 && endRowIndex == ( inputGridNrows - 1 ) && endColIndex == ( inputGridNcols - 1 ) ) {
                        setGridHashtable( ( Hashtable ) inGrid.getGridHashtable().clone() );
                    } else {
                        Hashtable inputGridHashtable = inGrid.getGridHashtable();
                        int size = inputGridHashtable.size();
                        Hashtable gridHashtable = new Hashtable( size );
                        Enumeration inputGridKeys = inputGridHashtable.keys();
                        Enumeration values = inputGridHashtable.elements();
                        int inGridID;
                        for ( int i = 0; i < size; i ++ ) {
                            inGridID = ( ( Integer ) inputGridKeys.nextElement() ).intValue();
                            gridHashtable.put( new Integer( ( ( inGrid.getRowIndex( inGridID ) - startRowIndex ) * ncols ) + ( inGrid.getColIndex( inGridID ) - startColIndex ) ), values.nextElement() );
                        }
                        setGridHashtable( gridHashtable );
                    }
                } else {
                    if ( startRowIndex == 0 && startColIndex == 0 && endRowIndex == ( inputGridNrows - 1 ) && endColIndex == ( inputGridNcols - 1 ) ) {
                        setGridArray( ( double[] ) inGrid.getGridArray().clone() );
                    } else {
                        // 0.8d is hardcoded but is an estimated not a calculated optimistation!
                        // see optimiseCollection( "memory" ) method
                        if ( sparseness > 0.8d ) {
                            inGrid.toHashtable();
                            Hashtable inputGridHashtable = inGrid.getGridHashtable();
                            int size = inputGridHashtable.size();
                            Hashtable gridHashtable = new Hashtable( size );
                            Enumeration inputGridKeys = inputGridHashtable.keys();
                            Enumeration values = inputGridHashtable.elements();
                            int inGridID;
                            for ( int i = 0; i < size; i ++ ) {
                                inGridID = ( ( Integer ) inputGridKeys.nextElement() ).intValue();
                                gridHashtable.put( new Integer( ( ( inGrid.getRowIndex( inGridID ) - startRowIndex ) * ncols ) + ( inGrid.getColIndex( inGridID ) - startColIndex ) ), values.nextElement() );
                            }
                            setGridHashtable( gridHashtable );
                        } else {
                            double[] inputGridArray = inGrid.getGridArray();
                            int gridArrayLength = nrows * ncols;
                            double[] gridArray = new double[gridArrayLength];
                            Arrays.fill( gridArray, 0, gridArrayLength, noDataValue );
                            for ( int i = 0; i < inputGridNrows; i ++ ) {
                                System.arraycopy( inputGridArray, i * inputGridNcols, gridArray, ( ( ( i - startRowIndex )*ncols ) + ( 0 - startColIndex ) ), inputGridNcols );
                            }
                            setGridArray( gridArray );
                        }
                    }
                }
            } else {
                if ( inGridIsHashtable ) {
                    Hashtable inputGridHashtable = inGrid.getGridHashtable();
                    int size = inputGridHashtable.size();
                    setGridHashtable( new Hashtable() );
                    Enumeration inputGridKeys = inputGridHashtable.keys();
                    Enumeration values = inputGridHashtable.elements();
                    int inGridID;
                    int inGridRow;
                    int inGridCol;
                    Double D1;
                    double d1;
                    for ( int i = 0; i < size; i ++ ) {
                        inGridID = ( ( Integer ) inputGridKeys.nextElement() ).intValue();
                        inGridRow = inGrid.getRowIndex( inGridID );
                        inGridCol = inGrid.getColIndex( inGridID );
                        if ( inGridRow >= startRowIndex && inGridRow <= endRowIndex && inGridCol >= startColIndex && inGridCol <= endColIndex ) {
                            D1 = ( Double ) values.nextElement();
                            d1 = D1.doubleValue();
                            initCell( ( ( inGridRow - startRowIndex ) * ncols ) + ( inGridCol - startColIndex ), d1 );
                            sparseness -= cellAsProportionOfGrid;
                        } else {
                            values.nextElement();
                        }
                    }
                } else {
                    setGridHashtable( new Hashtable() );
                    double d1;
                    for ( int i = 0; i < inputGridNrows; i ++ ) {
                        for ( int j = 0; j < inputGridNcols; j ++ ) {
                            if ( i >= startRowIndex && i <= endRowIndex && j >= startColIndex && j <= endColIndex ) {
                                d1 = inGrid.getCell( i, j );
                                if ( d1 != noDataValue ) {
                                    initCell( ( ( i - startRowIndex ) * nrows ) + ( j - startColIndex ), d1 );
                                    sparseness -= cellAsProportionOfGrid;
                                }
                            }
                        }
                    }
                }
            }
        } else {
            setGridHashtable( new Hashtable() );
            double d1;
            int i1 = Math.max( startRowIndex, 0 );
            int i2 = Math.min( endRowIndex, ( inputGridNrows - 1 ) );
            int j1 = Math.max( startColIndex, 0 );
            int j2 = Math.min( endColIndex, ( inputGridNcols - 1 ) );
            for ( int i = i1; i <= i2; i ++ ) {
                for ( int j = j1; j <= j2; j ++ ) {
                    d1 = inputGrid.getCell( i, j );
                    if ( d1 != noDataValue ) {
                        initCell( ( ( i - startRowIndex ) * ncols ) + ( j - startColIndex ), d1 );
                        sparseness -= cellAsProportionOfGrid;
                    }
                }
            }
        }
        optimiseCollection( "memory" );
        //System.out.println( toString() );
    }
    
    /**
     * Constructs from a File
     * @param gridFile - File containing the binary representation of a grid.
     * These are assumed to be double values written out in row major order with a footer of 13 doubles:
     * (nrows, ncols, xllcorner, yllcorner, cellsize, noDataValue, upToDate, sparseness, sum, min, minCount, max, maxCount).
     * The constructed grid has endColIndex-startColIndex + 1 columns and
     * endRowIndex-startRowIndex + 1 rows.
     */
    protected Grid2DSquareCellDouble( File gridFile, int startRowIndex, int startColIndex, int endRowIndex, int endColIndex, int gridStatisticsType ) {
        //System.out.println( "Grid2DSquareCellDouble( gridascii( " + gridascii + " ), gridStatistics1_0( " + gridStatistics1_0 + " ) ) {" );
        initGridStatistics( gridStatisticsType );
        String filename = gridFile.getName();
        initNrows( endRowIndex - startRowIndex + 1 );
        initNcols( endColIndex - startColIndex + 1 );
        if ( filename.endsWith( "asc" ) ) {
            try {
                StreamTokenizer st = new StreamTokenizer( new BufferedReader( new InputStreamReader( new FileInputStream( gridFile ) ) ) );
                st.wordChars( '_', '_' );
                st.parseNumbers();
                st.eolIsSignificant( false );
                st.lowerCaseMode( true );
                int type = st.nextToken();
                st.nextToken();
                int inputGridNcols = ( int )st.nval;
                st.nextToken();
                st.nextToken();
                int inputGridNrows = ( int )st.nval;
                st.nextToken();
                boolean b1 = true;
                if ( st.sval.equalsIgnoreCase( "xllcenter" ) || st.sval.equalsIgnoreCase( "xllcentre" ) ) {
                    b1 = false;
                }
                st.nextToken();
                double xllcorner = st.nval;
                st.nextToken();
                boolean b2 = true;
                if ( st.sval.equalsIgnoreCase( "yllcenter" ) || st.sval.equalsIgnoreCase( "yllcentre" ) ) {
                    b2 = false;
                }
                st.nextToken();
                double yllcorner = st.nval;
                st.nextToken();
                st.nextToken();
                double cellsize = st.nval;
                setCellsize( cellsize );
                if ( !b1 ) {
                    xllcorner -= ( cellsize / 2.0d );
                }
                if ( !b2 ) {
                    yllcorner -= ( cellsize / 2.0d );
                }
                setXllcorner( xllcorner + ( double ) startColIndex * cellsize );
                setYllcorner( yllcorner + ( double ) ( inputGridNrows - endRowIndex - 1 ) * cellsize );
                type = st.nextToken();
                double noDataValue = Double.NaN;
                if( type == StreamTokenizer.TT_NUMBER ) {
                    // Warning! - if there is no noDataValue and grid values are non-numeric this will break.
                    st.pushBack();
                } else {
                    type = st.nextToken();
                    noDataValue = st.nval;
                }
                initNoDataValue( noDataValue );
                // Having got passed the header st is set so that it recognises E and D as ordinary characters.
                // This is to cope with exponent numbers of the form 0.1E-5 or 0.1D-5; where, E or D means
                // multiplied by ten raised to the power of...
                st.ordinaryChar( 'E' );
                st.ordinaryChar( 'D' );
                // Initialise cell values.
                double d1 = getNoDataValue();
                setGridHashtable( new Hashtable() );
                for ( int i = 0; i < nrows; i ++ ) {
                    for ( int j = 0; j < ncols; j ++ ) {
                        st.nextToken();
                        if ( i >= startRowIndex && i <= endRowIndex ) {
                            if ( j >= startColIndex && j <= endColIndex ) {
                                d1 = st.nval;
                                type = st.nextToken();
                                if ( type != StreamTokenizer.TT_NUMBER && type != StreamTokenizer.TT_EOF ) {
                                    // Either encountered an exponent term number or end of file marker or grid value is non-numeric or something else.
                                    st.nextToken();
                                    d1 = d1 * Math.pow( 10.0, st.nval );
                                } else {
                                    st.pushBack();
                                }
                                initCell( i - startRowIndex, j - startColIndex, d1 );
                            }
                        }
                    }
                }
            } catch ( java.io.IOException e ) {
                System.out.println( e.toString() + " in Grid2DSquareCellDouble( File( " + gridFile + " ), startRowIndex( " + startRowIndex + " ), startColIndex( " + startColIndex + " ), endRowIndex( " + endRowIndex + " ), endColIndex( " + endColIndex + " ), gridStatisticsType( " + gridStatisticsType + " ) )" );
                System.exit( 0 );
            }
        }
        if ( filename.endsWith( "grd" ) ) {
            try {
                setGridHashtable( new Hashtable() );
                RandomAccessFile raf = new RandomAccessFile( gridFile, "rw" );
                raf.seek( raf.length() - ( 13 * 8 ) );
                int inputGridNrows = ( int ) raf.readDouble();
                int inputGridNcols = ( int ) raf.readDouble();
                setXllcorner( raf.readDouble() + ( double ) startColIndex * cellsize );
                setYllcorner( raf.readDouble() + ( double ) ( inputGridNrows - endRowIndex - 1 ) * cellsize );
                double cellsize = raf.readDouble();
                setCellsize( cellsize );
                initNoDataValue( raf.readDouble() );
                AbstractGridStatistics stats = getGridStatistics();
                if ( !( startColIndex >= inputGridNcols || endColIndex < 0 || startRowIndex >= inputGridNrows || endRowIndex < 0 ) ) { // Only need do something if there are coincident values.
                    int caseType=0;
                    // There are 16 cases to deal with.
                    if ( startColIndex <= 0 && startRowIndex <= 0 && endRowIndex >= ( inputGridNrows - 1 ) && endColIndex >= ( inputGridNrows - 1 ) ) {caseType = 1;}
                    if ( startColIndex <= 0 && startRowIndex <= 0 && endRowIndex >= ( inputGridNrows - 1 ) && endColIndex < ( inputGridNrows - 1 ) ) {caseType = 2;}
                    if ( startColIndex <= 0 && startRowIndex <= 0 && endRowIndex < ( inputGridNrows - 1 ) && endColIndex >= ( inputGridNrows - 1 ) ) {caseType = 3;}
                    if ( startColIndex <= 0 && startRowIndex <= 0 && endRowIndex < ( inputGridNrows - 1 ) && endColIndex < ( inputGridNrows - 1 ) ) {caseType = 4;}
                    if ( startColIndex > 0 && startRowIndex <= 0 && endRowIndex >= ( inputGridNrows - 1 ) && endColIndex >= ( inputGridNrows - 1 ) ) {caseType = 5;}
                    if ( startColIndex > 0 && startRowIndex <= 0 && endRowIndex >= ( inputGridNrows - 1 ) && endColIndex < ( inputGridNrows - 1 ) ) {caseType = 6;}
                    if ( startColIndex > 0 && startRowIndex <= 0 && endRowIndex < ( inputGridNrows - 1 ) && endColIndex >= ( inputGridNrows - 1 ) ) {caseType = 7;}
                    if ( startColIndex > 0 && startRowIndex <= 0 && endRowIndex < ( inputGridNrows - 1 ) && endColIndex < ( inputGridNrows - 1 ) ) {caseType = 8;}
                    if ( startColIndex <= 0 && startRowIndex>0 && endRowIndex >= ( inputGridNrows - 1 ) && endColIndex >= ( inputGridNrows - 1 ) ) {caseType = 9;}
                    if ( startColIndex <= 0 && startRowIndex>0 && endRowIndex >= ( inputGridNrows - 1 ) && endColIndex < ( inputGridNrows - 1 ) ) {caseType = 10;}
                    if ( startColIndex <= 0 && startRowIndex>0 && endRowIndex < ( inputGridNrows - 1 ) && endColIndex >= ( inputGridNrows - 1 ) ) {caseType = 11;}
                    if ( startColIndex <= 0 && startRowIndex>0 && endRowIndex < ( inputGridNrows - 1 ) && endColIndex < ( inputGridNrows - 1 ) ) {caseType = 12;}
                    if ( startColIndex > 0 && startRowIndex>0 && endRowIndex >= ( inputGridNrows - 1 ) && endColIndex >= ( inputGridNrows - 1 ) ) {caseType = 13;}
                    if ( startColIndex > 0 && startRowIndex>0 && endRowIndex >= ( inputGridNrows - 1 ) && endColIndex < ( inputGridNrows - 1 ) ) {caseType = 14;}
                    if ( startColIndex > 0 && startRowIndex>0 && endRowIndex < ( inputGridNrows - 1 ) && endColIndex >= ( inputGridNrows - 1 ) ) {caseType = 15;}
                    if ( startColIndex > 0 && startRowIndex>0 && endRowIndex < ( inputGridNrows - 1 ) && endColIndex < ( inputGridNrows - 1 ) ) {caseType = 16;}
                    //System.out.print( "caseType " + caseType );
                    switch ( caseType ) {
                        case 1 :
                            raf.seek( 0 );
                            for ( int i = 0; i < inputGridNrows; i ++ ) {
                                for ( int j = 0; j < inputGridNcols; j ++ ) {
                                    initCell( ( i - startRowIndex ), ( j - startColIndex ), raf.readDouble() );
                                }
                            }
                            break;
                        case 2 :
                            for ( int i = 0; i < inputGridNrows; i ++ ) {
                                raf.seek( ( i * inputGridNcols ) * 8 );
                                for ( int j = 0; j <= endColIndex; j ++ ) {
                                    initCell( ( i - startRowIndex ), ( j - startColIndex ), raf.readDouble() );
                                }
                            }
                            break;
                        case 3 :
                            for ( int i = 0; i <= endRowIndex; i ++ ) {
                                raf.seek( ( i * inputGridNcols ) * 8 );
                                for ( int j = 0; j < inputGridNcols; j ++ ) {
                                    initCell( ( i - startRowIndex ), ( j - startColIndex ), raf.readDouble() );
                                }
                            }
                            break;
                        case 4 :
                            for ( int i = 0; i <= endRowIndex; i ++ ) {
                                raf.seek( ( i * inputGridNcols ) * 8 );
                                for ( int j = 0; j <= endColIndex; j ++ ) {
                                    initCell( ( i - startRowIndex ), ( j - startColIndex ), raf.readDouble() );
                                }
                            }
                            break;
                        case 5 :
                            for ( int i = 0; i < inputGridNrows; i ++ ) {
                                raf.seek( ( ( i * inputGridNcols ) + startColIndex ) * 8 );
                                for ( int j = startColIndex; j < inputGridNcols; j ++ ) {
                                    initCell( ( i - startRowIndex ), ( j - startColIndex ), raf.readDouble() );
                                }
                            }
                            break;
                        case 6 :
                            for ( int i = 0; i < inputGridNrows; i ++ ) {
                                raf.seek( ( ( i * inputGridNcols ) + startColIndex ) * 8 );
                                for ( int j = startColIndex; j <= endColIndex; j ++ ) {
                                    initCell( ( i - startRowIndex ), ( j - startColIndex ), raf.readDouble() );
                                }
                            }
                            break;
                        case 7 :
                            for ( int i = 0; i <= endRowIndex; i ++ ) {
                                raf.seek( ( ( i * inputGridNcols ) + startColIndex ) * 8 );
                                for ( int j = startColIndex; j < inputGridNcols; j ++ ) {
                                    initCell( ( i - startRowIndex ), ( j - startColIndex ), raf.readDouble() );
                                }
                            }
                            break;
                        case 8 :
                            for ( int i = 0; i <= endRowIndex; i ++ ) {
                                raf.seek( ( ( i * inputGridNcols ) + startColIndex ) * 8 );
                                for ( int j = 0; j <= endColIndex; j ++ ) {
                                    initCell( ( i - startRowIndex ), ( j - startColIndex ), raf.readDouble() );
                                }
                            }
                            break;
                        case 9 :
                            for ( int i = startRowIndex; i <inputGridNrows; i ++ ) {
                                raf.seek( ( i * inputGridNcols ) * 8 );
                                for ( int j = 0; j < inputGridNcols; j ++ ) {
                                    initCell( ( i - startRowIndex ), ( j - startColIndex ), raf.readDouble() );
                                }
                            }
                            break;
                        case 10 :
                            for ( int i = startRowIndex; i < inputGridNrows; i ++ ) {
                                raf.seek( ( i * inputGridNcols ) * 8 );
                                for ( int j = 0; j <= endColIndex; j ++ ) {
                                    initCell( ( i - startRowIndex ), ( j - startColIndex ), raf.readDouble() );
                                }
                            }
                            break;
                        case 11 :
                            for ( int i = startRowIndex; i <= endRowIndex; i ++ ) {
                                raf.seek( ( i * inputGridNcols ) * 8 );
                                for ( int j = 0; j < inputGridNcols; j ++ ) {
                                    initCell( ( i - startRowIndex ), ( j - startColIndex ), raf.readDouble() );
                                }
                            }
                        case 12 :
                            for ( int i = startRowIndex; i <= endRowIndex; i ++ ) {
                                raf.seek( ( i * inputGridNcols ) * 8 );
                                for ( int j = 0; j <= endColIndex; j ++ ) {
                                    initCell( ( i - startRowIndex ), ( j - startColIndex ), raf.readDouble() );
                                }
                            }
                            break;
                        case 13 :
                            for ( int i = startRowIndex; i < inputGridNrows; i ++ ) {
                                raf.seek( ( ( i * inputGridNcols ) + startColIndex ) * 8 );
                                for ( int j = startColIndex; j < inputGridNcols; j ++ ) {
                                    initCell( ( i - startRowIndex ), ( j - startColIndex ), raf.readDouble() );
                                }
                            }
                            break;
                        case 14 :
                            for ( int i = startRowIndex; i < inputGridNrows; i ++ ) {
                                raf.seek( ( ( i * inputGridNcols ) + startColIndex ) * 8 );
                                for ( int j = startColIndex; j <= endColIndex; j ++ ) {
                                    initCell( ( i - startRowIndex ), ( j - startColIndex ), raf.readDouble() );
                                }
                            }
                            break;
                        case 15 :
                            for ( int i = startRowIndex; i <= endRowIndex; i ++ ) {
                                raf.seek( ( ( i * inputGridNcols ) + startColIndex ) * 8 );
                                for ( int j = startColIndex; j < inputGridNcols; j ++ ) {
                                    initCell( ( i - startRowIndex ), ( j - startColIndex ), raf.readDouble() );
                                }
                            }
                            break;
                        case 16 :
                            for ( int i = startRowIndex; i <= endRowIndex; i ++ ) {
                                raf.seek( ( ( i * inputGridNcols ) + startColIndex ) * 8 );
                                for ( int j = startColIndex; j <= endColIndex; j ++ ) {
                                    initCell( ( i - startRowIndex ), ( j - startColIndex ), raf.readDouble() );
                                }
                            }
                            break;
                    }
                }
            } catch ( java.io.IOException e ) {
                System.out.println( e.toString() + " in Grid2DSquareCellDouble( File( " + gridFile + " ), startRowIndex( " + startRowIndex + " ), startColIndex( " + startColIndex + " ), endRowIndex( " + endRowIndex + " ), endColIndex( " + endColIndex + " ), gridStatisticsType( " + gridStatisticsType + " ) )" );
                System.exit( 0 );
            }
        }
        optimiseCollection( "memory" );
        //System.out.println( toString() );
    }
    
    /**
     * Constructs from a uk.ac.leeds.ccg.raster.Raster
     * This can probably be made faster by cloning the raster data array or
     * Hashtable, converting sparseness and other summary stats. There is a
     * problem though:
     * raster.getMin() may not return the minimum value of the raster
     * raster.getMax() may not return the maximum value of the raster
     * There is no fast way of finding out ( i.e. no flag like isUpToDate )
     */
    protected Grid2DSquareCellDouble( Raster raster, int gridStatisticsType ) {
        //System.out.println( "Grid2DSquareCellDouble( Raster( " + raster.toString() + " ), gridStatistics1_0( " + gridStatistics1_0 + " ) ) {" );
        initGridStatistics( gridStatisticsType );
        int ncols = raster.getWidth();
        int nrows = raster.getHeight();
        initNrows( nrows );
        initNcols( ncols );
        setXllcorner( raster.getOriginx() );
        setYllcorner( raster.getOriginy() );
        setCellsize( raster.getCellSize() );
        initNoDataValue( raster.getMissingValueCode() );
        //initNoDataValue( raster.getNoDataValue() );
        setGridHashtable( new Hashtable() );
        for ( int i = 0; i < nrows; i ++ ) {
            for ( int j = 0; j < ncols; j ++ ) {
                initCell( i, j, raster.getCell( i, j ) );
            }
        }
        optimiseCollection( "memory" );
        //System.out.println( toString() );
    }
    
    /**
     * Returns the data type
     */
    public int getDataType() {
        return 0;
    }
    
    /**
     * Returns Hashtable containing data.
     */
    public Hashtable getGridHashtable() {
        if ( ! getIsHashtable() ) {
            toHashtable();
        }
        return this.gridHashtable;
    }
    
    /**
     * Sets gridHashtable.
     */
    private void setGridHashtable( Hashtable gridHashtable ) {
        this.gridHashtable = gridHashtable;
        setIsHashtable( true );
    }
    
    /**
     * Gets isHashtable
     */
    public boolean getIsHashtable() {
        return this.isHashtable;
    }
    
    /**
     * Sets isHashtable
     */
    private void setIsHashtable( boolean isHashtable ) {
        this.isHashtable = isHashtable;
    }
    
    /**
     * Returns double[] containing data.
     */
    public double[] getGridArray() {
        if ( getIsHashtable() ) {
            toArray();
        }
        return this.gridArray;
    }
    
    /**
     * sets gridArray.
     */
    private void setGridArray( double[] gridArray ) {
        this.gridArray = gridArray;
        setIsHashtable( false );
    }
    
    /**
     * Returns a description of the grid.
     */
    public String toString() {
        AbstractGridStatistics stats = getGridStatistics();
        return "isHashtable( " + getIsHashtable() + " ), nrows( " + getNrows() + " ), ncols( " + getNcols() + " ), xllcorner( " + getXllcorner() + " ), yllcorner( " + getYllcorner() + " ), cellsize( " + getCellsize() + " ), noDataValue( " + getNoDataValue() + " ), " + stats.toString();
    }
    
    /**
     * Returns the value at row i, column j or noDataValue.
     */
    public double getCell( int i, int j ) {
        //System.out.println( "getCell( row( " + i + " ), col( " + j + " ) )" );
        double result = getNoDataValue();
        if ( inGrid( i, j ) ) {
            if ( getIsHashtable() ) {
                Hashtable gridHashtable = getGridHashtable();
                Double d1 = ( Double )gridHashtable.get( new Integer( getCellID( i, j ) ) );
                if ( d1 != null ) {
                    result = d1.doubleValue();
                }
            } else {
                double[] gridArray = getGridArray();
                result = gridArray[ ( i * ncols ) + j ];
            }
        }
        return result;
    }
    
    /**
     * Sets the value at row i column j to newValue
     */
    public double setCell( int i, int j, double newValue ) {
        //System.out.println( "setCell( row( " + i + " ), col( " + j + " ), value( " + newValue + " ) ) {" );
        double noDataValue = getNoDataValue();
        double result = noDataValue;
        if ( inGrid( i, j ) ) {
            AbstractGridStatistics stats = getGridStatistics();
            int gridStatisticsType = stats.getGridStatisticsType();
            int ncols = getNcols();
            int nrows = getNrows();
            double cellAsProportionOfGrid = ( 1.0d / ( double ) ( nrows * ncols ) );
            //@TODO Check cellAsProportionOfGrid within tollerance of range [ 0, 1 ]
            if ( getIsHashtable() ) {
                Double d2;
                if ( newValue != noDataValue ) {
                    d2 = ( Double )gridHashtable.put( new Integer( ( i * ncols ) + j ), new Double( newValue ) );
                    if ( d2 == null ) {
                        stats.setSparseness( stats.getSparseness() - cellAsProportionOfGrid );
                        if ( gridStatisticsType == 0 ) {
                            stats.setSum( stats.getSum() + newValue );
                            stats.doMin( newValue );
                            stats.doMax( newValue );
                        }
                        result = noDataValue;
                    } else {
                        double d2doubleValue = d2.doubleValue();
                        if ( gridStatisticsType == 0 ) {
                            if ( d2doubleValue == stats.getMin() ) { stats.setMinCount( stats.getMinCount() - 1 ); }
                            if ( d2doubleValue == stats.getMax() ) { stats.setMaxCount( stats.getMaxCount() - 1 ); }
                            stats.setSum( stats.getSum() - d2doubleValue + newValue );
                            stats.doMin( newValue );
                            stats.doMax( newValue );
                        }
                        result = d2doubleValue;
                    }
                } else {
                    d2 = ( Double )gridHashtable.remove( new Integer( getCellID( i, j ) ) );
                    if ( d2 != null ) {
                        stats.setSparseness( stats.getSparseness() + cellAsProportionOfGrid );
                        double d2doubleValue = d2.doubleValue();
                        if ( gridStatisticsType == 0 ) {
                            if ( d2doubleValue == stats.getMin() ) { stats.setMinCount( stats.getMinCount() - 1 ); }
                            if ( d2doubleValue == stats.getMax() ) { stats.setMaxCount( stats.getMaxCount() - 1 ); }
                            stats.setSum( stats.getSum() - d2doubleValue );
                        }
                        result = d2doubleValue;
                    }
                }
            } else {
                double[] gridArray = getGridArray();
                double d2 = gridArray[( i * ncols ) + j];
                result = d2;
                if ( newValue == noDataValue ) {
                    stats.setSparseness( stats.getSparseness() + cellAsProportionOfGrid );
                } else {
                    if ( gridStatisticsType == 0 ) {
                        stats.setSum( stats.getSum() + newValue );
                        stats.doMin( newValue );
                        stats.doMax( newValue );
                    }
                }
                if ( d2 == noDataValue ) {
                    stats.setSparseness( stats.getSparseness() - cellAsProportionOfGrid );
                } else {
                    if ( gridStatisticsType == 0 ) {
                        if ( d2 == stats.getMin() ) { stats.setMinCount( stats.getMinCount() - 1 ); }
                        if ( d2 == stats.getMax() ) { stats.setMaxCount( stats.getMaxCount() - 1 ); }
                        stats.setSum( stats.getSum() - d2 );
                    }
                }
                gridArray[( i * ncols ) + j] = newValue;
            }
            optimiseCollection( "memory" );
        }
        return result;
    }
    
    /**
     * Initilises the value at row i column j to value.
     */
    protected void initCell( int i, int j, double value ) {
        //System.out.println( "initCell( row( " + i + " ), col( " + j + " ), value( " + value + " ) ) {" );
        if ( inGrid( i, j ) ) {
            AbstractGridStatistics stats = getGridStatistics();
            int gridStatisticsType = stats. getGridStatisticsType();
            int ncols = getNcols();
            double noDataValue = getNoDataValue();
            if ( getIsHashtable() ) {
                if ( value != noDataValue ) {
                    stats.setSparseness( stats.getSparseness() - ( 1.0d / ( double ) ( getNrows() * ncols ) ) );
                    if ( gridStatisticsType == 0 ) {
                        stats.doMin( value );
                        stats.doMax( value );
                        stats.setSum( stats.getSum() + value );
                    }
                    getGridHashtable().put( new Integer( ( i * ncols ) + j ), new Double( value ) );
                }
            } else {
                if ( value != noDataValue ) {
                    stats.setSparseness( stats.getSparseness() - ( 1.0d / ( double ) ( getNrows() * ncols ) ) );
                    if ( gridStatisticsType == 0 ) {
                        stats.doMin( value );
                        stats.doMax( value );
                        stats.setSum( stats.getSum() + value );
                    }
                }
                double[] gridArray = getGridArray();
                gridArray[( i * ncols ) + j] = value;
            }
        }
    }
    
    /**
     * Deletes collection that stores grid values rendering the grid useless and freeing up resources
     * The gridStatistics stay as they were.
     */
    public void clear() {
        //System.out.println( "clear() {" );
        this.gridHashtable = null;
        this.gridArray = null;
        System.gc();
    }
    
    /**
     * -----------------
     * Converter methods
     * -----------------
     */
    
    /**
     * Change internal storage from Array to Hashtable
     */
    private void toHashtable() {
        //System.out.println( "toHashtable() {" );
        if ( ! getIsHashtable() ) {
            int nrows = getNrows();
            int ncols = getNcols();
            double noDataValue = getNoDataValue();
            double d1 = noDataValue;
            Hashtable gridHashtable = new Hashtable();
            double[] gridArray = getGridArray();
            int cellID;
            for ( int i = 0; i < nrows; i ++ ) {
                for ( int j = 0; j < ncols; j ++ ) {
                    cellID = getCellID( i, j );
                    d1 = gridArray[ cellID ];
                    if ( d1 != noDataValue ) {
                        gridHashtable.put( new Integer( cellID ), new Double( d1 ) );
                    }
                }
            }
            setGridHashtable( gridHashtable );
            setIsHashtable( true );
            gridArray = null;
            System.gc();
        }
    }
    
    /**
     * Changes internal storage from Array to Hashtable if this.GetSparseness() is greater than value.
     */
    private void toHashtable( double value ) {
        double sparseness = getGridStatistics().getSparseness();
        if ( sparseness > value ) {
            toHashtable();
        }
    }
    
    /**
     * Changes internal storage from Hashtable to Array
     */
    private void toArray() {
        //System.out.println( "toArray() {" );
        if ( getIsHashtable() ) {
            int nrows = getNrows();
            int ncols = getNcols();
            int cellID;
            double[] gridArray = new double[ nrows * ncols ];
            double noDataValue = getNoDataValue();
            Hashtable gridHashtable = getGridHashtable();
            Double D1;
            for ( int i = 0; i < nrows; i ++ ) {
                for ( int j = 0; j < ncols; j ++ ) {
                    cellID = getCellID( i, j );
                    D1 = ( Double ) gridHashtable.get( new Integer( cellID ) );
                    if ( D1 != null ) {
                        gridArray[ cellID ] = D1.doubleValue();
                    } else {
                        gridArray[ cellID ] = noDataValue;
                    }
                }
            }
            setGridArray( gridArray );
            setIsHashtable( false );
            gridHashtable = null;
            System.gc();
        }
    }
    
    /**
     * Changes internal storage from Hashtable to Array if this.GetSparseness() is less than value.
     */
    private void toArray( double value ) {
        double sparseness = getGridStatistics().getSparseness();
        if ( sparseness < value ) {
            toArray();
        }
    }
    
    /**
     * -----------------
     * Optimiser methods
     * -----------------
     */
    protected void optimiseCollection( String command ) {
        if ( command.equalsIgnoreCase( "memory" ) ) {
            toArray( 0.4 );
            toHashtable( 0.8 );
        }
    }
}