package uk.ac.leeds.ccg.grids;
/**
 * -------------------------------------------------------------------------
 * This file is distributed as part of the uk.ac.leeds.ccg.grids library.
 * This library is a resource for manipulating spatial data stored as grids.
 * Copyright (C) 2002  Andy Turner
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 */
/**
 * ---------------------------
 * Last updated on 27 Nov 2002
 * @author andyt
 * @version 1.0
 * ---------------------------
 * Description:
 * This class is instantiated in the constructor of an AbstractsGrid2DSquareCellDouble
 * extended class (a grid) or replaces an existing instance of a class extended from
 * AbstractGridStatistics. The idea of this class is to enable various grid statistics
 * to be kept up-to-date and quickly retrievable as the underlying data in grid is
 * changed. It currently provides fast access to min, minCount, max, maxCount and sum
 * statistics for the grid.
 *
 * Notes:
 * 1. If the min, max or sum are not likely to be asked for then grid processing will
 *    generally be faster if instead a GridStatistics1 instance was invoked. This can
 *    sensibly be done either at initialisation or via the toGridStatistics converter
 *    method.
 *
 * @ToDo:
 * 1. Add a clone method
 */
public class GridStatistics0 extends AbstractGridStatistics {
    
    /**
     * ------
     * Fields
     * ------
     */
    
    /**
     * Indicates if the min needs recalulating
     * The min would need recalculating if the only min value in the grid was
     * replaced by a smaller value or a noDataValue
     */
    private boolean recalculateMin = true;
    
    /**
     * Indicates if the max needs recalulating
     * The max would need recalculating if the only max value in the grid was
     * replaced by a larger value or a noDataValue
     */
    private boolean recalculateMax = true;
    
    /**
     * -------------------
     * Constructor methods
     * -------------------
     */
    
    /**
     * Creates a new instance of GridStatistics0
     * @param grid - the AbstractGrid2DSquareCellDouble that holds a reference to this
     */
    public GridStatistics0( AbstractGrid2DSquareCellDouble grid ) {
        // Initialise
        init( grid );
        setRecalculateMax( false );
        setRecalculateMin( false );
    }
    
    /**
     * Creates a new instance of GridStatistics0 to replace gridStatistics
     * @param gridStatistics - the gridStatistics which this will replace
     */
    public GridStatistics0( AbstractGridStatistics gridStatistics ) {
        AbstractGrid2DSquareCellDouble grid = gridStatistics.getGrid();
        int gridStatisticsType = gridStatistics.getGridStatisticsType();
        switch ( gridStatisticsType ) {
            case 0 :
                GridStatistics0 dummy0 = ( GridStatistics0 ) gridStatistics;
                sum = dummy0.sum;
                setSparseness( dummy0.getSparseness() );
                setMax( dummy0.max );
                maxCount = dummy0.maxCount;
                recalculateMax = dummy0.getRecalculateMax();
                min = dummy0.min;
                minCount = dummy0.minCount;
                recalculateMin = dummy0.getRecalculateMin();
                break;
            default :
                GridStatistics1 dummy1 = ( GridStatistics1 ) gridStatistics;
                dummy1.doStats();
                setSum( dummy1.getSum() );
                setSparseness( dummy1.getSparseness() );
                setMax( dummy1.getMax() );
                setMaxCount( dummy1.getMaxCount() );
                setRecalculateMax( false );
                setMin( dummy1.getMin() );
                setMinCount( dummy1.getMinCount() );
                setRecalculateMin( false );
                break;
        }
        setGrid( grid );
        grid.setGridStatistics( this );
    }
    
    /**
     * ---------------
     * General methods
     * ---------------
     */
    
    /**
     * Returns the type of gridStatistics:
     * 0 - GridStatistics0
     * 1 - GridStatistics1
     */
    public int getGridStatisticsType() {
        return 0;
    }
    
    /**
     * Returns true iff all stats are up-to-date
     */
    public boolean getIsUpToDate() {
        return ! getRecalculateMin() && ! getRecalculateMax() ;
    }
    
    /**
     * Sets recalculateMax and recalculateMin
     */
    protected void setIsUpToDate( boolean value ) {
        setRecalculateMax( ! value );
        setRecalculateMin( ! value );
    }
    
    /**
     * Sets sum
     */
    protected void setSum( double sum ) {
        this.sum = sum;
    }
    
    /**
     * Sets min
     */
    protected void setMin( double min ) {
        this.min = min;
    }
    
    /**
     * Sets min to the minimum of it and value and corrects minCount.
     */
    protected void doMin( double value ) {
        AbstractGrid2DSquareCellDouble grid = getGrid();
        double noDataValue = grid.getNoDataValue();
        if ( value != noDataValue ) {
            double min = getMin();
            if ( value < min ) { setMinCount( 1 ); setMin( value ); setRecalculateMin( false ); }
            if ( value == min ) { setMinCount( getMinCount() + 1 ); }
        }
    }
    
    /**
     * Returns the minimum value in the grid
     */
    public double getMin() {
        if ( getRecalculateMin() ) {
            AbstractGrid2DSquareCellDouble grid = getGrid();
            int nrows = grid.getNrows();
            int ncols = grid.getNcols();
            double min = Double.POSITIVE_INFINITY;
            int minCount = maxCount;
            double noDataValue = grid.getNoDataValue();
            double d1;
            for ( int i = 0; i < nrows; i ++ ) {
                for ( int j = 0; j < ncols; j ++ ) {
                    d1 = grid.getCell( i, j );
                    if ( d1 != noDataValue ) {
                        if ( d1 < min ) { minCount = 1;min = d1; }
                        if ( d1 == min ) { minCount ++; }
                    }
                }
            }
            this.min = min;
            this.minCount = minCount;
            setRecalculateMin( false );
        }
        return this.min;
    }
    
    /**
     * Returns a count of the number of cells in the grid with the minimum value
     */
    public int getMinCount() {
        if ( getRecalculateMin() ) {
            getMin();
        }
        return this.minCount;
    }
    
    /**
     * Sets minCount
     */
    protected void setMinCount( int minCount ) {
        this.minCount = minCount;
    }
    
    /**
     * Gets recalulateMin.
     */
    public boolean getRecalculateMin() {
        return this.recalculateMin;
    }
    
    /**
     * Sets recalulateMin.
     */
    private void setRecalculateMin( boolean recalculateMin ) {
        this.recalculateMin = recalculateMin;
    }
    
    /**
     * Sets max.
     */
    protected void setMax( double max ) {
        this.max = max;
    }
    
    /**
     * Sets max to the maximum of it and value and corrects minCount.
     */
    protected void doMax( double value ) {
        AbstractGrid2DSquareCellDouble grid = getGrid();
        double noDataValue = grid.getNoDataValue();
        if ( value != noDataValue ) {
            double max = getMax();
            if ( value > max ) { setMaxCount( 1 ); setMax( value ); setRecalculateMax( false ); }
            if ( value == max ) { setMaxCount( getMaxCount() + 1 ); }
        }
    }
    
    /**
     * Returns the maximum value in the Grid.
     */
    public double getMax() {
        if ( getRecalculateMax() ) {
            AbstractGrid2DSquareCellDouble grid = getGrid();
            int nrows = grid.getNrows();
            int ncols = grid.getNcols();
            double noDataValue = grid.getNoDataValue();
            double max = Double.NEGATIVE_INFINITY;
            int maxCount = nrows * ncols;
            double d1;
            for ( int i = 0; i < nrows; i ++ ) {
                for ( int j = 0; j < ncols; j ++ ) {
                    d1 = grid.getCell( i, j );
                    if ( d1 != noDataValue ) {
                        if ( d1 > max ) { maxCount = 1; max = d1; }
                        if ( d1 == max ) { maxCount ++; }
                    }
                }
            }
            this.max = max;
            this.maxCount = maxCount;
            setRecalculateMax( false );
        }
        return this.max;
    }
    
    /**
     * Returns a count of the number of cells in the grid with the maximum value
     */
    public int getMaxCount() {
        if ( getRecalculateMax() ) {
            getMax();
        }
        return this.maxCount;
    }
    
    /**
     * Sets maxCount
     */
    protected void setMaxCount( int maxCount ) {
        this.maxCount = maxCount;
    }
    
    /**
     * Returns recalulateMax
     */
    public boolean getRecalculateMax() {
        return this.recalculateMax;
    }
    
    /**
     * Sets recalulateMax
     */
    private void setRecalculateMax( boolean recalculateMax ) {
        this.recalculateMax = recalculateMax;
    }
    
}
