package uk.ac.leeds.ccg.grids;
/**
 * -------------------------------------------------------------------------
 * This file is distributed as part of the uk.ac.leeds.ccg.grids library.
 * This library is a resource for manipulating spatial data stored as grids.
 * Copyright (C) 2002  Andy Turner
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 */
import uk.ac.leeds.ccg.raster.Raster;
import uk.ac.leeds.ccg.raster.circleRaster;
import uk.ac.leeds.ccg.geotools.GeoRectangle;
import uk.ac.leeds.ccg.geotools.GeoData;
import uk.ac.leeds.ccg.geotools.SimpleGeoData;
import uk.ac.leeds.ccg.geotools.GeoCircle;
import uk.ac.leeds.ccg.geotools.CircleLayer;
import java.io.File;
import java.io.StreamTokenizer;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.PrintWriter;
import java.util.Vector;
import javax.imageio.ImageIO;
/**
 * ---------------------------
 * Last updated on 25 Nov 2002
 * @author andyt
 * @version 1.0
 * ---------------------------
 * Description:
 *
 * This class holds general utility methods
 *
 * @TODO
 * 1. Make this static so there is no need for constructer and constructing where methods are used in package
 */
public class Utilities {
    
    /** Creates a new instance of Utilities */
    public Utilities() {
    }
    
    /**
     * Returns a temporary File
     */
    public static File makeTemporaryFile() {
        return makeTemporaryFile( System.getProperty( "java.io.tmpdir" ) );
    }
    public static File makeTemporaryFile( String dataDirectory ) {
        return makeTemporaryFile( dataDirectory, "f" );
    }
    public static File makeTemporaryFile( String dataDirectory, String firstPartOfFileName ) {
        return makeTemporaryFile( dataDirectory, firstPartOfFileName, ".tmp" );
    }
    public static File makeTemporaryFile( String dataDirectory, String firstPartOfFileName, String fileExtension ) {
        if ( dataDirectory == null ) {
            dataDirectory = System.getProperty( "java.io.tmpdir" );
        }
        boolean isNewFile = false;
        String filename = new String( dataDirectory + firstPartOfFileName + System.currentTimeMillis() + fileExtension );
        File file = new File( filename );
        try {
            if ( file.exists() ) {
                String addition = new String( "a" );
                while ( ! isNewFile ) {
                    filename = dataDirectory + firstPartOfFileName + System.currentTimeMillis() + addition + "." + fileExtension;
                    file = new File( filename );
                    if ( ! file.exists() ) {
                        file.createNewFile();
                        isNewFile = true;
                    }
                    addition = addition + "a";
                }
            } else {
                file.createNewFile();
            }
            // Set to delete on exiting the Java Virtual Machine
            file.deleteOnExit();
        } catch ( java.io.IOException e ) {
            System.out.println( e.toString() + " in Utilities makeTemporaryFile( dataDirectory( " + dataDirectory + " ), firstPartOfFileName( " + firstPartOfFileName + " ), fileExtension( " + fileExtension + " ) )" );
            System.exit( 0 );
        }
        return file;
    }
    
    /**
     * Returns the minimum double that can be added to a double in the range [ min, max ] such that
     * the result is larger than any double in the range.
     */
    public static double minTollerance( double min, double max ) {
        double tolleranceMin = min;
        double dummyMin = min;
        double tolleranceMax = max;
        double dummyMax = max;
        boolean calculated = false;
        while ( ! calculated ) {
            tolleranceMin /= 10.0d;
            tolleranceMax /= 10.0d;
            dummyMin = min + tolleranceMin;
            dummyMax = max + tolleranceMax;
            if ( ! ( dummyMin > min && dummyMax > max ) ) {
                calculated = true;
            }
        }
        return Math.min( tolleranceMin * 10.0d, tolleranceMax * 10 );
    }
    
    /**
     * Returns the maximum double that can be added to a double in the range [ min, max ] such that
     * it can be taken away again and the same number returned
     */
    public static double maxTollerance( double min, double max ) {
        double tolleranceMin = min;
        double dummyMin = min;
        double tolleranceMax = max;
        double dummyMax = max;
        boolean calculated = false;
        while ( ! calculated ) {
            tolleranceMin *= 10.0d;
            tolleranceMax *= 10.0d;
            dummyMin += tolleranceMin;
            dummyMax += tolleranceMax;
            dummyMin -= tolleranceMin;
            dummyMax -= tolleranceMax;
            if ( dummyMin != min || dummyMax != max ) {
                calculated = true;
            }
        }
        return Math.max( tolleranceMin / 10.0d, tolleranceMax / 10 );
    }

    /**
     * Returns true if writerType is available
     * @param writerType - usually a well known text String for an image type
     */
    public static boolean isImageWriterAvailable( String writerType ) {
        boolean result = false;
        String[] writerTypes = ImageIO.getWriterMIMETypes();
        for ( int i = 0 ; i < writerTypes.length; i ++ ) {
            if ( writerTypes[ i ].equalsIgnoreCase( "image/" + writerType ) ) {
                result = true;
            }
        }
        return result;
    }
    
    // Creates gamInputFile from the countFile and parFile.
    // The basic format is:
    // id x y count par
    public void GenerateGamInput( File gamInputFile, File countFile, File parFile, String gridOrAscii ) {
        String dataDirectory = countFile.getParent();
        //System.out.println( dataDirectory );
        Grid2DSquareCellDoubleFileFactory ff = new Grid2DSquareCellDoubleFileFactory();
        Grid2DSquareCellDoubleFactory f = new Grid2DSquareCellDoubleFactory();
        AbstractGrid2DSquareCellDouble countGrid;
        AbstractGrid2DSquareCellDouble parGrid;
        int gridStatisticsType = 0;
        try {
            if ( gridOrAscii.equalsIgnoreCase( "ascii" ) ) {
                countGrid = f.createGrid2DSquareCellDouble( countFile, gridStatisticsType );
                parGrid = f.createGrid2DSquareCellDouble( parFile, gridStatisticsType );
            } else {
                countGrid = f.createGrid2DSquareCellDouble( countFile );
                parGrid = f.createGrid2DSquareCellDouble( parFile );
            }
        } catch ( java.lang.OutOfMemoryError e ) {
            if ( gridOrAscii.equalsIgnoreCase( "ascii" ) ) {
                ff.setFile( new File( new String( dataDirectory + "count" ) ) );
                countGrid = ff.createGrid2DSquareCellDouble( countFile, gridStatisticsType );
                ff.setFile( new File( new String( dataDirectory + "par" ) ) );
                parGrid = ff.createGrid2DSquareCellDouble( parFile, gridStatisticsType );
            } else {
                ff.setFile( new File( new String( dataDirectory + "count" ) ) );
                countGrid = f.createGrid2DSquareCellDouble( countFile );
                ff.setFile( new File( new String( dataDirectory + "par" ) ) );
                parGrid = f.createGrid2DSquareCellDouble( parFile );
            }
        }
        System.out.println( countGrid.toString() );
        System.out.println( parGrid.toString() );
        GenerateGamInput( gamInputFile, countGrid, parGrid );
    }
    // Creates gamInputFile from the AbstractGrid2DSquareCellDouble count and AbstractGrid2DSquareCellDouble par grids.
    public void GenerateGamInput( File gamInputFile, AbstractGrid2DSquareCellDouble countGrid, AbstractGrid2DSquareCellDouble parGrid ) {
        //System.out.println( "GridToGam ( gamInputFile( " + gamInputFile + " ), countGrid( " + countGrid.toString() + " ), parGrid( " + parGrid.toString() + " ) )" );
        int nrows = countGrid.getNrows();
        int ncols = countGrid.getNcols();
        double xllcorner = countGrid.getXllcorner();
        double yllcorner = countGrid.getYllcorner();
        double cellsize = countGrid.getCellsize();
        double noDataValue = countGrid.getNoDataValue();
        // Generate Gam input file
        PrintWriter pw = null;
        try {
            pw = new PrintWriter( new BufferedWriter( new OutputStreamWriter( new FileOutputStream( gamInputFile ) ) ) );
        } catch (java.io.FileNotFoundException e) {
            System.out.println( e + " in GridToGam ( gamInputFile( " + gamInputFile + " ), countGrid( " + countGrid.toString() + " ), parGrid( " + parGrid.toString() + " ) )" );
            System.exit( 0 );
        }
        double countNoDataValue = countGrid.getNoDataValue();
        double parNoDataValue = parGrid.getNoDataValue();
        System.out.println( "parNoDataValue " + parNoDataValue );
        System.out.println( "countNoDataValue " + countNoDataValue );
        //double parNoDataValue = 0.0d;
        double x = 0.0d;
        double y = 0.0d;
        double count = 0.0d;
        double par = 0.0d;
        for ( int i = 0; i < nrows * ncols; i++ ) {
            count = countGrid.getCell( i );
            par = parGrid.getCell( i );
            if ( count != countNoDataValue && par != parNoDataValue ) {
                //System.out.println( i + " " + countGrid.getCellX( i ) + " " + countGrid.getCellY( i ) + " " + count + " " + par );
                pw.println( i + " " + countGrid.getCellX( i ) + " " + countGrid.getCellY( i ) + " " + count + " " + par );
            }
        }
        pw.flush();
        pw.close();
    }
    
    public Raster GamOutputToRaster( File gamOutputFile, AbstractGrid2DSquareCellDouble grid ) {
        //System.out.println( "GamOutputToRaster( gamOutputFile( " + gamOutputFile.toString() + " ), grid( " + grid.toString() +  " ) )" );
        int nrows = grid.getNrows();
        int ncols = grid.getNcols();
        double xllcorner = grid.getXllcorner();
        double yllcorner = grid.getYllcorner();
        double cellsize = grid.getCellsize();
        double noDataValue = grid.getNoDataValue();
        GeoRectangle geoRectangle = new GeoRectangle( xllcorner, yllcorner, ( cellsize * ncols ), ( cellsize * nrows ) );
        Raster raster = new Raster( geoRectangle, cellsize );
        Vector geoCircles = new Vector();
        SimpleGeoData geoData = new SimpleGeoData();
        GeoCircle geoCircle = null;
        try {
            StreamTokenizer st = new StreamTokenizer( new BufferedReader( new InputStreamReader( new FileInputStream( gamOutputFile ) ) ) );
            st.eolIsSignificant( false );
            st.parseNumbers();
            st.whitespaceChars( ',', ',' );
            st.wordChars( '"', '"' );
            int tokenType = st.nextToken();
            int read = 0;
            double x = 0.0d;
            double y = 0.0d;
            double radius = 0.0d;
            int id = 0;
            while ( tokenType != StreamTokenizer.TT_EOF ) {
                switch ( tokenType ) {
                    case StreamTokenizer.TT_NUMBER :
                        switch ( read ) {
                            case 0 :
                                read = 1;
                                break;
                            case 1 :
                                x = st.nval;
                                read = 2;
                                break;
                            case 2 :
                                y = st.nval;
                                read = 3;
                                break;
                            case 3 :
                                radius = st.nval;
                                read = 4;
                                break;
                            case 4 :
                                geoData.setValue( id, st.nval );
                                geoCircle = new GeoCircle( id, x, y, radius );
                                geoCircles.add( geoCircle );
                                id ++;
                                read = 0;
                                break;
                        }
                        break;
                    default :
                        break;
                }
                tokenType = st.nextToken();
            }
        }
        catch ( java.io.IOException e ) {
            System.out.println( e + " in GamOutputToRaster( gamOutputFile( " + gamOutputFile.toString() + " ), grid( " + grid.toString() +  " ) )" );
            System.exit( 0 );
        }
        return null;
        //return new circleRaster( geoCircles, geoData, cellsize, geoRectangle );
    }
    
    public CircleLayer GamOutputToCircleLayer( File gamOutputFile, AbstractGrid2DSquareCellDouble grid ) {
        System.out.println( "GamOutputToCircleLayer( gamOutputFile( " + gamOutputFile.toString() + " ), grid( " + grid.toString() +  " ) )" );
        int nrows = grid.getNrows();
        int ncols = grid.getNcols();
        double xllcorner = grid.getXllcorner();
        double yllcorner = grid.getYllcorner();
        double cellsize = grid.getCellsize();
        double noDataValue = grid.getNoDataValue();
        GeoRectangle geoRectangle = new GeoRectangle( xllcorner, yllcorner, ( cellsize * ncols ), ( cellsize * nrows ) );
        //Raster raster = new Raster( geoRectangle, cellsize );
        CircleLayer circleLayer = new CircleLayer();
        //Vector geoCircles = new Vector();
        SimpleGeoData geoData = new SimpleGeoData();
        GeoCircle geoCircle = null;
        try {
            StreamTokenizer st = new StreamTokenizer( new BufferedReader( new InputStreamReader( new FileInputStream( gamOutputFile ) ) ) );
            //st.eolIsSignificant( false );
            st.parseNumbers();
            int tokenType = st.nextToken();
            int read = 0;
            int id = 0;
            double x = 0.0d;
            double y = 0.0d;
            double radius = 0.0d;
            double value = 0.0d;
            while ( tokenType != StreamTokenizer.TT_EOF ) {
                switch ( tokenType ) {
                    case StreamTokenizer.TT_NUMBER :
                        switch ( read ) {
                            case 0 :
                                read = 1;
                                break;
                            case 1 :
                                x = st.nval;
                                read = 2;
                                break;
                            case 2 :
                                y = st.nval;
                                read = 3;
                                break;
                            case 3 :
                                radius = st.nval;
                                read = 4;
                                break;
                            case 4 :
                                value = st.nval;
                                geoData.setValue( id, value );
                                geoCircle = new GeoCircle( id, x, y, radius );
                                System.out.println( id + ", " + x + ", " + y + ", " + radius + ", " + value );
                                circleLayer.addGeoCircle( geoCircle );
                                id ++;
                                read = 0;
                                break;
                        }
                        break;
                    default :
                        break;
                }
                tokenType = st.nextToken();
            }
        }
        catch ( java.io.IOException e ) {
            System.out.println( e + " in GamOutputToRaster( gamOutputFile( " + gamOutputFile.toString() + " ), grid( " + grid.toString() +  " ) )" );
            System.exit( 0 );
        }
        return circleLayer;
    }
       
    public void XYFileToGrid( File XYFile, AbstractGrid2DSquareCellDouble grid ) {
        System.out.println( "XYFileToGrid( XYFile( " + XYFile.toString() + " ), grid ( " + grid.toString() + " ) )" );
        int nrows = grid.getNrows();
        int ncols = grid.getNcols();
        double xllcorner = grid.getXllcorner();
        double yllcorner = grid.getYllcorner();
        double cellsize = grid.getCellsize();
        double noDataValue = grid.getNoDataValue();
        try {
            StreamTokenizer st = new StreamTokenizer( new BufferedReader( new InputStreamReader( new FileInputStream( XYFile ) ) ) );
            st.eolIsSignificant( false );
            st.parseNumbers();
            st.whitespaceChars( ',', ',' );
            st.wordChars( '"', '"' );
            int tokenType = st.nextToken();
            int read = 0;
            double x = 0.0d;
            double y = 0.0d;
            double radius = 0.0d;
            int id = 0;
            while ( tokenType != StreamTokenizer.TT_EOF ) {
                switch ( tokenType ) {
                    case StreamTokenizer.TT_NUMBER :
                        switch ( read ) {
                            case 0 :
                                x = st.nval;
                                read = 1;
                                break;
                            case 1 :
                                y = st.nval;
                                grid.addToCell( x, y, 1.0d );
                                read = 0;
                                break;
                        }
                        break;
                    default :
                        break;
                }
                tokenType = st.nextToken();
            }
        }
        catch ( java.io.IOException e ) {
            System.out.println( e + " in XYFileToGrid( XYFile( " + XYFile.toString() + " ), grid ( " + grid.toString() + " ) )" );
            System.exit( 0 );
        }
    }

}
